#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef _MSC_VER
  #include <unistd.h>
#else
  #include <direct.h>
#endif
#include <string.h>
#include <sys/stat.h>
#include <ctype.h>
#include "vhdl_tab.h"
#include "stack.h"
#include "semval.h"
#include "tstst.h"
#include "semerr.h"
#include "Tmpvar.h"
#include "bodypars.h"


extern YYSTYPE yylval;
extern DECL_STATUS decl_status;
extern struct PBGLOBAL *SemGlobal;


int IsFunctionType(int toknum, char *tokstr)
{
	int i;
	for(i=0 ; i<SemGlobal->protoscnt ; i++)
	{
		if(!stricmp(SemGlobal->fprotos[i].name, tokstr))
		{
			SemGlobal->IsResolvedFunction=1;
			strcpy(SemGlobal->LastResName, tokstr);
			return 1;
		}
	}
	return 0;
}


int NewFindTypeInformation( int toknum, char *tokstr, char *typestr)
{
	int retval;
	char ltokstr[BUFSIZE];
	toknum=GetOnlyType(ltokstr);
	strcpy(tokstr, ltokstr);
	decl_status.type_toknum=toknum;
	/*******        INTEGER         *******/
	if(!stricmp(tokstr, "integer"))
    {
		sprintf(typestr, "%d\t%d\t%d\t32\t-2147483647\t2147483647\t0",
			toknum, decl_status.Class, decl_status.mode);
		return 0;
    }
	/*******         CHARACTER      *******/
	if(!stricmp(tokstr, "character"))
    {
		sprintf(typestr, "%d\t%d\t%d\t8\t0\t255\t0",
			toknum, decl_status.Class, decl_status.mode);
		return 0;
    }
	/********        BIT           *******/
	if(!(stricmp(tokstr, "bit") && stricmp(tokstr, "std_ulogic") && stricmp(tokstr, "std_logic")))
    {
		sprintf(typestr, "%d\t%d\t%d\t1\t0\t1\t0",
			toknum, decl_status.Class, decl_status.mode);
		return 0;
    }
	/********         BOOLEAN     ********/
	if( !stricmp(tokstr,"boolean"))
    {
		sprintf( typestr,"%d\t%d\t%d\t1\t0\t1\t0",
			toknum, decl_status.Class, decl_status.mode);
		return 0;
    }
	/********        BIT_VECTOR   ********/
	if( !( stricmp(tokstr,"bit_vector") && stricmp(tokstr,"std_ulogic_vector") && stricmp(tokstr,"std_logic_vector")))
    {
		sprintf( typestr,"%d\t%d\t%d\t32\t0\t31\t0",
			toknum, decl_status.Class, decl_status.mode); 
		return 0;
    }
	/****             NATURAL    *********/
	if( !stricmp(tokstr,"natural"))
    {
		sprintf( typestr,"%d\t%d\t%d\t31\t0\t2147483647\t0",
			toknum, decl_status.Class, decl_status.mode);
		return 0;
    }
	/******          POSITIVE     *********/
	if( !stricmp(tokstr,"positive"))
    {
		sprintf( typestr,"%d\t%d\t%d\t31\t1\t2147483647\t0",
			toknum, decl_status.Class, decl_status.mode);
		return 0;
    }
	/*****           STRING      **********/
	if( !stricmp(tokstr,"string"))
    {
		sprintf( typestr,"%d\t%d\t%d\t1024\t1\t1024\t0",
			toknum, decl_status.Class, decl_status.mode); 
		return 0;
    }
	/* When type is not one of the predefined types.*/
	strcpy(SemGlobal->tmpactpath, SemGlobal->actpath);
	if((retval=FindTypeString(toknum, typestr, yylval.tokstr))==-1)
	{
		if(IsFunctionType(toknum, yylval.tokstr))
			return 0;
		else
		{
			if((retval=TryFindTypeStringInPack(toknum, typestr, yylval.tokstr))==-1)
			{
				yylval.line=SemGlobal->lineno;
				fprintf(stderr, "Could not find type declaration!\n");
				fprintf(stderr, "line: %d\ttoken: %s\n", yylval.line, yylval.tokstr);
				return -1;
			}
		}
	}
	return retval;
}//NewFindTypeInformation


int GetTypeStringForAlias(char *typestr)
{
	memset(typestr,'\0',128);
	/* skip keywords: in out inout buffer linkage*/
	SemGlobal->DoesLastIsBuffer=0;
	decl_status.mode=SemGlobal->defaultmode; /*  default mode like in context */
	switch( yylex())
    {
    case tok_IN: decl_status.mode=1;
		yylex();
		break;
    case tok_OUT: decl_status.mode=2;
		yylex();
		break;
    case tok_BUFFER:
		SemGlobal->DoesLastIsBuffer=1;
    case tok_LINKAGE:
    case tok_INOUT:
		decl_status.mode=3;
		yylex();
		break;
    default: break;
    }
	if(NewFindTypeInformation( yylval.toknum, yylval.tokstr, typestr)<0)
	{
		printf("Unknown type!");
		exit(-1);
	}
	else
	{
		yylex();
		return 1;
	}
	
	/* Generating type information */
	while( yylval.semval!=tok_IS && yylval.semval!=tok_RightParen )
	{
		if( yylval.semval==tok_RANGE || yylval.semval==tok_LeftParen)
			break;
		if(yylval.semval==tok_ARRAY)
			ArrayTypeInformation(yylval.toknum, yylval.tokstr, typestr);
		else
		{
			if( FindTypeInformation( yylval.toknum, yylval.tokstr, typestr)<0)
			{
				printf("Unknown type!");
				exit(-1);
			}
		}
		if( yylval.semval!=tok_IS && yylval.semval!=tok_RightParen )
			yylex();
	}
	return 1;
}//GetTypeStringForAlias

int NewGetTypeString(char *typestr)
{
	int retval;
	memset(typestr,'\0',128);
	/* skip keywords: in out inout buffer linkage*/
	SemGlobal->DoesLastIsBuffer=0;
	decl_status.mode=SemGlobal->defaultmode; /*  default mode like in context */
	switch( yylex())
    {
    case tok_IN: decl_status.mode=1;
		yylex();
		break;
    case tok_OUT: decl_status.mode=2;
		yylex();
		break;
    case tok_BUFFER:
		SemGlobal->DoesLastIsBuffer=1;
    case tok_LINKAGE:
    case tok_INOUT:
		decl_status.mode=3;
		yylex();
		break;
    default: break;
    }
	SemGlobal->IsParentNonComputable=0;
	if((retval=NewFindTypeInformation(yylval.toknum, yylval.tokstr, typestr))<0)
	{
		printf("Unknown type!");
		exit(-1);
	}
	else
	{
		if(!SemGlobal->TypeCollected)
		{
			yylex();
			SemGlobal->TypeCollected=0;
		}
		return retval;
	}
	
	/* Generating type information */
	while( yylval.semval!=tok_Semicolon && yylval.semval!=tok_RightParen )
	{
		if( yylval.semval==tok_RANGE || yylval.semval==tok_LeftParen)
			break;
		if(yylval.semval==tok_ARRAY)
			ArrayTypeInformation(yylval.toknum, yylval.tokstr, typestr);
		else
		{
			if(FindTypeInformation(yylval.toknum, yylval.tokstr, typestr)<0)
			{
				printf("Unknown type!");
				exit(-1);
			}
		}
		if( yylval.semval!=tok_Semicolon && yylval.semval!=tok_RightParen )
			yylex();
	}
	return 1;
}//NewGetTypeString

/**********************************************************************
 * Generator of string containing information about type of identifier
 **********************************************************************
 */
int GetTypeString(char *typestr)
{
	memset(typestr,'\0',128);
	/* skip keywords: in out inout buffer linkage*/
	SemGlobal->DoesLastIsBuffer=0;
	decl_status.mode=SemGlobal->defaultmode; /*  default mode like in context */
	switch( yylex())
    {
    case tok_IN: decl_status.mode=1;
		yylex();
		break;
    case tok_OUT: decl_status.mode=2;
		yylex();
		break;
    case tok_BUFFER:
		SemGlobal->DoesLastIsBuffer=1;
    case tok_LINKAGE:
    case tok_INOUT:
		decl_status.mode=3;
		yylex();
		break;
    default: break;
    }
	
	/* Generating type information */
	while( yylval.semval!=tok_Semicolon && yylval.semval!=tok_RightParen )
	{
		if( yylval.semval==tok_RANGE || yylval.semval==tok_LeftParen)
			break;
		if(yylval.semval==tok_ARRAY)
			ArrayTypeInformation(yylval.toknum, yylval.tokstr, typestr);
		else
		{
			if(FindTypeInformation(yylval.toknum, yylval.tokstr, typestr)<0)
			{
				printf("Unknown type!");
				exit(-1);
			}
		}
		if(SemGlobal->TypeCollected)
		{
			SemGlobal->TypeCollected=0;
			break;
		}
		if( yylval.semval!=tok_Semicolon && yylval.semval!=tok_RightParen )
			yylex();
	}
	return 1;
}

/*******************************************************************
 * Generator of string containing information about array type 
 * identifier
 *******************************************************************
 */

int ArrayTypeInformation( int toknum, char *tokstr, char *typestr)
{
	int left, right;
	int to_or_downto, element_type;
	/* skip ( */
	yylex();
	yylex();
	sscanf(yylval.tokstr, "%d", &left);
	to_or_downto=yylex();/* get direction: to || downto*/
	yylex();
	sscanf(yylval.tokstr, "%d", &right);
	if( ((to_or_downto==tok_TO)     && (left > right)) ||
      ((to_or_downto==tok_DOWNTO) && (left < right)) ) 
	{
		/* incorrect direction! */
		fprintf(stderr, "Incorrect direction description!\n");
		fprintf(stderr, "line: %d\ntoken: %s\n", yylval.line, yylval.tokstr);
		exit(1);
	}
	/* skip keyword of */
	yylex();
	/* 
   * get element subtype indication 
   * CAUTION! _Simple_ element subtype indication!
   */
	yylex();
	element_type=AnonymousTypeGeneration();
	sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t%d",
		toknum, decl_status.Class, decl_status.mode,
		max(right-left,left-right), min(left,right), max(left,right),
		element_type); 
	return 0;
}//ArrayTypeInformation


int is_ge_zero(int x)
{
	if(x>0)
		return x;
	if(x<0)
		return 0;
	return 0;
}//is_ge_zero


int TryFindTypeStringInPack(int toknum, char *typestr, char *tokstr)
{
	char Path[BUFSIZE], SPath[BUFSIZE];
	int l, retval;
	if(SemGlobal->LastFullType[0]==0)
		return -1;
	strcpy(Path, SemGlobal->LastFullType);
	l=strlen(Path);
	while(l--)
	{
		if(Path[l]=='.')
			break;
	}
	if(!l)
		return -1;
	Path[l]=0;
	if(!strnicmp(Path, "work.", 5) || !strnicmp(Path, "IEEE.", 5))
	{
		if(tolower(Path[0])=='w')
			sprintf(SPath, "library\\%s\\%s\n", Path+5, Path+5);
		else
			sprintf(SPath, "..\\syslib\\Ieee\\%s\\%s\n", Path+5, Path+5);
		strcpy(SemGlobal->tmpactpath, SPath);
		retval=FindTypeStringInPack(toknum, typestr, tokstr);
		return retval;
	}
	SemGlobal->LastFullType[0]=0;
	return -1;
}




/******************************************************************
 * Produces string containing all information about type          *
 * 0 means simple type information                                *
 * 1 means complex type information                               *
 * -1 means error                                                 *
 ******************************************************************/

/******************************************************************
 * Produces string containing all information about type          *
 ******************************************************************/
int FindTypeInformation(int toknum, char *tokstr, char *typestr)
{
	int retval, ltoknum, IsSemicolonRead=0;
	char ltokstr[BUFSIZE];
	ltoknum=toknum;
	toknum=GetOnlyType(ltokstr);
	if(!stricmp(yylval.tokstr, ";"))
		IsSemicolonRead=1;
	if(yylval.semval!=tok_Identifier)
		toknum=ltoknum;
	if(yylval.semval==tok_RETURN)
		strcpy(ltokstr, tokstr);
	strcpy(tokstr, ltokstr);
	decl_status.type_toknum=toknum;
	/****************************************************
	 *                       INTEGER                    * 
	 ****************************************************/
	if(!stricmp(tokstr, "integer"))
    {
		int left, right, to_or_downto;
		int left_sign, right_sign;
		if(!IsSemicolonRead)
			yylex();
		if(yylval.semval!=tok_RANGE) 
		{
			sprintf(typestr, "%d\t%d\t%d\t32\t-2147483647\t2147483647\t0",
				toknum, decl_status.Class, decl_status.mode);
			/* skip ; */
			// yylex();
			return 0;
		}
		/* get left limit */
		if( yylex()==tok_Minus)
		{
			left_sign=-1;
			yylex();
		}
		else
			left_sign=1;
		sscanf(yylval.tokstr, "%d", &left);
		left*=left_sign;
		/* skip tok_TO, tok_DOWNTO */
		yylex();
		to_or_downto=yylval.semval;
		/* get right limit */
		if( yylex()==tok_Minus)
		{
			right_sign=-1;
			yylex();
		}
		else
			right_sign=1;
		sscanf(yylval.tokstr, "%d", &right);
		right*=right_sign;
		if( to_or_downto==tok_TO)
		{
			left=max(left,MIN_VHDL_INT);
			right=min(right,MAX_VHDL_INT);
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t0",
				toknum, decl_status.Class, decl_status.mode,
        HowManyBits(is_ge_zero(right-(left<0?left:0))), /* is it null slice? */
				min(left,right), max(left,right)
				); 
		}
		else
		{
			left=min(left,MAX_VHDL_INT);
			right=max(right,MIN_VHDL_INT);
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t0",
				toknum, decl_status.Class, decl_status.mode,
        HowManyBits(is_ge_zero(left-(right<0?right:0))), /* is it null slice? */
				min(left,right), max(left,right)
				);
		}
		/* skip ; */ 
		// yylex();
		return 1;
    }
	/****************************************************
	 *                       CHARACTER                  * 
	 ****************************************************/
	
#if 1
	if(!stricmp(tokstr, "character"))
    {
		sprintf(typestr, "%d\t%d\t%d\t8\t0\t255\t0",
			toknum, decl_status.Class, decl_status.mode);
		return 0;
    }
#endif
	/****************************************************
	 *                       BIT                        *
	 ****************************************************/

	if(!(stricmp(tokstr, "bit") && stricmp(tokstr, "std_ulogic") && stricmp(tokstr, "std_logic")))
    {
		sprintf(typestr, "%d\t%d\t%d\t1\t0\t1\t0",
			toknum, decl_status.Class, decl_status.mode);
		return 0;
    }
	
	/****************************************************
	 *                       BOOLEAN                    * 
	 ****************************************************/

	if(!stricmp(tokstr, "boolean"))
    {
		sprintf(typestr, "%d\t%d\t%d\t1\t0\t1\t0",
			toknum, decl_status.Class, decl_status.mode);
		return 0;
    }

	/****************************************************
	 *                       BIT_VECTOR                 * 
	 ****************************************************/

	if(!(stricmp(tokstr, "bit_vector") && stricmp(tokstr, "std_ulogic_vector") && stricmp(tokstr, "std_logic_vector")))
    {
		int left, right, to_or_downto;
		if(yylval.semval != tok_LeftParen && yylval.semval != tok_Semicolon)
			yylex();
		if( yylval.semval != tok_LeftParen)
		{
			sprintf( typestr,"%d\t%d\t%d\t32\t0\t2147483647\t0",
				toknum, decl_status.Class, decl_status.mode); 
			return 0;
		}
		/* get left limit */
		yylex();
		sscanf(yylval.tokstr, "%d", &left);
		/* skip tok_TO, tok_DOWNTO */
		yylex();
		to_or_downto=yylval.semval;
		/* get right limit */
		yylex();
		sscanf(yylval.tokstr, "%d", &right);
		if( to_or_downto==tok_TO)
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t0",
			toknum, decl_status.Class, decl_status.mode,
			is_ge_zero(right-left+1), /* is it null slice? */
			left, right
			); 
		else
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t0",
			toknum, decl_status.Class, decl_status.mode,
			is_ge_zero(left-right+1), /* is it null slice? */
			left, right
			); 
		/* skip ) */
		yylex();
		/* skip ; */
		yylex();
		return 1;
    }
	/****************************************************
	 *                       NATURAL                    * 
	 ****************************************************/
	if(!stricmp(tokstr, "natural"))
    {
		int left, right, to_or_downto;
		if(!IsSemicolonRead)
			yylex();
		if( yylval.semval!=tok_RANGE) 
		{
			sprintf( typestr,"%d\t%d\t%d\t31\t0\t2147483647\t0",
				toknum, decl_status.Class, decl_status.mode);
			/* skip ; */
			// yylex();
			return 0;
		}
		/* get left limit */
		yylex();
		sscanf(yylval.tokstr, "%d", &left);
		/* skip tok_TO, tok_DOWNTO */
		yylex();
		to_or_downto=yylval.semval;
		/* get right limit */
		yylex();
		sscanf(yylval.tokstr, "%d", &right);
		if( to_or_downto==tok_TO)
		{
			left=max(left,MIN_VHDL_NATURAL);
			right=min(right,MAX_VHDL_NATURAL);
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t0",
				toknum, decl_status.Class, decl_status.mode,
				HowManyBits(is_ge_zero(right)), /* is it null slice? */
				//HowManyBits(is_ge_zero(right-left)), /* is it null slice? */
				min(left,right), max(left,right)
				); 
		}
		else
		{
			left=min(left,MAX_VHDL_NATURAL);
			right=max(right,MIN_VHDL_NATURAL);
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t0",
				toknum, decl_status.Class, decl_status.mode,
				HowManyBits(is_ge_zero(left)), /* is it null slice? */
				//HowManyBits(is_ge_zero(left-right)), /* is it null slice? */
				min(left,right), max(left,right)
				);
		}
		/* skip ; */ 
		// yylex();
		return 1;
    }
	/****************************************************
	 *                       POSITIVE                   * 
	 ****************************************************/
	if(!stricmp(tokstr, "positive"))
    {
		int left, right, to_or_downto;
		
		if(!IsSemicolonRead)
			yylex();
		if( yylval.semval!=tok_RANGE) 
		{
			sprintf( typestr,"%d\t%d\t%d\t31\t1\t2147483647\t0",
				toknum, decl_status.Class, decl_status.mode);
			/* skip ; */
			// yylex();
			return 0;
		}
		/* get left limit */
		yylex();
		sscanf(yylval.tokstr, "%d", &left);
		/* skip tok_TO, tok_DOWNTO */
		yylex();
		to_or_downto=yylval.semval;
		/* get right limit */
		yylex();
		sscanf(yylval.tokstr, "%d", &right);
		if( to_or_downto==tok_TO)
		{
			left=max(left,MIN_VHDL_POSITIVE);
			right=min(right,MAX_VHDL_POSITIVE);
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t0",
				toknum, decl_status.Class, decl_status.mode,
				HowManyBits(is_ge_zero(right)), /* is it null slice? */
				//HowManyBits(is_ge_zero(right-left)), /* is it null slice? */
				min(left,right), max(left,right)
				); 
		}
		else
		{
			left=min(left,MAX_VHDL_POSITIVE);
			right=max(right,MIN_VHDL_POSITIVE);
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t0",
				toknum, decl_status.Class, decl_status.mode,
				HowManyBits(is_ge_zero(left)), /* is it null slice? */
				//HowManyBits(is_ge_zero(left-right)), /* is it null slice? */
				min(left,right), max(left,right)
				);
		}
		/* skip ; */ 
		// yylex();
		return 1;
    }
	/****************************************************
	*                       STRING                     * 
	****************************************************/
	if(!stricmp(tokstr, "string"))
    {
		int left, right, to_or_downto;
		
		if(!IsSemicolonRead)
			yylex();
		if( yylval.semval != tok_LeftParen)
		{
			sprintf( typestr,"%d\t%d\t%d\t1024\t1\t1024\t0",
				toknum, decl_status.Class, decl_status.mode); 
			return 0;
		}
		/* get left limit */
		if( yylex()==tok_Minus)
		{
			semerror(ERRBADARRAYBOUNDS, NULL);
			yylex();
			left=1;
		}
		sscanf(yylval.tokstr, "%d", &left);
		/* skip tok_TO, tok_DOWNTO */
		yylex();
		to_or_downto=yylval.semval;
		/* get right limit */
		if( yylex()==tok_Minus)
		{
			semerror(ERRBADARRAYBOUNDS, NULL);
			yylex();
			right=1;
		}
		sscanf(yylval.tokstr, "%d", &right);
		if( to_or_downto==tok_TO)
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t0",
			toknum, decl_status.Class, decl_status.mode,
			is_ge_zero(right-left+1), /* is it null slice? */
			left, right
			); 
		else
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t0",
			toknum, decl_status.Class, decl_status.mode,
			is_ge_zero(left-right+1), /* is it null slice? */
			left, right
			); 
		/* skip ) */
		yylex();
		/* skip ; */
		yylex();
		return 1;
    }


	/*
	 * When type is not one of the predefined types...
	 */
	strcpy(SemGlobal->tmpactpath, SemGlobal->actpath);
	if((retval=FindTypeString(toknum, typestr, yylval.tokstr))==-1)
	{
		if((retval=TryFindTypeStringInPack(toknum, typestr, yylval.tokstr))==-1)
		{
			yylval.line=SemGlobal->lineno;
			fprintf(stderr, "Could not find type declaration!\n");
			fprintf(stderr, "line: %d\ttoken: %s\n", yylval.line, yylval.tokstr);
			return -1;
		}
		//fclose(typesfile);
	}
	return retval;
}//FindTypeInformation


int FindTypeStringInPack( int toknum, char *typestr, char *looktype)
{
	FILE *typesfile;
	char line[16384], typespath[PATHSIZE];
	int len;
	len = strlen(SemGlobal->tmpactpath);
	strncpy(typespath, SemGlobal->tmpactpath, len-1);
	typespath[len-1]=0;
	if(!strnicmp(typespath, "..\\syslib\\Ieee\\std_logic_", 25))
	{
		if( !strnicmp(typespath, "..\\syslib\\Ieee\\std_logic_arith", 30) ||
			!strnicmp(typespath, "..\\syslib\\Ieee\\std_logic_unsigned", 33) ||
			!strnicmp(typespath, "..\\syslib\\Ieee\\std_logic_signed", 31)
			)
		{
			if(toknum==UNSIGNEDLV || !stricmp(looktype, "unsigned"))
			{
				strcpy(SemGlobal->basetype, "ARRAY");
				sprintf(typestr, "214\t%d\t%d\t32\t0\t2147483647\t210", decl_status.Class, decl_status.mode);
				return 1;
			}
			if(toknum==SIGNEDLV || !stricmp(looktype, "signed"))
			{
				strcpy(SemGlobal->basetype, "ARRAY");
				sprintf(typestr, "213\t%d\t%d\t32\t0\t2147483647\t210", decl_status.Class, decl_status.mode);
				return 1;
			}
			if(toknum==SMALL_INTLV || !stricmp(looktype, "small_int"))
			{
				strcpy(SemGlobal->basetype, "INT");
				sprintf(typestr, "215\t%d\t%d\t1\t1\t0\t201", decl_status.Class, decl_status.mode);
				return 1;
			}
		}

	}
	strcat(typespath, ".tdf");
	typesfile=fopen(typespath,"rt");
	if(typesfile!=NULL)
	{
		while( fgets( line, 16383, typesfile))
		{
			int typenumber, typelength, leftlimit, rightlimit, parenttype;
			char typename[256], basetype[256]; /* useless, for (s)scan(f)ning only */ 
			
			sscanf(line,"%d%s%s%d%d%d%d", &typenumber, typename, basetype, &typelength,
				&leftlimit, &rightlimit, &parenttype);
			if(stricmp(looktype, typename) && (typenumber!=toknum))
				continue;
			strcpy(SemGlobal->basetype, basetype);
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t%d",
				typenumber, decl_status.Class, decl_status.mode,
				typelength, leftlimit, rightlimit, parenttype);
			strcpy(SemGlobal->LastTypeName, typename);
			fclose(typesfile);
			return 1;
		}
		fclose(typesfile);
	}
	strncpy(typespath, SemGlobal->tmpactpath, len-1);
	typespath[len-1]=0;
	strcat(typespath, ".tnc");
	typesfile=fopen(typespath,"rt");
	if(typesfile!=NULL)
	{
		while( fgets( line, 16383, typesfile))
		{
			int typenumber, typelength, leftlimit, rightlimit, parenttype;
			char typename[256], basetype[256]; /* useless, for (s)scan(f)ning only */ 
			
			sscanf(line,"%d%s%s%d%d%d%d", &typenumber, typename, basetype, &typelength,
				&leftlimit, &rightlimit, &parenttype);
			if(stricmp(looktype, typename) && (typenumber!=toknum))
				continue;
			strcpy(SemGlobal->basetype, basetype);
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t%d",
				typenumber, decl_status.Class, decl_status.mode,
				typelength, leftlimit, rightlimit, parenttype);
			fclose(typesfile);
			SemGlobal->IsParentNonComputable=1;
			return 2;
		}
		fclose(typesfile);
	}
	return 0;
}//FindTypeStringInPack



int LookForUses(char *path, int toknum, char *typestr, char *looktype)
{
	FILE *srcf;
	char bufor[BUFSIZE];
	int isPaths=0;
	if((srcf=fopen(path, "rt"))==NULL)
		return -1;
	while(fgets(bufor, BUFSIZE, srcf))
	{
		if(bufor[0]=='#')
		{
			if(bufor[2]=='3' || bufor[2]=='7')
			{
				if(SemGlobal->IsSTLACompiled)
					isPaths=0;
				else
					isPaths=1;
			}
			else
				isPaths=0;
			continue;
		}
		if(isPaths==1)
		{
			int retval;
			strcpy(SemGlobal->tmpactpath, bufor);
			retval=FindTypeStringInPack(toknum, typestr, looktype);
			if(retval==1 || retval==2)
			{
				fclose(srcf);
				return retval;
			}
		}
	}
	fclose(srcf);
	return 0;
}//LookForUses

int CutPath(int toknum, char *typestr, char *looktype)
{
	unsigned len, i;
	FILE *typesfile;
	int retval;
	char path[PATHSIZE], tmppath[PATHSIZE];
	sprintf(path, "%s\\clause.use", SemGlobal->tmpactpath);
	strcpy(tmppath, SemGlobal->tmpactpath);
	if(LookForUses(path, toknum, typestr, looktype)==1)
		return 1;
	strcpy(SemGlobal->tmpactpath, tmppath);
	sprintf(path, "%s.val", SemGlobal->tmpactpath);
	retval=LookForUses(path, toknum, typestr, looktype);
	if(retval==1 || retval==2)
		return retval;
	strcpy(SemGlobal->tmpactpath, tmppath);
	len=strlen(SemGlobal->tmpactpath);
	i=len;
	while(i>0)
	{
		if(SemGlobal->tmpactpath[i]=='\\')
		{
			SemGlobal->tmpactpath[i]=0;
			break;
		}
		i--;
	}
	if(i>8)
	{
		while(i>0)
		{
			if(SemGlobal->tmpactpath[i]!=0 && !isdigit(SemGlobal->tmpactpath[i]))
			{
				if(SemGlobal->tmpactpath[i]=='_')
					SemGlobal->tmpactpath[i]=0;
				else
					SemGlobal->tmpactpath[i+1]=0;
				break;
			}
			if(SemGlobal->tmpactpath[i]=='_')
			{
				SemGlobal->tmpactpath[i]=0;
				break;
			}
			i--;
		}
	}
	if(!stricmp("toplevel", SemGlobal->tmpactpath))
		return -1;//zostalo tylko toplevel
	if(!stricmp("library", SemGlobal->tmpactpath))
	{
		char PackageName[PATHSIZE], PackageTypes[PATHSIZE];
		char line[16384];
		strcpy(SemGlobal->tmpactpath, SemGlobal->actpath);
		len=strlen(SemGlobal->tmpactpath);
		i=len;
		while(i>0)
		{
			if(SemGlobal->tmpactpath[i]=='\\')
			{
				SemGlobal->tmpactpath[i]=0;
				break;
			}
			i--;
		}
		while(i>0)
		{
			if(SemGlobal->tmpactpath[i]=='\\')
				break;
			i--;
		}
		strcpy(PackageName, SemGlobal->tmpactpath+i);
		sprintf(PackageTypes, "%s%s.tdf", SemGlobal->tmpactpath, PackageName);
		typesfile=fopen(PackageTypes,"rt");
		if(typesfile!=NULL)
		{
			while( fgets( line, 16383, typesfile))
			{
				int typenumber, typelength, leftlimit, rightlimit, parenttype;
				char typename[256], basetype[256]; /* useless, for (s)scan(f)ning only */ 
				
				sscanf(line,"%d%s%s%d%d%d%d", &typenumber, typename, basetype, &typelength,
					&leftlimit, &rightlimit, &parenttype);
				if( typenumber != toknum)
					continue;
				strcpy(SemGlobal->basetype, basetype);
				sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t%d",
					typenumber, decl_status.Class, decl_status.mode,
					typelength, leftlimit, rightlimit, parenttype);
				fclose(typesfile);
				return 1;
			}
		}
		sprintf(PackageTypes, "%s%s.tnc", SemGlobal->tmpactpath, PackageName);
		typesfile=fopen(PackageTypes,"rt");
		if(typesfile!=NULL)
		{
			while( fgets( line, 16383, typesfile))
			{
				int typenumber, typelength, leftlimit, rightlimit, parenttype;
				char typename[256], basetype[256]; /* useless, for (s)scan(f)ning only */ 
				
				sscanf(line,"%d%s%s%d%d%d%d", &typenumber, typename, basetype, &typelength,
					&leftlimit, &rightlimit, &parenttype);
				if( typenumber != toknum)
					continue;
				strcpy(SemGlobal->basetype, basetype);
				sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t%d",
					typenumber, decl_status.Class, decl_status.mode,
					typelength, leftlimit, rightlimit, parenttype);
				fclose(typesfile);
				return 2;
			}
		}
		return -1;//zostalo tylko library (w pakiecie) i nie znaleziono zadnego typu
	}
	return 0;
}


int FindTypeString( int toknum, char *typestr, char *looktype)
{
	FILE *typesfile;
	char line[16384], typespath[PATHSIZE];
	int retval;
	sprintf(typespath, "%s.tdf", SemGlobal->tmpactpath);
	typesfile=fopen(typespath,"rt");
	if(typesfile!=NULL)
	{
		while( fgets( line, 16383, typesfile))
		{
			int typenumber, typelength, leftlimit, rightlimit, parenttype;
			char typename[256], basetype[256]; /* useless, for (s)scan(f)ning only */ 
			
			sscanf(line,"%d%s%s%d%d%d%d", &typenumber, typename, basetype, &typelength,
				&leftlimit, &rightlimit, &parenttype);
			if( typenumber != toknum)
				continue;
			strcpy(SemGlobal->basetype, basetype);
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t%d",
				typenumber, decl_status.Class, decl_status.mode,
				typelength, leftlimit, rightlimit, parenttype);
			fclose(typesfile);
			return 1;
		}
		fclose(typesfile);
	}
	sprintf(typespath, "%s.tnc", SemGlobal->tmpactpath);
	typesfile=fopen(typespath,"rt");
	if(typesfile!=NULL)
	{
		while( fgets( line, 16383, typesfile))
		{
			int typenumber, typelength, leftlimit, rightlimit, parenttype;
			char typename[256], basetype[256]; /* useless, for (s)scan(f)ning only */ 
			
			sscanf(line,"%d%s%s%d%d%d%d", &typenumber, typename, basetype, &typelength,
				&leftlimit, &rightlimit, &parenttype);
			if( typenumber != toknum)
				continue;
			strcpy(SemGlobal->basetype, basetype);
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t%d",
				typenumber, decl_status.Class, decl_status.mode,
				typelength, leftlimit, rightlimit, parenttype);
			fclose(typesfile);
			return 2;
		}
		fclose(typesfile);
	}
	retval=CutPath(toknum, typestr, looktype);
	if(retval==-1)
		return -1;
	if(retval==1 || retval==2)//znaleziono typ w sciezce z use
		return retval;
	return 	FindTypeString(toknum, typestr, looktype);
}


int FindLexemStringInPack( int toknum, char *typestr)
{
	FILE *valuesfile, *varfile;
	char line[16384], valuespath[PATHSIZE], looktype[BUFSIZE], bufor[BUFSIZE];
	int len, numval, isFound=0;
	varfile=fopen("alu.var", "rt");//pozniej staly plik
	while(fgets(bufor, BUFSIZE, varfile))
	{
		sscanf(bufor, "%d\t%s", &numval, looktype);
		if(numval+200==toknum)
		{
			isFound=1;
			break;
		}
	}
	fclose(varfile);
	if(!isFound)
		return 0;
	len = strlen(SemGlobal->tmpactpath);
	strncpy(valuespath, SemGlobal->tmpactpath, len-1);
	valuespath[len-1]=0;
	strcat(valuespath, ".val");
	valuesfile=fopen(valuespath,"rt");
	if(valuesfile!=NULL)
	{
		while( fgets( line, 16383, valuesfile))
		{
			int typenumber, typelength, leftlimit, rightlimit, parenttype;
			char typename[256], basetype[256]; /* useless, for (s)scan(f)ning only */ 
			
			sscanf(line,"%d%s%s%d%d%d%d", &typenumber, typename, basetype, &typelength,
				&leftlimit, &rightlimit, &parenttype);
			if(stricmp(looktype, typename))
				continue;
			strcpy(SemGlobal->basetype, basetype);
			sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t%d",
				typenumber, decl_status.Class, decl_status.mode,
				typelength, leftlimit, rightlimit, parenttype);
			fclose(valuesfile);
			return 1;
		}
		fclose(valuesfile);
	}
	strncpy(valuespath, SemGlobal->tmpactpath, len-1);
	valuespath[len-1]=0;
	strcat(valuespath, ".id");
	valuesfile=fopen(valuespath,"rt");
	if(valuesfile!=NULL)
	{
		int lBaseType, lvalue;
		while( fgets( line, 16383, valuesfile))
		{
			int typenumber;
			char *t, *s;
			sscanf(line, "%d", &typenumber);
			if( typenumber != toknum) continue;
			t=line;
			while(*t!=' ' && *t!='\t' && *t!='\n') t++;
			if(*t=='\n') continue;
			t++;
			while(*t!='\n')
			{
				s=t;
				while(*t!=' ' && *t!='\t' && *t!='\n') t++;
				if(*t=='\n') break;
				t++;
				while(*t!=' ' && *t!='\t' && *t!='\n') t++;
				sscanf(s, "%d %d", &lBaseType, &lvalue);
				if(SemGlobal->EnumBaseType==lBaseType)
				{
					sprintf(typestr, "%d\t%d\t%d\tLENUM", toknum, lBaseType, lvalue);
					fclose(valuesfile);
					return 1;
				}
				if(*t=='\n') break;
				t++;
			}
		}
		fclose(valuesfile);
	}
	return 0;
}//FindLexemStringInPack


int LookForUsesForLexem(char *path, int toknum, char *typestr)
{
	FILE *srcf;
	char bufor[BUFSIZE];
	int isPaths=0;
	if((srcf=fopen(path, "rt"))==NULL)
		return -1;
	while(fgets(bufor, BUFSIZE, srcf))
	{
		if(bufor[0]=='#')
		{
			if(bufor[2]=='3' || bufor[2]=='7')
			{
				if(SemGlobal->IsSTLACompiled)
					isPaths=0;
				else
					isPaths=1;
			}
			else
				isPaths=0;
			continue;
		}
		if(isPaths==1)
		{
			strcpy(SemGlobal->tmpactpath, bufor);
			if(FindLexemStringInPack(toknum, typestr)==1)
			{
				fclose(srcf);
				return 1;
			}
		}
	}
	fclose(srcf);
	return 0;
}

int CutPathForLexem(int toknum, char *typestr)
{
	unsigned len, i;
	char path[PATHSIZE];
	if(strlen(SemGlobal->tmpactpath)<1)
		return -1;//nie zostalo nic w sciezkce przeszukiwania
	sprintf(path, "%s.val", SemGlobal->tmpactpath);
	if(LookForUsesForLexem(path, toknum, typestr)==1)
		return 1;
	len=strlen(SemGlobal->tmpactpath);
	i=len;
	while(i>0)
	{
		if(SemGlobal->tmpactpath[i]=='\\')
		{
			SemGlobal->tmpactpath[i]=0;
			break;
		}
		i--;
	}
	if(i>8)
	{
		while(i>0)
		{
			if(SemGlobal->tmpactpath[i]!=0 && !isdigit(SemGlobal->tmpactpath[i]))
			{
				if(SemGlobal->tmpactpath[i]=='_')
					SemGlobal->tmpactpath[i]=0;
				else
					SemGlobal->tmpactpath[i+1]=0;
				break;
			}
			if(SemGlobal->tmpactpath[i]=='_')
			{
				SemGlobal->tmpactpath[i]=0;
				break;
			}
			i--;
		}
	}
	if(!stricmp("toplevel", SemGlobal->tmpactpath))
		return -1;//zostalo tylko toplevel
	if(!stricmp("library", SemGlobal->tmpactpath))
		return -1;//zostalo tylko library (w pakiecie)
	if(!stricmp("..\\syslib", SemGlobal->tmpactpath))
		return -1;//zostalo tylko syslib (w pakiecie)
	if(strlen(SemGlobal->tmpactpath)<1)
		return -1;//nie zostalo nic w sciezkce przeszukiwania
//	if(!stricmp("..", SemGlobal->tmpactpath))
//		return -1;//zostalo tylko .. przed syslibem 
	return 0;
}

void DeleteLastChar(char *str)
{
	int tmplen=strlen(str);
	str[strlen(str)-1]=0;
}
/* Funkcja do znajdowania ciagu znakow przedstawiajacego
 * wartosci semantyczne dla podanego lexemu */

int FindLexemInformation( int toknum, char *typestr)
{
	FILE *typesfile;
	char line[16384], typespath[PATHSIZE];
	int retval;
	sprintf(typespath, "%s.val", SemGlobal->tmpactpath);
	typesfile=fopen(typespath,"rt");
	if(typesfile!=NULL)
	{
		while( fgets( line, 16383, typesfile))
		{
			int typenumber;
			char typename[256];
			sscanf(line, "%d%s", &typenumber, typename);
			if( typenumber != toknum)
				continue;
			DeleteLastChar(line);
			strcpy(typestr, line);
			fclose(typesfile);
			return 1;
		}
	}
	sprintf(typespath, "%s.tnc", SemGlobal->tmpactpath);
	typesfile=fopen(typespath,"rt");
	if(typesfile!=NULL)
	{
		while( fgets( line, 16383, typesfile))
		{
			int typenumber;
			char typename[256];
			sscanf(line, "%d%s", &typenumber, typename);
			if( typenumber != toknum)
				continue;
			DeleteLastChar(line);
			strcpy(typestr, line);
			fclose(typesfile);
			return 2;
		}
	}
	sprintf(typespath, "%s.id", SemGlobal->tmpactpath);
	typesfile=fopen(typespath,"rt");
	if(typesfile!=NULL)
	{
		int lBaseType, lvalue;
		while( fgets( line, 16383, typesfile))
		{
			int typenumber;
			char *t, *s;
			sscanf(line, "%d", &typenumber);
			if( typenumber != toknum) continue;
			t=line;
			while(*t!=' ' && *t!='\t' && *t!='\n') t++;
			if(*t=='\n') continue;
			t++;
			while(*t!='\n')
			{
				s=t;
				while(*t!=' ' && *t!='\t' && *t!='\n') t++;
				if(*t=='\n') break;
				t++;
				while(*t!=' ' && *t!='\t' && *t!='\n') t++;
				sscanf(s, "%d %d", &lBaseType, &lvalue);
				if(SemGlobal->EnumBaseType==lBaseType)
				{
					sprintf(typestr, "%d\t%d\t%d\tLENUM", toknum, lBaseType, lvalue);
					fclose(typesfile);
					return 1;
				}
				if(*t=='\n') break;
				t++;
			}
		}
	}
	retval=CutPathForLexem(toknum, typestr);
	if(retval==-1)
		return -1;
	if(retval==1)//znaleziono typ w sciezce z use
		return 1;
	return FindLexemInformation(toknum, typestr);
}

/*
 * Looking for declared internal type
 */
int RDFindTypeString( int toknum, char *typestr)
{
#if 0
	fprintf(stderr, "This kind of declaration is not yet implemented!\n");
	fprintf(stderr, "line: %d\ntoken: %s\n", yylval.line, yylval.tokstr);
	exit(1);
	return -1;
#else
	FILE *typesfile;
	char line[16384], typespath[PATHSIZE];
	sprintf(typespath, "%s.tdf", SemGlobal->actpath);
	typesfile=fopen(typespath,"rt");
	
	while( fgets( line, 16383, typesfile))
	{
		int typenumber, typelength, leftlimit, rightlimit, parenttype;
		char typename[256], basetype[256]; /* useless, for (s)scan(f)ning only */ 
		
		sscanf(line,"%d%s%s%d%d%d%d", &typenumber, typename, basetype, &typelength,
			&leftlimit, &rightlimit, &parenttype);
		if( typenumber != toknum)
			continue;
			/*
			* Found exact type in types.tdf 
			* CAUTION! Only strict definition is support here (yet)
		*/
		strcpy(SemGlobal->basetype, basetype);
		sprintf( typestr,"%d\t%d\t%d\t%d\t%d\t%d\t%d",
			typenumber, decl_status.Class, decl_status.mode,
			typelength, leftlimit, rightlimit, parenttype);
		fclose(typesfile);
		return 1;
	}
	
	fprintf(stderr, "Could not find type declaration!\n");
	fprintf(stderr, "line: %d\ntoken: %s\n", yylval.line, yylval.tokstr);
	fclose(typesfile);
	return -1;
#endif
}



//Parsowanie list parametrow procedury
int ParseParameterList( FILE *parameters)
{
	int idnum;
	char idstr[255];
	stack_id portlist=NULL;
	stack_id tmplist=NULL;
	char typestr[65535];
	
	/* skip left parenthesis: */
	if( yylex()!=tok_LeftParen)
		return 0;
	/* Parsing parameter list */
	while( yylex() != tok_IS)
   {
		if( yylval.semval==tok_TYPE)
			break;
		switch(yylval.semval)
		{
		case tok_Comma      :
		case tok_Colon      :break;
			/* Build the stack of identifiers with the same type */
		case tok_Identifier :stack_push(&portlist, yylval.toknum, yylval.tokstr);
			break;
		}
		if( yylval.semval != tok_Colon)
			continue;
		/* Type description expected */
		GetTypeString(typestr);
		/* Reducing stack of identifiers */
		/* Reversing stack of identifiers 
		 * portlist - source stack
		 * tmplist  - reversed stack 
		 */
		while( stack_pop( &portlist, &idnum, idstr)!=EMPTY_STACK)       
			stack_push( &tmplist, idnum, idstr);
		/* Reducing reversed stack of identifiers */
		while( stack_pop( &tmplist, &idnum, idstr)!=EMPTY_STACK)
			fprintf(parameters,"%d\t%s\t%s\n",idnum,idstr,typestr);
		if(yylval.semval==tok_IS)
			break;
    }
	/* Procedure parameter list succesfully parsed */
	return 0;
}


//Parsowanie list parametrow funkcji
int ParseFunctionParameterList( FILE *parameters)
{
	int idnum;
	char idstr[255];
	stack_id portlist=NULL;
	stack_id tmplist=NULL;
	char typestr[65535];
	
	/* skip left parenthesis: */
	if( yylex()!=tok_LeftParen)
		return 0;
	/* Parsing parameter list */
	while( yylex() != tok_RETURN  )
    {
		if( yylval.semval==tok_TYPE)
			break;
		switch(yylval.semval)
		{
		case tok_Comma      :
		case tok_Colon      :break;
			/* Build the stack of identifiers with the same type */
		case tok_Identifier :stack_push(&portlist, yylval.toknum, yylval.tokstr);
			break;
		}
		if( yylval.semval != tok_Colon)
			continue;
		/* Type description expected */
		GetTypeString(typestr);
		/* Reducing stack of identifiers */
		/* Reversing stack of identifiers 
		 * portlist - source stack
		 * tmplist  - reversed stack 
		*/
		while( stack_pop( &portlist, &idnum, idstr)!=EMPTY_STACK)       
			stack_push( &tmplist, idnum, idstr);
		/* Reducing reversed stack of identifiers */
		while( stack_pop( &tmplist, &idnum, idstr)!=EMPTY_STACK)
			fprintf(parameters,"%d\t%s\t%s\n",idnum,idstr,typestr);
		if(yylval.semval==tok_RETURN)
			break;
    }
	/* Function parameter list succesfully parsed */
	return 0;
}//ParseFunctionParameterList

//Parsowanie list parametrow funkcji
int ParseProcedureParameterList( FILE *parameters)
{
	int idnum, ParenCnt=1;
	char idstr[255];
	stack_id portlist=NULL;
	stack_id tmplist=NULL;
	char typestr[65535];
	
	/* skip left parenthesis: */
	if( yylex()!=tok_LeftParen)
		return 0;
	/* Parsing parameter list */
	while(ParenCnt)
    {
		yylex();
		if( yylval.semval==tok_TYPE)
			break;
		switch(yylval.semval)
		{
		case tok_LeftParen	:ParenCnt++;break;
		case tok_RightParen	:ParenCnt--;break;
		case tok_Comma			:
		case tok_Colon			:break;
			/* Build the stack of identifiers with the same type */
		case tok_Identifier :stack_push(&portlist, yylval.toknum, yylval.tokstr);
			break;
		}
		if(yylval.semval != tok_Colon)
			continue;
		/* Type description expected */
		GetTypeString(typestr);
		if(yylval.semval==tok_RightParen)
			ParenCnt--;
		/* Reducing stack of identifiers */
		/* Reversing stack of identifiers 
		 * portlist - source stack
		 * tmplist  - reversed stack 
		*/
		while( stack_pop( &portlist, &idnum, idstr)!=EMPTY_STACK)       
			stack_push( &tmplist, idnum, idstr);
		/* Reducing reversed stack of identifiers */
		while( stack_pop( &tmplist, &idnum, idstr)!=EMPTY_STACK)
			fprintf(parameters,"%d\t%s\t%s\n",idnum,idstr,typestr);
    }
	/* Function parameter list succesfully parsed */
	return 0;
}//ParseProcedureParameterList

void WritePredefinedTypes(char *fname)
{
	FILE *tfile=fopen(fname, "wt");
	if(!tfile)
		return;
	fprintf( tfile, "201\tinteger\tINT\t32\t-2147483647\t2147483647\t0\t0\n");
	fprintf( tfile, "202\tbit\tBIT\t1\t0\t1\t0\t0\n");
	fprintf( tfile, "203\tbit_vector\tBVECT\t32\t0\t2147483647\t202\t0\n");
	fprintf( tfile, "204\tboolean\tBOOL\t1\t0\t1\t0\t0\n");
	fprintf( tfile, "205\tcharacter\tCHAR\t8\t0\t255\t0\t0\n");
	fprintf( tfile, "206\tstring\tSTRING\t1024\t0\t1024\t205\t0\n");
	fprintf( tfile, "207\tnatural\tINT\t31\t0\t2147483647\t0\t201\n");
	fprintf( tfile, "208\tpositive\tINT\t31\t1\t2147483647\t0\t201\n");
	fprintf( tfile, "209\tstd_ulogic\tBIT\t1\t0\t1\t0\t0\n");
	fprintf( tfile, "210\tstd_logic\tBIT\t1\t0\t1\t0\t209\n");
	fprintf( tfile, "211\tstd_ulogic_vector\tBVECT\t32\t0\t2147483647\t209\t0\n");
	fprintf( tfile, "212\tstd_logic_vector\tBVECT\t32\t0\t2147483647\t210\t0\n");
	fprintf( tfile, "216\tUX01\tBIT\t1\t0\t1\t0\t209\n");
	fprintf( tfile, "217\tUX01Z\tBIT\t1\t0\t1\t0\t209\n");
	fprintf( tfile, "218\tX01Z\tBIT\t1\t0\t1\t0\t209\n");
	fprintf( tfile, "219\tX01\tBIT\t1\t0\t1\t0\t209\n");
	fclose(tfile);
}

int AnonymousTypeGeneration(void)
{
	FILE *anontypes;
	char typename[64], typedescr[16384], anonpath[PATHSIZE];
	int typenumber, localtypenumber;
	
	/*
	* Is this a simple type dclaration?
	*/
	if( !GetTypeString(typedescr))
    {
		sscanf(typedescr,"%d",&typenumber);
		return typenumber;
    }
	/*
	* This is a complex type declaration
	*/
	sprintf(anonpath, "%s.ann", SemGlobal->tmpactpath);
	anontypes=fopen(anonpath,"rt+");
	if(!anontypes)
		anontypes=fopen(anonpath,"wt+");
	//SkipFile(anontypes);
	assert(anontypes);
	typename[0]='\0';
	typenumber=TmpVarNumber(typename);
	localtypenumber=ConvertVarToTypeDecl(typedescr);
	if( localtypenumber)
    {
		printf("Internal error:\nCould not find type declaration for type number "
			"%d.\n",localtypenumber);
		exit(-1);
    }
	fprintf(anontypes,"%d\t%s\t%s\n",typenumber,typename,typedescr);
	fclose(anontypes);
	return typenumber;
}


int ConvertVarToTypeDecl(char *typedescr)
{
	FILE *typesfile=fopen("types.tdf","rt");
	int typenum, idclass, mode, bitnum, beg, end;
	int localtypenum, found=0;
	char typename[64], typeclass[64], line[1024];
	
	sscanf(typedescr,"%d\t%d\t%d\t%d\t%d\t%d", &typenum, &idclass, &mode,
		&bitnum, &beg, &end);
	while(fgets(line,1023,typesfile))
    {
		sscanf(line,"%d\t%s\t%s", &localtypenum, typename, typeclass);
		if(localtypenum==typenum)
		{
			found=1;
			break;
		}
    }
	if(found)
		sprintf(typedescr,"%s\t%d\t%d\t%d\t%d",
		typeclass,bitnum,beg,end,localtypenum);
	fclose(typesfile);
	return ( found==0? typenum: 0);
}


int ConstantBitString( int toknum, char *bitstring)
{
	FILE *cfile=fopen("alu.con", "rt");
	char line[1024];
	assert(cfile);
	while( fgets(line,1023,cfile))
    {
		int token;
		sscanf(line,"%d", &token);
		if( token==toknum )
		{
			char name[256];
			int nrtype, idclass, mode, nrofbits, left, right; 
			sscanf(line,"%d%s%d%d%d%d%d%d%s", &token, name, &nrtype, &idclass,
				&mode, &nrofbits, &left, &right, bitstring);
			fclose(cfile);
			return 0;
		}
    }
	fclose(cfile);
	return -1;
} 

void ChangeLHTo01(char *bitstring)
{
	int l=0;
	while(bitstring[l]!=0)
	{
		if(toupper(bitstring[l])=='L')
			bitstring[l]='0';
		if(toupper(bitstring[l])=='H')
			bitstring[l]='1';
		l++;
	}
}//ChangeLHTo01


int BitStringToIntValue(char *bitstring)
{
	int result=0;
	int i, nrofbits=strlen(bitstring);
	assert( 0<=nrofbits && nrofbits<=33);
	for( i=0; i<nrofbits; i++)
		result=(result << 1) + (bitstring[i]=='1'? 1: 0);
	return result;
}//BitStringToIntValue


int InEquationParse(void)
{
	int gval;
	char bitstring[256], tokstr[BUFSIZE];
	if(yylval.semval>=tok_AbstractLit&&yylval.semval<=tok_BitStringLit)
	{
		ConstantBitString( yylval.toknum, bitstring);
		gval=BitStringToIntValue(bitstring);;
	}
	else
	{
		char bufor[BUFSIZE];
		int value;
		strcpy(SemGlobal->tmpactpath, SemGlobal->actpath);
		sprintf(bufor, "%d\tidentifier:\t%d", yylval.toknum, yylval.tokstr);
		AliasCheckAndReplace(bufor);
		value=getvalue(bufor, NULL);
		if(value!=yylval.toknum)
			yylval.toknum=value;
		FindLexemInformation(yylval.toknum, tokstr);
		ValueFromLine(tokstr, &gval);
	}
	return gval;
}


int ParseBoundsToRightParen(int *left, int *right, int *direction)
{
	int left_sign, right_sign;
	char bitstring[256];
	//int Arg1Legal=0, Arg2Legal=0;
	if( yylex()==tok_Minus)
	{
		left_sign=-1;
		yylex();
	}
	else
		left_sign=1;
	*left=left_sign*InEquationParse();

	while(yylex()!=tok_TO)
	{
		if(yylval.semval==tok_DOWNTO)
			break;
		if(yylval.semval==tok_Semicolon)
			return 1;
		if(yylval.semval==tok_Minus)
		{
			yylex();
			*left=*left-InEquationParse();;
			continue;
		}
		if(yylval.semval==tok_Plus)
		{
			*left=*left+InEquationParse();
			continue;
		}
	}

	*direction=( yylval.semval == tok_TO ? 1 : -1);
	/* right limit */
	if( yylex()==tok_Minus)
	{
		right_sign=-1;
		yylex();
	}
	else
		right_sign=1;
	ConstantBitString( yylval.toknum, bitstring);
	*right=right_sign*BitStringToIntValue(bitstring);
	yylex(); /* skip tok_RightParen */
	return 0;
}//ParseBoundsToRightParen


void GenerateTypeFromYYList()
{
//	int i;
//	for(i=0 ; i<SemGlobal->LexemsInListCnt ; i++)
//		;
	return;
}

int ParseConstantDeclaration( FILE *semantics)
{
	int constnumber, typeofconstant, value;
	char NotComputablePath[PATHSIZE], TmpNotComputablePath[PATHSIZE], tmpstr[BUFSIZE];
	char constname[256], bitstring[BUFSIZE], bufor[BUFSIZE], *OldBufor, *OldBufor2;
	int nrofbits, left=-1, right=-1, direction=1, left_sign=1, right_sign=1;
	FILE *ncmpfile, *tmpncmpfile;
	stack_id ConstList=NULL;
	stack_id ConstList2=NULL;
	yylval.semval=0;
	while(yylval.semval!=tok_Colon)
	{
		yylex();
		strcpy(constname, yylval.tokstr);
		stack_push(&ConstList, yylval.toknum, constname);
		/* read tok_Colon or tok_Comma*/
		yylex();
	}
	yylex();
	typeofconstant=yylval.toknum;
	if(yylval.semval==tok_FILE || !stricmp(yylval.tokstr, "time") || !stricmp(yylval.tokstr, "real"))
	{
		ReadToSemiC();
		return 0;
	}
	/* skip tok_VarAsgn ? */ 
	/* Added 23rd of November 1999 */
	if( yylex() != tok_VarAsgn)
		switch(yylval.semval)
	{
		case tok_RANGE: /* integer */
			/* left limit */
			if( yylex()==tok_Minus)
			{
				left_sign=-1;
				yylex();
			}
			else
				left_sign=1;
			ConstantBitString( yylval.toknum, bitstring);
			left=left_sign*BitStringToIntValue(bitstring);
			direction=( yylex() == tok_TO ? 1 : -1);
			/* right limit */
			if( yylex()==tok_Minus)
			{
				right_sign=-1;
				yylex();
			}
			else
				right_sign=1;
			ConstantBitString( yylval.toknum, bitstring);
			right=right_sign*BitStringToIntValue(bitstring);
			yylex(); /* finally skip tok_VarAsgn */
			break;
		case tok_LeftParen: /* bit_vector */
			/* left limit */
			if(ParseBoundsToRightParen(&left, &right, &direction))
			{
				GenerateTypeFromYYList();
			}
			yylex(); /* finally skip tok_VarAsgn */
			break;
		default:
			printf("Unsupported constant declaration!\n");
			if(yylval.line<SemGlobal->lineno)
				yylval.line=SemGlobal->lineno;
			printf("Error appeared in line %d!\n", yylval.line);
		  exit(-1);
	}
	/*
	* everything happens here: 
	*/  
	fgets(bufor, BUFSIZE, SemGlobal->srcf);
	value=getvalue(bufor, NULL);
	yylval.toknum=value;
	getname(bufor, tmpstr);
	strcpy(yylval.tokstr, tmpstr);
	OldBufor=strdup(bufor);
	//yylex();

	switch(typeofconstant)
    {
    case 201: /* integer */
		ConstantBitString( yylval.toknum, bitstring);
		left=( left==-1 ? MIN_VHDL_INT : left);
		right=( right==-1 ? MAX_VHDL_INT : right);
		nrofbits=HowManyBitsInt(right, left);
		break;
    case 207: /* natural */
		ConstantBitString( yylval.toknum, bitstring);
		left=( left==-1 ? 0 : left);
		right=( right==-1 ? MAX_VHDL_INT : right);
		nrofbits=HowManyBits(direction*(right-left));
		break;
    case 208: /* positive */
		ConstantBitString( yylval.toknum, bitstring);
		left=( left==-1 ? 1 : left);
		right=( right==-1 ? MAX_VHDL_INT : right);
		nrofbits=HowManyBits(direction*(right-left));
		break;
    case 202: /* bit */
		if(!strcmp(yylval.tokstr, "'0'"))
			strcpy(bitstring, "0");
		else
			strcpy(bitstring, "1");
		left=0;
		right=1;
		nrofbits=1;
		break;
    case 209: /* std_ulogic */
    case 210: /* std_logic */
		if(!strcmp(yylval.tokstr, "'0'"))
			strcpy(bitstring, "0");
		else
			if(!strcmp(yylval.tokstr, "'1'"))
				strcpy(bitstring, "1");
			else
				strcpy(bitstring, "0");
			left=0;
			right=1;
			nrofbits=1;
			break;
    case 203: /* bit_vector */
    case 211: /* std_ulogic_vector */
    case 212: /* std_logic_vector */
    case 213: /* signed */
    case 214: /* unsigned */
			strcpy(bitstring, yylval.tokstr);
			if(bitstring[0]=='x' && bitstring[1]=='"')
				ConstantBitString(yylval.toknum, bitstring);				
			/*Sprawdzic jakie kodowanie std_logica*/
			if(typeofconstant==211 || typeofconstant==212)
				ChangeLHTo01(bitstring);
			left=( left==-1 ? 0 : left);
			if(right==-1)
			{
				right=255;
				if((nrofbits=strlen(bitstring))>2)
					right=nrofbits-3;
			}
			nrofbits=abs(right-left)+1;
			break;
    case 204: /* boolean */
		if(!stricmp(yylval.tokstr, "false"))
			strcpy(bitstring, "0");
		else
			strcpy(bitstring, "1");
		left=0;
		right=1;
		nrofbits=1;
		break;
    case 205: /* character */
		strcpy(bitstring, yylval.tokstr);
		left=0;
		right=255;
		nrofbits=8;
		break;
	case 206: /* string */
		strcpy(bitstring, yylval.tokstr);
		left=( left==-1 ? 0 : left);
		if(right==-1)
		{
			right=255;
			if((nrofbits=strlen(bitstring))>2)
				right=nrofbits-3;
		}
		nrofbits=abs(right-left)+1;
		break;
    default:
		strcpy(bitstring, yylval.tokstr);
		SemGlobal->TypeCollected=2;
		if(FindTypeInformation(typeofconstant, yylval.tokstr, yylval.tokstr)==-1)
		{
			printf("Unsupported (yet?) constant declaration!\n");
			if(yylval.line<SemGlobal->lineno)
				yylval.line=SemGlobal->lineno;
			printf("Error appeared in line %d!\n", yylval.line);
			exit(-1);
		}
		else
		{
			int r1, r2, r3, r4;
			sscanf(yylval.tokstr, "%d\t%d\t%d\t%d\t%d\t%d", &r1, &r2, &r3, &r4, &left, &right);
			if(r4==TYPELV || r4 ==SUBTYPELV)
			{//kiedy typ stalej jest obliczany w trakcie generacji kodu
				nrofbits=0;
				right=0;
				left=0;
			}
			else
				nrofbits=right>left ?right-left+1: left-right+1;
		}
		SemGlobal->TypeCollected=0;
	}
	fgets(bufor, BUFSIZE, SemGlobal->srcf);
	value=getvalue(bufor, NULL);
	yylval.toknum=value;
	getname(bufor, yylval.tokstr);
	OldBufor2=strdup(bufor);
	if(value==SEMICV)
	{
		while(stack_pop(&ConstList, &constnumber, constname)!=EMPTY_STACK)
		{
			CheckNewName(constname, &constnumber, CONSTANTV);
			fprintf(semantics,"%d\t%s\t%d\t5\t1\t%d\t%d\t%d\t%s\n",
				constnumber, constname, typeofconstant,
				nrofbits, left, right, bitstring);
		}
		free(OldBufor2);
		free(OldBufor);
		return 0;
	}
	else
	{
		/*sprawdzenie czy stala nie jest np. tablica bitvectorow */
		//FindTypeInformation(typeofconstant, yylval.tokstr, yylval.tokstr);
		/* Koniec sprawdzenia*/
		sprintf(TmpNotComputablePath, "%s.tnc", SemGlobal->actpath);
		if(SemGlobal->hasNonComputables)
			tmpncmpfile=fopen(TmpNotComputablePath, "at");
		else
			tmpncmpfile=fopen(TmpNotComputablePath, "wt");
		if(!tmpncmpfile)
		{
			free(OldBufor2);
			free(OldBufor);
			return -1;
		}
		while(stack_pop(&ConstList, &constnumber, constname)!=EMPTY_STACK)
		{
			CheckNewName(constname, &constnumber, CONSTANTV);
			fprintf(tmpncmpfile,"%d\t%s\t%d\t5\t1\t%d\t%d\t%d\t0\n",
				constnumber, constname, typeofconstant,
				nrofbits, left, right);
			stack_push(&ConstList2, constnumber, constname);
		}
		fclose(tmpncmpfile);
	}
	sprintf(NotComputablePath, "%s.ncp", SemGlobal->actpath);
	if(SemGlobal->hasNonComputables)
		ncmpfile=fopen(NotComputablePath, "at");
	else
		ncmpfile=fopen(NotComputablePath, "wt");
	SemGlobal->hasNonComputables=1;
	if(!ncmpfile)
	{
		free(OldBufor2);
		free(OldBufor);
		return -1;
	}
	while(stack_pop(&ConstList2, &constnumber, constname)!=EMPTY_STACK)
	{
		fprintf(ncmpfile, "; %d\n", constnumber);
		fprintf(ncmpfile, "%s", OldBufor);
		fprintf(ncmpfile, "%s", OldBufor2);
		if(value!=SEMICV)
		{
			while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
			{
				value=getvalue(bufor, NULL);
				fprintf(ncmpfile, "%s", bufor);
				if(value==SEMICV)
					break;
			}
		}
		else
			fprintf(ncmpfile, "%s", bufor);
	}
	fclose(ncmpfile);
	free(OldBufor);
	free(OldBufor2);
	return 0;
}
