#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <search.h>
#include "vhdl_tab.h"
#include "tstst.h"

#ifndef tok_YET_UNKNOWN
#define tok_YET_UNKNOWN -1
#endif

char *GetTokStr(char *);
void GetTokStrPB(char *, char *);


static int tokentable[]=
{
    tok_YET_UNKNOWN,	/* there are some reasons ... */
	/*
	 * Keyword tokens
	 */
	tok_ABS, 		/* abs    */
	tok_ACCESS,		/* access */
    tok_AFTER,		/* after  */
    tok_ALIAS,		/* alias  */
    tok_ALL,		/* all    */
    tok_AND,		/* and    */
    tok_ARCHITECTURE,	/* architecture */
    tok_ARRAY,		/* array */
    tok_ASSERT,		/* assert */
    tok_ATTRIBUTE,	/* attribute */

    tok_BEGIN,		/* begin */
    tok_BLOCK,		/* block */
    tok_BODY,		/* body  */
    tok_BUFFER,		/* buffer */
    tok_BUS,		/* bus */

    tok_CASE,		/* case */
    tok_COMPONENT,	/* component */
    tok_CONFIGURATION,	/* configuration */
    tok_CONSTANT,	/* constant */

    tok_DISCONNECT,	/* disconnect */
    tok_DOWNTO,		/* downto */

    tok_ELSE,		/* else */
    tok_ELSIF,		/* elsif */
    tok_END,		/* end */
    tok_ENTITY,		/* entity */
    tok_EXIT,		/* exit */

    tok_FILE,		/* file */
    tok_FOR,		/* for */
    tok_FUNCTION,	/* function */

    tok_GENERATE,	/* generate */
    tok_GENERIC,	/* generic */
    tok_GROUP,		/* group */
    tok_GUARDED,	/* guarded */

    tok_IF,		/* if */
    tok_IMPURE,		/* impure */
    tok_IN,		/* in */
    tok_INERTIAL,	/* inertial */
    tok_INOUT,		/* inout */
    tok_IS,		/* is */

    tok_LABEL,		/* label */
    tok_LIBRARY,	/* library */
    tok_LINKAGE,	/* linkage */
    tok_LITERAL,        /* literal */
    tok_LOOP,		/* loop */

    tok_MAP,		/* map */
    tok_MOD,		/* mod */

    tok_NAND,		/* nand */
    tok_NEW,		/* new */
    tok_NEXT,		/* next */
    tok_NOR,		/* nor */
    tok_NOT,		/* not */
    tok_NULL,		/* null */

    tok_OF,		/* of */
    tok_ON,		/* on */
    tok_OPEN,		/* open */
    tok_OR,		/* or */
    tok_OTHERS,		/* others */
    tok_OUT,		/* out */

    tok_PACKAGE,	/* package */
    tok_PORT,		/* port */
    tok_POSTPONED,	/* postponed */
    tok_PROCEDURE,	/* procedure */
    tok_PROCESS,	/* process */
    tok_PURE,		/* pure */

    tok_RANGE,		/* range */
    tok_RECORD,		/* record */
    tok_REGISTER,	/* register */
    tok_REJECT,		/* reject */
    tok_REM,		/* rem */
    tok_REPORT,		/* report */
    tok_RETURN,		/* return */
    tok_ROL,		/* rol */
    tok_ROR,		/* ror */

    tok_SELECT,		/* select */
    tok_SEVERITY,	/* severity */
    tok_SIGNAL,		/* signal */
    tok_SHARED,		/* shared */
    tok_SLA,		/* sla */
    tok_SLL,		/* sll */
    tok_SRA,		/* sra */
    tok_SRL,		/* srl */
    tok_SUBTYPE,	/* subtype */

    tok_THEN,		/* then */
    tok_TO,		/* to */
    tok_TRANSPORT,	/* transport */
    tok_TYPE,		/* type */

    tok_UNAFFECTED,	/* unaffected */
    tok_UNITS,		/* units */
    tok_UNTIL,		/* until */
    tok_USE,		/* use */

    tok_VARIABLE,	/* variable */

    tok_WAIT,		/* wait */
    tok_WHEN,		/* when */
    tok_WHILE,		/* while */
    tok_WITH,		/* with */

    tok_XNOR,		/* xnor */
    tok_XOR,		/* xor */

    tok_YET_UNKNOWN,	/* reserved (for future use or something...) */
    tok_YET_UNKNOWN,	/* reserved (for future use or something...) */
    tok_YET_UNKNOWN,	/* reserved (for future use or something...) */

	/*
	 * Special character tokens
	 */
    tok_YET_UNKNOWN,	/* " */ 
    tok_YET_UNKNOWN,	/* # */
    tok_Ampersand,	/* & */
    tok_Apostrophe, 	/* ' */
    tok_LeftParen,	/* ( */
    tok_RightParen,	/* ) */
    tok_Star,		/* * */
    tok_Plus,		/* + */
    tok_Comma,		/* , */
    tok_Minus,		/* - */
    tok_Dot,		/* . */
    tok_Slash,		/* / */
    tok_Colon,		/* : */
    tok_Semicolon,	/* ; */
    tok_LTSym,		/* < */
    tok_EQSym,		/* = */
    tok_GTSym,		/* > */

    tok_YET_UNKNOWN,	/* _ */

    tok_Bar,		/* | */
    tok_Bar,		/* ! */

    tok_YET_UNKNOWN,	/* $ */
    tok_YET_UNKNOWN,	/* % */
    tok_YET_UNKNOWN,	/* ? */
    tok_YET_UNKNOWN,	/* @ */
    tok_YET_UNKNOWN,	/* [ */
    tok_YET_UNKNOWN,	/* \ */
    tok_YET_UNKNOWN,	/* ] */
    tok_YET_UNKNOWN,	/* ^ */
    tok_YET_UNKNOWN,	/* ` */
    tok_YET_UNKNOWN,	/* { */
    tok_YET_UNKNOWN,	/* } */
    tok_YET_UNKNOWN,	/* ~ */
    tok_YET_UNKNOWN,	/* inverted ! */
    tok_YET_UNKNOWN,	/* cent sign */
    tok_YET_UNKNOWN,	/* pound sign */
    tok_YET_UNKNOWN,	/* currency sign */
    tok_YET_UNKNOWN,	/* yen sign */
    tok_YET_UNKNOWN,	/* broken bar */
    tok_YET_UNKNOWN,	/* paragraph sign */
    tok_YET_UNKNOWN,	/* diaeresis */
    tok_YET_UNKNOWN,	/* copyright sign */
    tok_YET_UNKNOWN,	/* feminine ordinal indicator */
    tok_YET_UNKNOWN,	/* left angle quotation mark */
    tok_YET_UNKNOWN,	/* not sign */
    tok_YET_UNKNOWN,	/* soft hyphen */
    tok_YET_UNKNOWN,	/* registered trade mark sign */
    tok_YET_UNKNOWN,	/* macron */
    tok_YET_UNKNOWN,	/* degree sign */
    tok_YET_UNKNOWN,	/* plus-minus sign */
    tok_YET_UNKNOWN,	/* superscript two */
    tok_YET_UNKNOWN,	/* superscript three */
    tok_YET_UNKNOWN,	/* acute accent */
    tok_YET_UNKNOWN,	/* micro sign */
    tok_YET_UNKNOWN,	/* pilcrow sign */
    tok_YET_UNKNOWN,	/* middle dot */
    tok_YET_UNKNOWN,	/* cedilla */
    tok_YET_UNKNOWN,	/* superscript one */
    tok_YET_UNKNOWN,	/* masculine ordinal indicator */
    tok_YET_UNKNOWN,	/* right angle quotation mark */
    tok_YET_UNKNOWN,	/* vulgar fraction one quarter */
    tok_YET_UNKNOWN,	/* vulgar fraction one half */
    tok_YET_UNKNOWN,	/* vulgar fraction three quarters */
    tok_YET_UNKNOWN,	/* inverted question mark */
    tok_YET_UNKNOWN,	/* multiplication sign */
    tok_YET_UNKNOWN,	/* division sign */

	/*
	* Special double character tokens
	*/
    tok_Arrow,		/* => */
    tok_DoubleStar,	/* ** */
    tok_VarAsgn,	/* := */
    tok_NESym,		/* /= */
    tok_GESym,		/* >= */
    tok_LESym,		/* <= */
    tok_Box		/* <> */
};


static char *tokenstrtable[]=
{
    "",	                /* there are some reasons ... */
	/*
	 * Keyword tokens
	 */
	 "abs", "access", "after", "alias", "all",
	 "and", "architecture", "array",
	 "assert", "attribute",
	 "begin", "block", "body", "buffer", "bus", 
	 "case", "component", "configuration", "constant",
	 "disconnect", "downto",
	 "else", "elsif", "end", "entity", "exit",
	 "file", "for", "function",
	 "generate", "generic", "group", "guarded", 
	 "if", "impure", "in", "inertial", "inout", "is",
	 "label", "library",	"linkage", "literal", "loop",
	 "map", "mod",
	 "nand", "new", "next", "nor", "not", "null",
	 "of", "on", "open", "or", "others", "out",
	 "package", "port", "postponed", "procedure", "process", "pure",
	 "range", "record", "register", "reject", "rem", "report", "return", "rol",
	 "ror",
	 "select", "severity", "signal", "shared", "sla", "sll", "sra", "srl",
	 "subtype",
	 "then", "to", "transport", "type",
	 "unaffected", "units", "until", "use",
	 "variable",
	 "wait", "when", "while", "with",
	 "xnor", "xor",
	 "",	/* reserved (for future use or something...) */
	 "",	/* reserved (for future use or something...) */
	 "",	/* reserved (for future use or something...) */
	 
	/*
	 * Special character tokens
	 */
	 "\"", "#", "&", "'", "(", ")", "*", "+", ",", "-", ".", "/",   ":", ";",
	 "<",  "=", ">", "_", "|", "!", "$", "%%", "?", "@", "[", "\\", "]",	"^",
	 "`",  "{", "}", "~",
	 "",	/* inverted ! */
	 "",	/* cent sign */
	 "",	/* pound sign */
	 "",	/* currency sign */
	 "",	/* yen sign */
	 "",	/* broken bar */
	 "",	/* paragraph sign */
	 "",	/* diaeresis */
	 "",	/* copyright sign */
	 "",	/* feminine ordinal indicator */
	 "",	/* left angle quotation mark */
	 "",	/* not sign */
	 "",	/* soft hyphen */
	 "",	/* registered trade mark sign */
	 "",	/* macron */
	 "",	/* degree sign */
	 "",	/* plus-minus sign */
	 "",	/* superscript two */
	 "",	/* superscript three */
	 "",	/* acute accent */
	 "",	/* micro sign */
	 "",	/* pilcrow sign */
	 "",	/* middle dot */
	 "",	/* cedilla */
	 "",	/* superscript one */
	 "",	/* masculine ordinal indicator */
	 "",	/* right angle quotation mark */
	 "",	/* vulgar fraction one quarter */
	 "",	/* vulgar fraction one half */
	 "",	/* vulgar fraction three quarters */
	 "",	/* inverted question mark */
	 "",	/* multiplication sign */
	 "",	/* division sign */
	 
			/*
			* Special double character tokens
			*/
			"=>", "**",	":=", "/=", ">=", "<=", "<>"
};



extern YYSTYPE yylval;
extern struct PBGLOBAL *SemGlobal;

int yylex(void)
{
	static char line[256];
	char  wrd1[32];
	
	if( !fgets( line, 255, SemGlobal->srcf))
		return -1;
	if( line[0]=='#')
    {
		sscanf( line,"%s%d", wrd1, &(yylval.line));         /* yylval.line */
		return yylex();
    }
	GetTokStrPB(line, yylval.tokstr);							/* yylval.tokstr */
	sscanf(line,"%d%s", &(yylval.toknum), wrd1);          /* yylval.toknum */
	if( yylval.toknum <= 173)
		return (yylval.semval=tokentable[yylval.toknum]);
	if( !stricmp(wrd1,"identifier:")          ||
		!stricmp(wrd1,"extended_identifier:") )
		return (yylval.semval=tok_Identifier);
	if( !stricmp(wrd1,"character_constant:"))
		return (yylval.semval=tok_CharacterLit);
	if( !stricmp(wrd1,"string_constant:"))
		return (yylval.semval=tok_StringLit);
	if( !stricmp(wrd1,"bit_string_constant_bin:") ||
		!stricmp(wrd1,"bit_string_constant_oct:") ||
		!stricmp(wrd1,"bit_string_constant_hex:") )
		return (yylval.semval=tok_BitStringLit);
	if( !stricmp(wrd1,"integer_constant:")              || 
		!stricmp(wrd1,"integer_based_constant:")        ||
		!stricmp(wrd1,"floating_point_constant:")       ||
		!stricmp(wrd1,"floating_point_based_constant:") )
		return (yylval.semval=tok_AbstractLit);
	return -1;
}//yylex

int IdString2(int lexem, char *idstring);

int KeyWordString( int lexem, char *lexemstring)
{
	if( (lexem >=0) && (lexem <= 97))
    {
		sprintf(lexemstring,"%s",tokenstrtable[lexem]);
		return 0;
    }
	if( (lexem >=101) && (lexem <= 165))
    {
		sprintf(lexemstring,"%s",tokenstrtable[lexem]);
		return 0;
    }
	if( (lexem >=166) && (lexem <= 172))
    {
		sprintf(lexemstring,"%s",tokenstrtable[lexem]);
		return 0;
    }
	if( lexem >=201)
    {
		//char idstr[64];
		IdString2(lexem,lexemstring);
		return 0;
    }
	return 1;
}//KeyWordString


int IdString2(int lexem, char *idstring)
{
	int number;
	char line[1024], probe[1024];
	FILE *varfile;
	if(lexem<200)
		return KeyWordString(lexem, idstring);
	varfile=fopen("alu.var","rt");
	while( fgets(line,1023,varfile)!=NULL)
    {
		sscanf(line,"%d\t%s",&number, probe);
		if(lexem==number+200)
		{
			sprintf(idstring,"%s",probe);
			fclose(varfile);
			return 0;
		}
    }
	fclose(varfile);
	return 1;
}//IdString

int Tokenize(int lexem)
{
	if( lexem <= 173)
		return tokentable[lexem];
	else
		return tok_Identifier;
}//Tokenize


int LexemStringDescr( int lexem, char *lexemstring)
{
	if((lexem>=0) && (lexem<=97))
    {
		sprintf(lexemstring,"%d\tkeyword:\t%s",lexem,tokenstrtable[lexem]);
		return Tokenize(lexem);
    }
	if((lexem>=101) && (lexem<=165))
    {
		sprintf(lexemstring,"%d\tspecial_character:\t%s",lexem,tokenstrtable[lexem]);
		return Tokenize(lexem);
    }
	if((lexem>=166) && (lexem<=172))
    {
		sprintf(lexemstring,"%d\tdouble_char_symbol:\t%s",lexem,tokenstrtable[lexem]);
		return Tokenize(lexem);
    }
	if(lexem>=201)
    {
		char idstr[64];
		IdString2(lexem,idstr);
		sprintf(lexemstring,"%d\tidentifier:\t%s",lexem,idstr);
		return Tokenize(lexem);
    }
	return -1;
}//LexemStringDescr

char *GetTokStr(char *line)
{
	char *tokstr=strchr(line,':');
	int toklength=strlen(tokstr);
	tokstr=(char *)malloc(toklength*sizeof(char));
	strncpy(tokstr, strchr(line,':')+2, toklength-3);
	tokstr[toklength-3]='\0';
	return tokstr;
}//GetTokStr


void GetTokStrPB(char *line, char *instr)
{
	char *tokstr=strchr(line,':');
	int toklength=strlen(tokstr);
	tokstr=(char *)malloc(toklength*sizeof(char));
	strncpy( tokstr, strchr(line,':')+2, toklength-3);
	tokstr[toklength-3]='\0';
	strcpy(instr, tokstr);
	free(tokstr);
}//GetTokStrPB