/* Tworzenie drzewa katalogow dla analizatora semantycznego. Piotr Blaszynski*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <stdlib.h>
#include "tstst.h"
#include "bodypars.h"
#include "semerr.h"
#include "Konwert.h"
#include "Vhdl_tab.h"
#include "semval.h"
#include "def2id.h"

#ifdef MSVC_VER
#include <dir.h>
#else
#include <direct.h>
#endif


int proparse(char *, int);
int funcparse(char *, int);
int FuncPrototypeParse(char *, int);
int ProcedurePrototypeParse(char *, int);
int BlockParse(char *, int, char *);
int ProcessParse(char *, int, char *);
int NoNamedProcessParse(char *, int);
int GenerateParse(char *, int, char *, char *);
int NoNamedGenerateParse(char *, int, char *);
int entityparse(char *, int);



struct PBGLOBAL *SemGlobal;
YYSTYPE yylval;
extern DECL_STATUS decl_status;

char *mstrdup (const char * string)
{
        char *memory;

        if (!string)
                return(NULL);

        if (memory = malloc(strlen(string) + 1))
                return(strcpy(memory,string));

        return(NULL);
}


void semerror(int errno, char *addmesg)
{

	char line[16384], *msg, err[16384];
	FILE *msgfile;
	
	msgfile=fopen("..\\errmsg.txt", "rt");
	if(!msgfile)
		return;
	while(fgets(line,16383,msgfile))
	{
		int code;
		if(!sscanf(line,"%d",&code))
			continue;
		if( code!=errno)
			continue;
		for( msg=line; !isalpha(*msg) && *msg; msg++)
			;
		break;
	}
	fclose(msgfile);
	strcpy(err, msg);
	err[strlen(err)-1]=' ';
	if(addmesg)
		strcat(err, addmesg);

	printf("%d %s\nfile: <%s> line: <%d>\n",
		errno, err, SemGlobal->actfile, SemGlobal->lineno);
	fprintf(SemGlobal->errfile, "%d %s\nfile: <%s> line: <%d>\n",
		errno, err, SemGlobal->actfile, SemGlobal->lineno);
	if(!SemGlobal->errcnt)
	{
		FILE *errfile;
		errfile=fopen("lineno.err", "wt");
		if(errfile)
			fprintf(errfile, "%d", SemGlobal->lineno);
		fclose(errfile);
	}
	SemGlobal->errcnt++;
	return;
}//semerror

void NotImplemented(char *construct)
{
	printf("Construction not implemented: in file: %s, line: %d  %s\n",
		SemGlobal->actfile,
		SemGlobal->lineno,
		construct);
	fprintf(SemGlobal->errfile, "Construction not implemented: in file: %s, line: %d  %s\n",
		SemGlobal->actfile,
		SemGlobal->lineno,
		construct);
}

void semwarn(char *warn)
{
	printf("Semantic analyzis warning: in file: %s, line: %d  %s\n",
		SemGlobal->actfile,
		SemGlobal->lineno,
		warn);
	fprintf(SemGlobal->errfile, "Semantic analysis warning: in file: %s, line: %d  %s\n",
		SemGlobal->actfile,
		SemGlobal->lineno,
		warn);
	return;
}//semwarn



int getvalue(char *bufor, FILE *dest)//bufor w zadnym wypadku nie moze byc const
{
	char c, svalue[16];
	unsigned i;
	c=bufor[0];
	if(isdigit(c))
	{
		i=0;
		while(isdigit(c))
		{
			svalue[i]=c;
			i++;
			c=bufor[i];
		}
		svalue[i]=0;
		return atoi(svalue);
	}
	if(c=='#')
	{
		if (!strncmp(bufor+1, "line", 4))
		{
			i=6;
			c=bufor[i];
			while(isdigit(c))
			{
				svalue[i-6]=c;
				i++;
				c=bufor[i];
			}
			svalue[i-6]=0;
			SemGlobal->lineno=atoi(svalue);
			if(dest!=NULL)
				fprintf(dest, "%s", bufor);
		}
		//odczyt kolejnej linii i modyfikacja zawartosci bufora
		if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
			return 0;
		return getvalue(bufor, dest);
		
	}
	return 0;
}//getvalue


int getValueForResolved(char *bufor, FILE *dest)//bufor w zadnym wypadku nie moze byc const
{
	char c, svalue[16];
	unsigned i;
	c=bufor[0];
	if(isdigit(c))
	{
		i=0;
		while(isdigit(c))
		{
			svalue[i]=c;
			i++;
			c=bufor[i];
		}
		svalue[i]=0;
		return atoi(svalue);
	}
	if(c=='#')
	{
		if (!strncmp(bufor+1, "line", 4))
		{
			i=6;
			c=bufor[i];
			while(isdigit(c))
			{
				svalue[i-6]=c;
				i++;
				c=bufor[i];
			}
			svalue[i-6]=0;
			SemGlobal->lineno=atoi(svalue);
			if(dest!=NULL)
				fprintf(dest, "%s", bufor);
		}
		else
		{
			int i=1, isResFun=0;
			char MethodName[NAMESIZE];
			while(bufor[i])
			{
				if(isalpha(bufor[i]))
					break;
				i++;
			}
#define SYNLEN 26
#define PRALEN 24
#define ALDLEN 23
			if(!strnicmp(bufor+i, "synopsys resolution_method", SYNLEN))
			{
				strcpy(MethodName, bufor+i+SYNLEN+1);
				isResFun=1;
			}
			if(!isResFun && !strnicmp(bufor+i, "aldec resolution_method", ALDLEN))
			{
				strcpy(MethodName, bufor+i+ALDLEN+1);
				isResFun=1;
			}
			if(!isResFun && !strnicmp(bufor+i, "pragma resolution_method", PRALEN))
			{
				strcpy(MethodName, bufor+i+PRALEN+1);
				isResFun=1;
			}
			if(isResFun)
			{
				if(!strnicmp(MethodName, "wired_and", 9))
				{
					SemGlobal->LastResolvedType=WIREDAND;
				}
				if(!strnicmp(MethodName, "wired_or", 8))
				{
					SemGlobal->LastResolvedType=WIREDOR;
				}
				if(!strnicmp(MethodName, "three_state", 11))
				{
					SemGlobal->LastResolvedType=THREESTATE;
				}
			}
		}
		//odczyt kolejnej linii i modyfikacja zawartosci bufora
		if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
			return 0;
		return getValueForResolved(bufor, dest);
	}
	return 0;
}//getvalue



void getname(char *bufor, char *name)
{
	char c;
	int i, j;
	c=bufor[0];
	i=0;
	while(c!=':')
	{
		i++;
		c=bufor[i];
	}
	i++;
	c=bufor[i];
	while((c=='\t' || c==' '))
	{
		i++;
		c=bufor[i];
	}
	name[0]=c;
	j=0;
	while(c!='\n')
	{
		j++;
		i++;
		c=bufor[i];
		name[j]=c;
	}
	name[j]=0;
}//getname



void copytoend(FILE *dest)
{
	char bufor[BUFSIZE];
	int value, openb=0;
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		fprintf(dest, "%s", bufor);
		if(value==LEFTBRV)
			openb++;
		if(value==RIGHTBRV)
			openb--;
		if(value==SEMICV && openb==0)
			break;
	}
}//copytoend

void CopyToSemicolon(FILE *dest)
{
	char bufor[BUFSIZE];
	int value;
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		fprintf(dest, "%s", bufor);
		if(value==SEMICV)
			break;
	}
}//CopyToSemicolon


int ParseNamedComponentMap(FILE *dest, int ie, int im, int isCompon)
{
	//ie - index jednostki mapujacej (toplevel), im - index komponentu
	//mv - wartosc do ewentualnej modyfikacji
	int i, j, mv, value, WasModified, ParenCnt=0;
	char bufor[BUFSIZE];
	for(i=0 ; i<SemGlobal->LexemsInListCnt ; i++)
	{
		WasModified=0;
		mv=SemGlobal->ListOfYY[i].toknum;
		if(mv>200)
		{
			if(im!=-1)
			{
				if(isCompon)
				{
					for(j=0 ; j<SemGlobal->ChangedCom[im].modifierscnt ; j++)
					{
						if(mv==SemGlobal->ChangedCom[im].modifiers[j].origlexno)//wyszukiwanie pasujacego modyfikatora
						{
							fprintf(dest, "%d\tidentifier:\t%s\n", SemGlobal->ChangedCom[im].modifiers[j].modlexno, SemGlobal->ChangedCom[im].modifiers[j].ModifiedName);
							WasModified=1;
							break;
						}
					}
				}
				else
				{
					for(j=0 ; j<SemGlobal->entities[im].modifierscnt ; j++)
					{
						if(mv==SemGlobal->entities[im].modifiers[j].origlexno)//wyszukiwanie pasujacego modyfikatora
						{
							fprintf(dest, "%d\tidentifier:\t%s\n", SemGlobal->entities[im].modifiers[j].modlexno, SemGlobal->entities[im].modifiers[j].ModifiedName);
							WasModified=1;
							break;
						}
					}
				}
			}
		}
		if(!WasModified)
			fprintf(dest, "%s", SemGlobal->ListOfYY[i].tokstr);
	}
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		if(value==LEFTBRV)
		{	
			ParenCnt++;
			fprintf(dest, "%s", bufor);
			continue;
		}
		if(value==RIGHTBRV)
		{
			ParenCnt--;
			fprintf(dest, "%s", bufor);
			continue;
		}
		if(value==COMMAV)//nastepne mapowanie
		{
			fprintf(dest, "%s", bufor);
			break;
		}
		if(value==SEMICV)//koniec mapowania
		{
			fprintf(dest, "%s", bufor);
			break;
		}
		mv=value;
		WasModified=0;
		for(j=0 ; j<SemGlobal->entities[ie].modifierscnt ; j++)
		{
			if(mv==SemGlobal->entities[ie].modifiers[j].origlexno)
			{
				fprintf(dest, "%d\tidentifier:\t%s\n", SemGlobal->entities[ie].modifiers[j].modlexno, SemGlobal->entities[ie].modifiers[j].ModifiedName);
				WasModified=1;
				break;
			}
		}
		for(j=0 ; j<SemGlobal->ModifiersCnt.real ; j++)
		{
			if(SemGlobal->Modifiers[j].isInArchitecture &&
				mv==SemGlobal->Modifiers[j].origlexno && 
				SemGlobal->Modifiers[j].level==0)
			{
				if(SemGlobal->Modifiers[j].ComponName && !stricmp(SemGlobal->Modifiers[j].ComponName, SemGlobal->actEntityContext))
				{
					fprintf(dest, "%d\tidentifier:\t%s\n", SemGlobal->Modifiers[j].modlexno, SemGlobal->Modifiers[j].ModifiedName);
					WasModified=1;
					break;
				}
			}
		}
		if(!WasModified)
			fprintf(dest, "%s", bufor);
	}
	FreeListOfYY();
	return value;
}//ParseComponentMap


void ParseOrderComponentMap(FILE *dest, int ie, int isCompon)
{
	//ie - index jednostki mapujacej (toplevel), im - index komponentu
	//mv - wartosc do ewentualnej modyfikacji
	int i, j, mv, WasModified, ParenCnt=0;
	for(i=0 ; i<SemGlobal->LexemsInListCnt ; i++)
	{
		WasModified=0;
		mv=SemGlobal->ListOfYY[i].toknum;
		if(isCompon)
		{
			for(j=0 ; j<SemGlobal->ChangedCom[ie].modifierscnt ; j++)
			{
				if(mv==SemGlobal->ChangedCom[ie].modifiers[j].origlexno)
				{
					fprintf(dest, "%d\tidentifier:\t%s\n", SemGlobal->ChangedCom[ie].modifiers[j].modlexno, SemGlobal->ChangedCom[ie].modifiers[j].ModifiedName);
					WasModified=1;
					break;
				}
			}
		}
		else
		{
			for(j=0 ; j<SemGlobal->entities[ie].modifierscnt ; j++)
			{
				if(mv==SemGlobal->entities[ie].modifiers[j].origlexno)
				{
					fprintf(dest, "%d\tidentifier:\t%s\n", SemGlobal->entities[ie].modifiers[j].modlexno, SemGlobal->entities[ie].modifiers[j].ModifiedName);
					WasModified=1;
					break;
				}
			}
		}
		for(j=0 ; j<SemGlobal->ModifiersCnt.real ; j++)
		{
			if(SemGlobal->Modifiers[j].isInArchitecture &&
				mv==SemGlobal->Modifiers[j].origlexno && 
				SemGlobal->Modifiers[j].level==0)
			{
				fprintf(dest, "%d\tidentifier:\t%s\n", SemGlobal->Modifiers[j].modlexno, SemGlobal->Modifiers[j].ModifiedName);
				WasModified=1;
				break;
			}
		}

		if(!WasModified)
			fprintf(dest, "%s", SemGlobal->ListOfYY[i].tokstr);
	}
	FreeListOfYY();
}//ParseOrderComponentMap


void CopyPortMap(FILE *dest, char *mapped)
{
	char bufor[BUFSIZE]/*, tokstr[NAMESIZE]*/;
	int value, i, itop, ie, ic, ModifiedNames=0, MapIdx=-1, ParenCnt=0, isCompon=0;
	ie=SemGlobal->ileencji-1;//numer ostatniej jednostki
	for(i=0 ; i<SemGlobal->ileencji ; i++)
	{
		if(!stricmp(SemGlobal->actEntityContext, SemGlobal->entities[i].name))
		{
			itop=i;//jednostka mapujaca
			break;
		}
	}
	ic=SemGlobal->ChangedComcnt;//numer ostatniego componentu
	if(SemGlobal->ModifiersCnt.real)
	{
		if(SemGlobal->entities[ie].modifierscnt || SemGlobal->ChangedComcnt)
			ModifiedNames=1;
		for(i=0 ; i<=ie; i++)
		{
			if(!stricmp(SemGlobal->entities[i].name, mapped))//wyszukiwanie mapowanej jednostki
			{
				if(SemGlobal->entities[i].modifierscnt)
				{
					ModifiedNames=1;
					MapIdx=i;
					break;
				}
				else
				{
					MapIdx=i;
					break;
				}
			}
		}
		for(i=0 ; i<SemGlobal->ChangedComcnt ; i++)
		{
			if(!stricmp(SemGlobal->ChangedCom[i].name, mapped))//wyszukiwanie mapowanego componentu
			{
				MapIdx=i;
				isCompon=1;
				break;
			}
		}
	}
	//Zawsze zakladany jest przeklad ze sprawdzaniem modyfikacji nazw
	//MapIdx - indeks componentu, itop - indeks jednostki mapujacej (toplevel)
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		if(value==LEFTBRV)
			ParenCnt++;
		if(value==RIGHTBRV)
			ParenCnt--;
		yylval.toknum=value;
		yylval.semval=value;
		strcpy(yylval.tokstr, bufor);
		//fprintf(dest, "%s", bufor);
		if(value==IMPLICV)//mapowany component => toplevel mapujacy
		{
			NextYY();
			value=ParseNamedComponentMap(dest, itop, MapIdx, isCompon);
			ParenCnt--;
			if(value==SEMICV)//koniec mapowania
				break;
			continue;
		}
		if(value==COMMAV)//nastepne mapowanie
		{
			NextYY();
			ParseOrderComponentMap(dest, itop, isCompon);
			continue;
		}
		if(value==SEMICV)//koniec mapowania
		{
			NextYY();
			ParseOrderComponentMap(dest, itop, isCompon);
			break;
		}
		NextYY();
	}
	FreeListOfYY();
}//CopyPortMap





void ReadToSemiC(void)
{
	char bufor[BUFSIZE];
	int value;
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, NULL);
		if(value==SEMICV)
			break;
	}
}//ReadToSemiC


void ReadToIsOrSemiC(void)
{
	char bufor[BUFSIZE];
	int value;
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, NULL);
		if(value==ISV)
			break;
		if(value==SEMICV)
			break;
	}
}//ReadToIsOrSemiC


void WriteLastLexemId(char *tokstr)
{
	FILE *varfile;
	varfile=fopen("alu.var", "at");
	if(!varfile)
		return;
	fprintf(varfile, "%d\t%s\n", SemGlobal->LastLexemId, tokstr);
	SemGlobal->LastLexemId++;
	fclose(varfile);
}//WriteLastLexemId



int GenericParse(int What)
{
	char bufor[BUFSIZE], valbuf[NAMESIZE], namebuf[BUFSIZE];
	int value, NotEnd=1, BrCnt, defaultmode, TypeOf;
	defaultmode=SemGlobal->defaultmode;
	SemGlobal->defaultmode=3;/* default mode INOUT */
	fgets(bufor, BUFSIZE, SemGlobal->srcf);/* skip ( */
	value=getvalue(bufor, NULL);
	while(NotEnd)
	{
		fgets(bufor, BUFSIZE, SemGlobal->srcf);/* get LexemId */
		value=getvalue(bufor, NULL);
		getname(bufor, namebuf);
		//tmpidnum=value;
		if(What==ENTITYV)
			NewEntityPort(&value, namebuf);
		if(What==COMPONV)
			CheckNamesInEntity(&value, namebuf);
		if(SemGlobal->ListOfGenerics==NULL)
			SemGlobal->ListOfGenerics=(GENERIC*)malloc(sizeof(GENERIC));
		else
			SemGlobal->ListOfGenerics=(GENERIC*)realloc((GENERIC*)SemGlobal->ListOfGenerics, (SemGlobal->GenericCnt+1)*sizeof(GENERIC));
		SemGlobal->ListOfGenerics[SemGlobal->GenericCnt].LexemId=value;
		SemGlobal->ListOfGenerics[SemGlobal->GenericCnt].name=strdup(namebuf);
		BrCnt=0;
		while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
		{
			value=getvalue(bufor, NULL);
			if(value==COLONV)
			{
				int value;
				if(fgets(bufor, BUFSIZE, SemGlobal->srcf))
				{
					value=getvalue(bufor, NULL);
					if(value==INV || value==INOUTV || value==OUTV)
					{
						fgets(bufor, BUFSIZE, SemGlobal->srcf);
						value=getvalue(bufor, NULL);
					}
					SemGlobal->ListOfGenerics[SemGlobal->GenericCnt].TypeOf=value;
				}
			}
			if(value==LEFTBRV)
				BrCnt++;
			if(value==RIGHTBRV)
			{
				BrCnt--;
				if(BrCnt<0)
					break;
			}
			if(value==ASSIGNV || value==SEMICV)
				break;
		}
		if(value==SEMICV || value==RIGHTBRV)
		{
			SemGlobal->ListOfGenerics[SemGlobal->GenericCnt].value=0;
			SemGlobal->GenericCnt++;
			if(value==RIGHTBRV)
			{
				if(fgets(bufor, BUFSIZE, SemGlobal->srcf))
				{
					value=getvalue(bufor, NULL);
					if(value==SEMICV)
					{
						NotEnd=0;
						break;
					}
				}
			}
			continue;
		}
		fgets(bufor, BUFSIZE, SemGlobal->srcf);/* get value of generic */
		getname(bufor, valbuf);
		TypeOf=SemGlobal->ListOfGenerics[SemGlobal->GenericCnt].TypeOf;
		if(TypeOf==INTEGERLV || TypeOf==POSITIVELV || TypeOf==NATURALLV)
		{
			SemGlobal->ListOfGenerics[SemGlobal->GenericCnt].value=atoi(valbuf);
		}
		if(TypeOf==BOOLEANLV)
		{
			int len;
			len=strlen(valbuf);
			if(!stricmp(valbuf, "true"))
				SemGlobal->ListOfGenerics[SemGlobal->GenericCnt].value=1;
			if(!stricmp(valbuf, "false"))
				SemGlobal->ListOfGenerics[SemGlobal->GenericCnt].value=0;
		}
		if(TypeOf==BITLV || TypeOf==STDULOGICLV || TypeOf==STDLOGICLV)
		{
			int len;
			len=strlen(valbuf);
			valbuf[len-1]=0;
			SemGlobal->ListOfGenerics[SemGlobal->GenericCnt].value=atoi(valbuf+1);
		}
		if(TypeOf==CHARLV)
		{
			int len;
			len=strlen(valbuf);
			valbuf[len-1]=0;
			SemGlobal->ListOfGenerics[SemGlobal->GenericCnt].value=*(valbuf+1);
		}
		if(TypeOf==BITVECTORLV || TypeOf==SULVECTORLV || TypeOf==SLVECTORLV || TypeOf==STRINGLV)
		{
			SemGlobal->ListOfGenerics[SemGlobal->GenericCnt].bitvalue=strdup(valbuf);
		}
		SemGlobal->GenericCnt++;
		while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
		{
			value=getvalue(bufor, NULL);
			if(value==RIGHTBRV)
			{
				if(fgets(bufor, BUFSIZE, SemGlobal->srcf))
				{
					value=getvalue(bufor, NULL);
					if(value==SEMICV)
					{
						NotEnd=0;
						break;
					}
				}
			}
			if(value==SEMICV)
				break;
		}
	}
	SemGlobal->defaultmode=defaultmode;/*return to previous value*/
	return 0;
}//GenericParse



void NewComponentName(char *componname)
{
	if(SemGlobal->componcnt==0)
	{
		SemGlobal->componcnt=1;
		SemGlobal->components=(COMPONENT*) malloc (sizeof(COMPONENT));
		strcpy(SemGlobal->components[0].name, componname);
	}
	else
	{
		int i;
		for(i=0 ; i<SemGlobal->componcnt ; i++)
		{
			if(!stricmp(SemGlobal->components[i].name, componname))
				break;
		}
		if(i==SemGlobal->componcnt)
		{
			SemGlobal->componcnt++;
			SemGlobal->components =
				(COMPONENT*) realloc ((COMPONENT*)SemGlobal->components, sizeof(COMPONENT)*SemGlobal->componcnt);
			strcpy(SemGlobal->components[i].name, componname);
		}
	}
}//NewComponentName


int componparse(char *destpath, int level)
{
	char bufor[BUFSIZE], componname[NAMESIZE];
	char dpath[PATHSIZE], valuef[PATHSIZE], usepath[PATHSIZE];	
	FILE *dest, *destval;
	int value, IsIsRead, ComponValue;
	SemGlobal->ListOfGenerics=NULL;
	SemGlobal->GenericCnt=0;
	SemGlobal->isComponentPort=1;
	mkdir(destpath);
	//na poczatku komponentu znajduje sie jego identyfikator, ktory bedzie nazwa pliku
	fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt nazwy komponentu
	ComponValue=getvalue(bufor, NULL);
	getname(bufor, componname);
	sprintf(valuef, "%s\\%s.val", destpath, componname);
	if((dest=fopen(valuef, "wt")))
	{
		fprintf(dest, "# 0 Czy jest dostepny pakiet IEEE.std_logic_1164. 1-jest, 0-nie jest\n");
		fprintf(dest, "%d\n", SemGlobal->isIEEE1164avail);
		fprintf(dest, "# 1 Wartosc rodzica aktualnej jednostki.\n");
		fprintf(dest, "%s\n", SemGlobal->actroot);
		SemGlobal->varlevel++;
	}
	else
		return -1;
	fclose(dest);
	SemGlobal->indexno++;
	fprintf(SemGlobal->indexf, "%d 8 %s\t%s\\%s\n", SemGlobal->indexno, componname, destpath, componname);
	sprintf(dpath, "%s\\%s", destpath, componname);
	sprintf(usepath, "%s\\%s.val", destpath, componname);
	strcpy(SemGlobal->actpath, dpath);
	
	if(dest=fopen(usepath, "at"))
	{
		fprintf(dest, "# 5 Wartosci semantyczne\n");
		fclose(dest);
	}
	else
		return -1;
	
	if((dest=fopen(dpath, "wt"))==NULL)
		return -1;
	
	IsIsRead=0;
	//odczyt do slowa kluczowego end, dozwolone tylko: port i generic
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		if(value==ISV && IsIsRead==0)
		{	
			IsIsRead=1;
			continue;
		}
		if(value==GENERICV)
		{
			strcpy(SemGlobal->actcompo, componname);
			GenericParse(COMPONV);
			continue;
		}
		if(value==PORTV)
		{
			strcpy(SemGlobal->actcompo, componname);
			ParseEntityPortList(dpath, COMPONV);
			continue;
		}
		if(value==ENDV)
		{
			int i;
			char tokstr[BUFSIZE];
			fprintf(dest, "%s", bufor);
			/* Lets check generics */
			destval=fopen(valuef, "at");
			for(i=0 ; i<SemGlobal->GenericCnt ; i++)
			{
				int tmpmode=decl_status.mode;
				char BinStr[128], DecStr[128];
				int TypeOf=SemGlobal->ListOfGenerics[i].TypeOf;
				if(TypeOf==INTEGERLV || TypeOf==POSITIVELV || TypeOf==NATURALLV || TypeOf==BITLV || TypeOf==BITVECTORLV || TypeOf==STRINGLV || TypeOf==STDULOGICLV || TypeOf==STDLOGICLV || TypeOf==SULVECTORLV || TypeOf==SLVECTORLV)
				{
					int lType, lClass, lMode, lBits, lLeft, lRight;
					decl_status.mode=3;//inout
					strcpy(SemGlobal->tmpactpath, SemGlobal->actpath);
					FindTypeString(SemGlobal->ListOfGenerics[i].TypeOf, tokstr, tokstr);
					ReduceTo6Cols(tokstr);
					sscanf(tokstr, "%d\t%d\t%d\t%d\t%d\t%d", &lType, &lClass, &lMode, &lBits, &lLeft, &lRight);
					switch(TypeOf)
					{
					case INTEGERLV:
					case POSITIVELV:
					case NATURALLV: 
					case BITLV: 
					case STDULOGICLV: 
					case STDLOGICLV: 
						sprintf(DecStr, "%d", SemGlobal->ListOfGenerics[i].value);
						strcpy(BinStr, anytobin(DecStr, 10));
						break;
					case CHARLV: 
						sprintf(DecStr, "%d", SemGlobal->ListOfGenerics[i].value);
						strcpy(BinStr, anytobin(DecStr, 10));
						break;
					case SULVECTORLV:
					case SLVECTORLV:
					case BITVECTORLV:
					case STRINGLV:
						strcpy(BinStr, SemGlobal->ListOfGenerics[i].bitvalue);
						lBits=strlen(BinStr)-2;
						lLeft=0;
						lRight=lBits-1;
						if(TypeOf==STRINGLV)
						{
							lLeft=1;lRight=lBits;
						}
						break;
					}
					sprintf(tokstr, "%d\t16\t%d\t%d\t%d\t%d", lType, lMode, lBits, lLeft, lRight);
					fprintf(destval, "%d\t%s\t%s\t%s\n", SemGlobal->ListOfGenerics[i].LexemId, SemGlobal->ListOfGenerics[i].name, tokstr, BinStr);
					decl_status.mode=tmpmode;
				}
				else
				{
					;//printf("\nTypeOf:%d\n", TypeOf);
				}
			}
			fclose(destval);
			break;
		}
		if(value==ASSERTV || value==ATTRIBV)
		{
			ReadToSemiC();
			continue;
		}
		if(value)
			semerror(ERRCOMPFORBID, NULL);
	}
	fprintf(dest, "%s", bufor);
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		if(value==SEMICV)
		{
			fprintf(dest, "%s", bufor);
			break;
		}
		fprintf(dest, "%s", bufor);
	}
	fclose(dest);
	SemGlobal->varlevel--;
	strcpy(SemGlobal->LastFoundName, componname);
	SemGlobal->LastFoundValue=ComponValue;	
	if(SemGlobal->GenericCnt)
	{
		int i;
		for(i=0 ; i<SemGlobal->GenericCnt ; i++)
			free(SemGlobal->ListOfGenerics[i].name);
		free(SemGlobal->ListOfGenerics);
		SemGlobal->ListOfGenerics=NULL;
		SemGlobal->GenericCnt=0;
	}
	SemGlobal->isComponentPort=0;
	return 0;
}//componparse



void newentity(void)
{
	if(SemGlobal->ileencji==0)
	{
		SemGlobal->ileencji=1;
		SemGlobal->entities=(ENTITY*) malloc (sizeof(ENTITY));
	}
	else
	{
//		ENTITY *tmpent;
//		tmpent=malloc(SemGlobal->ileencji*sizeof(ENTITY));
		SemGlobal->ileencji++;
		SemGlobal->entities =
			(ENTITY*) realloc ((ENTITY*)(SemGlobal->entities), sizeof(ENTITY)*SemGlobal->ileencji);
	}
	SemGlobal->entities[SemGlobal->ileencji-1].isarch=0;
	SemGlobal->entities[SemGlobal->ileencji-1].isentity=0;
	SemGlobal->entities[SemGlobal->ileencji-1].isconfig=0;
	SemGlobal->entities[SemGlobal->ileencji-1].isTopLevel=1;
	//kazda jednostka z zalozenia bedzie typu TopLevel, o ile nie bedzie nastepowalo jej mapowanie
	SemGlobal->entities[SemGlobal->ileencji-1].portcnt=0;
	SemGlobal->entities[SemGlobal->ileencji-1].portids=NULL;
	SemGlobal->entities[SemGlobal->ileencji-1].isbuffers=NULL;
	SemGlobal->entities[SemGlobal->ileencji-1].HasResolvFile=0;
	SemGlobal->entities[SemGlobal->ileencji-1].modifierscnt=0;
	SemGlobal->entities[SemGlobal->ileencji-1].modifiers=NULL;
}//newentity



int UseParse(void)
{
	char bufor[BUFSIZE], tmppath[PATHSIZE];
	char usepath[BUFSIZE], part[NAMESIZE], *token;
	FILE *packfile, *usefile;
	int WorkPackNotPresent=0;
	int value, iswork=0, i, PartsCnt=0;
	usepath[0]=0;
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, NULL);
		if(value==SEMICV)
			break;
		else
		{
			if(value==DOTV)
			{
				sprintf(usepath,"%s.", usepath);
				PartsCnt++;
			}
			else
			{
				getname(bufor, part);
				if(usepath[0]==0 && !stricmp("work", part))
					iswork=1;
				sprintf(usepath,"%s%s", usepath, part);
			}
		}
	}
	if(stricmp(part, "all") && isalnum(*part))
	{
		if(iswork && PartsCnt==1)
		{
			strcat(usepath, ".all"); 
			strcpy(part, "all");
		}
		else
		{
			int l1, l2;
			l1=strlen(part);
			l2=strlen(usepath);
			strncpy(usepath+l2-l1, "all", 4); 
			strcpy(part, "all");
		}
	}
	if(!stricmp("ieee.std_logic_1164.all", usepath) || !stricmp("std_logic_1164.all", usepath) )
	{
		SemGlobal->isIEEE1164avail=1;
		return 0;
	}
	if(!stricmp("synopsys.attributes.all", usepath))
		return 0;
	if((packfile=fopen("packages.pck", "rt"))==NULL)
		WorkPackNotPresent=1;
	if(!WorkPackNotPresent)
	{
		while(fgets(bufor, BUFSIZE, packfile))
		{
			token=strtok(bufor, " \t");
			if(!stricmp(token, usepath))
			{
				if(!stricmp("ieee.std_logic_1164.all", usepath))
					SemGlobal->isIEEE1164avail=1;
				token=strtok(NULL, "\n");
				if(SemGlobal->pathcnt==0)
				{
					SemGlobal->pathcnt=1;
					SemGlobal->usepaths=(USEPATH*) malloc (sizeof(USEPATH));
				}
				else
				{
					SemGlobal->pathcnt++;
					SemGlobal->usepaths =
						(USEPATH*) realloc ((USEPATH*)SemGlobal->usepaths, sizeof(USEPATH)*SemGlobal->pathcnt);
				}
				strcpy(SemGlobal->usepaths[SemGlobal->pathcnt-1].path, token); 
				return 0;
			}
		}
	}

	if((packfile=fopen("..\\syspack.pck", "rt"))!=NULL)
	{
		while(fgets(bufor, BUFSIZE, packfile))
		{
			token=strtok(bufor, " \t");
			if(!stricmp(token, usepath))
			{
				if(!stricmp("ieee.std_logic_1164.all", usepath))
					SemGlobal->isIEEE1164avail=1;
				token=strtok(NULL, "\n");
				if(SemGlobal->pathcnt==0)
				{
					SemGlobal->pathcnt=1;
					SemGlobal->usepaths=(USEPATH*) malloc (sizeof(USEPATH));
				}
				else
				{
					SemGlobal->pathcnt++;
					SemGlobal->usepaths =
						(USEPATH*) realloc ((USEPATH*)SemGlobal->usepaths, sizeof(USEPATH)*SemGlobal->pathcnt);
				}
				strcpy(SemGlobal->usepaths[SemGlobal->pathcnt-1].path, token);
				strcpy(tmppath, token);
				i=strlen(tmppath);
				while(tmppath[i--]!='\\');
				tmppath[i+1]=0;
				strcat(tmppath, "\\clause.use");
				usefile=fopen(tmppath, "rt");
				if(usefile)
				{
					while(fgets(bufor, 2048, usefile))
					{
						DeleteLastChar(bufor);
						SemGlobal->pathcnt++;
						SemGlobal->usepaths =
							(USEPATH*) realloc ((USEPATH*)SemGlobal->usepaths, sizeof(USEPATH)*SemGlobal->pathcnt);
						strcpy(SemGlobal->usepaths[SemGlobal->pathcnt-1].path, bufor);
					}
					fclose(usefile);
				}
				return 0;
			}
		}
	}
	else
	{
		if(WorkPackNotPresent)
			return -1;
	}
	if(iswork)
	{
		if(SemGlobal->nfoundcnt==0)
		{
			SemGlobal->nfoundcnt=1;
			SemGlobal->usesnfound=(USENFOUND*) malloc (sizeof(USENFOUND));
		}
		else
		{
			SemGlobal->nfoundcnt++;
			SemGlobal->usesnfound =
				(USENFOUND*) realloc ((USENFOUND*)SemGlobal->usesnfound, sizeof(USENFOUND)*SemGlobal->nfoundcnt);
		}
		strcpy(SemGlobal->usesnfound[SemGlobal->nfoundcnt-1].useclause, usepath);
		SemGlobal->usesnfound[SemGlobal->nfoundcnt-1].lineno=SemGlobal->lineno;
		strcpy(SemGlobal->usesnfound[SemGlobal->nfoundcnt-1].filename,SemGlobal->actfile);
		SemGlobal->usesnfound[SemGlobal->nfoundcnt-1].issth=0;
	}
	else
		semerror(ERRNFPACKAGE, usepath);
	return 0;
}//UseParse



int UseParseInSubProgram(char *dpath)
{
	char bufor[BUFSIZE];
	char usepath[BUFSIZE], part[NAMESIZE], *token;
	int WorkPackNotPresent=0, PartsCnt=0;
	FILE *packfile, *destfile;
	int value, iswork=0;
	usepath[0]=0;
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, NULL);
		if(value==SEMICV)
			break;
		else
		{
			if(value==DOTV)
			{
				sprintf(usepath,"%s.", usepath);
				PartsCnt++;
			}
			else
			{
				getname(bufor, part);
				if(usepath[0]==0 && !stricmp("work", part))
					iswork=1;
				sprintf(usepath,"%s%s", usepath, part);
			}
		}
	}
	if((packfile=fopen("packages.pck", "rt"))==NULL)
		WorkPackNotPresent=1;
	if(iswork && PartsCnt==1)
	{
		strcat(usepath, ".all"); 
		strcpy(part, "all");
	}
	if(!WorkPackNotPresent)
	{
		
		while(fgets(bufor, BUFSIZE, packfile))
		{
			token=strtok(bufor, " \t");
			if(!stricmp(token, usepath))
			{
				if(!stricmp("ieee.std_logic_1164.all", usepath))
					SemGlobal->isIEEE1164avail=1;
				token=strtok(NULL, "\n");
				destfile=fopen(dpath, "at");
				fprintf(destfile, "# 3 Sciezki do pakietow uzywanych przez podprogram"); 
				fprintf(destfile, "%s\n", token);
				fclose(destfile);
				return 0;
			}
		}
	}
	if((packfile=fopen("..\\syslib\\packages.pck", "rt"))!=NULL)
	{
		while(fgets(bufor, BUFSIZE, packfile))
		{
			token=strtok(bufor, " \t");
			if(!stricmp(token, usepath))
			{
				if(!stricmp("ieee.std_logic_1164.all", usepath))
					SemGlobal->isIEEE1164avail=1;
				token=strtok(NULL, "\n");
				destfile=fopen(dpath, "at");
				fprintf(destfile, "# 3 Sciezki do pakietow uzywanych przez podprogram"); 
				fprintf(destfile, "%s\n", token);
				fclose(destfile);
				return 0;
			}
		}
	}
	else
	{
		if(WorkPackNotPresent)
			return -1;
	}
	if(iswork)
	{
		if(SemGlobal->nfoundcnt==0)
		{
			SemGlobal->nfoundcnt=1;
			SemGlobal->usesnfound=(USENFOUND*) malloc (sizeof(USENFOUND));
		}
		else
		{
			SemGlobal->nfoundcnt++;
			SemGlobal->usesnfound =
				(USENFOUND*) realloc ((USENFOUND*)SemGlobal->usesnfound, sizeof(USENFOUND)*SemGlobal->nfoundcnt);
		}
		strcpy(SemGlobal->usesnfound[SemGlobal->nfoundcnt-1].useclause, usepath);
		SemGlobal->usesnfound[SemGlobal->nfoundcnt-1].lineno=SemGlobal->lineno;
		strcpy(SemGlobal->usesnfound[SemGlobal->nfoundcnt-1].filename,SemGlobal->actfile);
		SemGlobal->usesnfound[SemGlobal->nfoundcnt-1].issth=0;
	}
	else
		semerror(ERRNFPACKAGE, usepath);
	return 0;
}//UseParseInSubProgram


int entityparse(char *destpath, int level)
{
	char bufor[BUFSIZE], slxpath[PATHSIZE], dpath[PATHSIZE], valuef[PATHSIZE], usepath[PATHSIZE], mode[4];
	char entityname[NAMESIZE], typespath[PATHSIZE], endname[BUFSIZE], defpath[PATHSIZE], idpath[PATHSIZE];
	FILE *dest,*srcdest;
	int value, i, hasNonComputables=0, hasTypes=0;
	SemGlobal->defaultmode=1;/* default mode IN */
	SemGlobal->ListOfGenerics=NULL;
	SemGlobal->GenericCnt=0;
	mkdir(destpath);
	//na poczatku jednostki znajduje sie jej identyfikator, ktory bedzie nazwa pliku
	fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt nazwy jednostki
	getname(bufor, entityname);
	for( i=0 ; i<SemGlobal->ileencji ; i++)
	{
		if(!stricmp(SemGlobal->entities[i].name, entityname))
			break;
	}
	if(i!=SemGlobal->ileencji && SemGlobal->ileencji)
	{
		if(SemGlobal->entities[i].isentity==1)
		{
			semerror(ERRISSUCHENTITY, entityname);
			return -1;
		}
		else//byla architektura, juz jest jednostka
			SemGlobal->entities[SemGlobal->ileencji-1].isentity=1;
	}
	else
	{
		newentity();
		strcpy(SemGlobal->entities[SemGlobal->ileencji-1].name, entityname);
		SemGlobal->entities[SemGlobal->ileencji-1].isarch=0;
		SemGlobal->entities[SemGlobal->ileencji-1].isentity=1;
	}
	sprintf(typespath, "%s\\%s.tdf", destpath, entityname);
	strcpy(SemGlobal->actEntityContext, entityname); 
	WritePredefinedTypes(typespath);
	sprintf(valuef, "%s\\%s.val", destpath, entityname);
	if((dest=fopen(valuef, "wt")))
	{
		fprintf(dest, "# 0 Czy jest dostepny pakiet IEEE.std_logic_1164. 1-jest, 0-nie jest\n");
		fprintf(dest, "%d\n", SemGlobal->isIEEE1164avail);
		fprintf(dest, "# 1 Wartosc rodzica aktualnej jednostki.\n");
		fprintf(dest, "%s\n", SemGlobal->actroot);
	}
	else
		return -1;
	fclose(dest);
	sprintf(dpath, "%s\\%s_%d", destpath, entityname, level);
	mkdir(dpath);
	SemGlobal->indexno++;
	SemGlobal->entities[SemGlobal->ileencji-1].entityno=SemGlobal->indexno;
	fprintf(SemGlobal->indexf, "%d 6 %s\t%s\\%s\n", SemGlobal->indexno, entityname, destpath, entityname);
	sprintf(dpath, "%s\\%s", destpath, entityname);
	sprintf(defpath, "%s.def", dpath);
	sprintf(idpath, "%s.id", dpath);
	sprintf(usepath, "%s\\%s.val", destpath, entityname);
	strcpy(SemGlobal->actpath, dpath);
	if(SemGlobal->pathcnt)
	{
		int j;
		if(dest=fopen(usepath, "at"))
			fprintf(dest, "# 3 Sciezki do pakietow uzywanych przez jednostke\n");
		else
			return -1;
		for( i=0 ; i<SemGlobal->pathcnt ; i++)
		{
			for(j=i+1 ; j<SemGlobal->pathcnt ; j++)
			{
				if(!stricmp(SemGlobal->usepaths[i].path, SemGlobal->usepaths[j].path))
					break;
			}
			if(j==SemGlobal->pathcnt)
				fprintf(dest, "%s\n", SemGlobal->usepaths[i].path);
		}
		fclose(dest);
		free(SemGlobal->usepaths);
		SemGlobal->pathcnt=0;
	}
	if(dest=fopen(usepath, "at"))
	{
		fprintf(dest, "# 5 Wartosci semantyczne\n");
		fclose(dest);
	}
	else
		return -1;
	
	if(SemGlobal->nfoundcnt)
	{
		for( i=0 ; i<SemGlobal->nfoundcnt ; i++)
		{
			if(!SemGlobal->usesnfound[i].issth)
			{
				strcpy(SemGlobal->usesnfound[i].path, usepath);
				SemGlobal->usesnfound[i].issth=1;
			}
		}
	}
	sprintf(slxpath, "%s.slx", dpath);
	if((srcdest=fopen(slxpath, "wt")))
		fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt 'is'
	else
		return -1;
	//odczyt do slowa kluczowego end, dozwolone tylko: port i generic
	hasTypes=1;
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, srcdest);
		SemGlobal->hasNonComputables=hasNonComputables;
		SemGlobal->hasTypes=hasTypes;
		if(hasTypes)
			strcpy(mode, "at");
		else
			strcpy(mode, "wt");
		
		if(value==GENERICV)
		{
			GenericParse(ENTITYV);
			continue;
		}
		if(value==PORTV)
		{
			ParseEntityPortList(dpath, ENTITYV);
			continue;
		}
		if(value==SIGNALV)
		{
			yylval.semval=tok_SIGNAL;
			dest=fopen(valuef, "at");
			ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==TYPEV)
		{
			yylval.semval=tok_TYPE;
			ParseTypeDeclaration();
			hasTypes=1;
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			continue;
		}
		if(value==CONSTANTV)
		{
			yylval.semval=tok_CONSTANT;
			dest=fopen(valuef, "at");
			ParseConstantDeclaration(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			fclose(dest);
			continue;
		}
		if(value==ASSERTV || value==ATTRIBV)
		{
			ReadToSemiC();
			continue;
		}
		if(value==ENDV)
		{
			int i;
			char tokstr[BUFSIZE];
			fprintf(srcdest, "%s", bufor);
			/* Lets check generics */
			dest=fopen(valuef, "at");
			if(!dest)
				return -1;
			for(i=0 ; i<SemGlobal->GenericCnt ; i++)
			{
				int tmpmode=decl_status.mode;
				char BinStr[128], DecStr[128];
				int TypeOf=SemGlobal->ListOfGenerics[i].TypeOf;
				if(TypeOf==INTEGERLV || TypeOf==POSITIVELV || TypeOf==NATURALLV || TypeOf==BITLV || TypeOf==BITVECTORLV || TypeOf==STDULOGICLV || TypeOf==STDLOGICLV || TypeOf==SULVECTORLV || TypeOf==SLVECTORLV || TypeOf==STRINGLV || TypeOf==CHARLV || TypeOf==BOOLEANLV)
				{
					int lType, lClass, lMode, lBits, lLeft, lRight;
					char TmpStr[128];
					decl_status.mode=3;//inout
					strcpy(SemGlobal->tmpactpath, SemGlobal->actpath);
					FindTypeString(SemGlobal->ListOfGenerics[i].TypeOf, tokstr, tokstr);
					ReduceTo6Cols(tokstr);
					sscanf(tokstr, "%d\t%d\t%d\t%d\t%d\t%d", &lType, &lClass, &lMode, &lBits, &lLeft, &lRight);
					switch(TypeOf)
					{
					case INTEGERLV:
					case POSITIVELV:
					case NATURALLV: 
						sprintf(DecStr, "%d", SemGlobal->ListOfGenerics[i].value);
						strcpy(BinStr, anytobin(DecStr, 10));
						break;
					case CHARLV: 
						sprintf(DecStr, "%d", SemGlobal->ListOfGenerics[i].value);
						strcpy(TmpStr, anytobin(DecStr, 10));
						strcpy(BinStr, TmpStr+24);
						break;
					case BOOLEANLV: 
					case BITLV: 
					case STDULOGICLV: 
					case STDLOGICLV: 
						sprintf(DecStr, "%d", SemGlobal->ListOfGenerics[i].value);
						strcpy(BinStr, DecStr);
						break;
					case SULVECTORLV:
					case SLVECTORLV:
					case BITVECTORLV:
					case STRINGLV:
						strcpy(BinStr, SemGlobal->ListOfGenerics[i].bitvalue);
						lBits=strlen(BinStr)-2;
						lLeft=0;
						lRight=lBits-1;
						if(TypeOf==STRINGLV)
						{
							lLeft=1;lRight=lBits;
						}
						break;
					}
					sprintf(tokstr, "%d\t16\t%d\t%d\t%d\t%d", lType, lMode, lBits, lLeft, lRight);
					fprintf(dest, "%d\t%s\t%s\t%s\n", SemGlobal->ListOfGenerics[i].LexemId, SemGlobal->ListOfGenerics[i].name, tokstr, BinStr);
					decl_status.mode=tmpmode;
				}
				else
				{
					semwarn("Use of generic of unsupported type in entity");
					printf("\nTypeOf:%d\n", TypeOf);
				}
			}
			fclose(dest);
			break;
		}
		if(value)
			semerror(ERRENTIFORBID, NULL);
	}
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, srcdest);
		if(value==ENTITYV)
		{
			fgets(bufor, BUFSIZE, SemGlobal->srcf);
			value=getvalue(bufor, srcdest);
			if(value==SEMICV)
				break;
			else
			{
				getname(bufor, endname);
				if(stricmp(endname, entityname))
				{
					sprintf(bufor, "entity: %s -> %s", entityname, endname);
					semerror(ERRNAMESMISMATCH, bufor);
					return -1;
				}
				else
				{
					fgets(bufor, BUFSIZE, SemGlobal->srcf);
					value=getvalue(bufor, srcdest);
					if(value==SEMICV)
						break;
					else
					{
						sprintf(bufor, "entity: %s", entityname);
						semerror(ERRFORBIDCHAR, bufor);
					}
				}
			}
		}
		if(value==SEMICV)
			break;
		getname(bufor, endname);
		if(stricmp(endname, entityname))
		{
			sprintf(bufor, "entity: %s -> %s", entityname, endname);
			semerror(ERRNAMESMISMATCH, bufor);
			return -1;
		}
		else
		{
			fgets(bufor, BUFSIZE, SemGlobal->srcf);
			value=getvalue(bufor, srcdest);
			if(value==SEMICV)
				break;
			else
			{
				sprintf(bufor, "entity: %s", entityname);
				semerror(ERRFORBIDCHAR, bufor);
				return -1;
			}
		}
	}
	fclose(srcdest);
	if(SemGlobal->GenericCnt)
	{
		int i;
		for(i=0 ; i<SemGlobal->GenericCnt ; i++)
			free(SemGlobal->ListOfGenerics[i].name);
		free(SemGlobal->ListOfGenerics);
		SemGlobal->ListOfGenerics=NULL;
		SemGlobal->GenericCnt=0;
	}
	SemGlobal->actEntityContext[0]=0;
	SemGlobal->defaultmode=3;/* default mode INOUT */
	SemGlobal->isIEEE1164avail=0;
	return 0;
}//entityparse

void WriteAliasInfo(char *dpath, int hasAliases)
{
	FILE *dest;
	int i, rightlen;
	if(hasAliases)
		dest=fopen(dpath, "at");
	else
		dest=fopen(dpath, "wt");
	if(dest)
	{
		i=0;
		if(!hasAliases)
			fprintf(dest, "11\tkeyword:\tbegin\n");
		NewFullAlias();
		while(SemGlobal->ListForAlias[i].toknum!=SEMICV)
		{
			if(i==0)
				strcpy(SemGlobal->FullListForAlias[SemGlobal->AliasCnt].left, SemGlobal->ListForAlias[i].tokstr);
			//fprintf(dest, "%s", SemGlobal->ListForAlias[i].tokstr);
			i++;
		}
		rightlen=i-2;
		SemGlobal->FullListForAlias[SemGlobal->AliasCnt].rightlen=rightlen;
		SemGlobal->FullListForAlias[SemGlobal->AliasCnt].right=(YYSTYPE*)
			malloc(sizeof(YYSTYPE)*rightlen);
		for(i=0 ; i<rightlen ; i++)
		{
				strcpy(SemGlobal->FullListForAlias[SemGlobal->AliasCnt].right[i].tokstr, SemGlobal->ListForAlias[i+2].tokstr);
		}
		//fprintf(dest, "%s", SemGlobal->ListForAlias[i+2].tokstr);
		fclose(dest);
		SemGlobal->AliasCnt++;
		FreeListForAlias();
	}
}


void TranslateIdent(const char *extname, char *name)//Funkcja konwertujaca extended identifier na standartowa nazwe (dla sciezek, itp.)
{
	if(*extname=='\\')
	{
		sprintf(name, "__extid%d", SemGlobal->LastLexemId);
		SemGlobal->LastLexemId++;
		/* Generacja nowej nazwy sciezkowej,poniewaz
		 * extended idetifiers nie moga byc nazwami plikow */
	}
	else
		strcpy(name, extname);
}//TranslateIdent


int archparse(char *destpath, int level)
{
	char bufor[BUFSIZE], dpath[PATHSIZE], valuef[PATHSIZE], usepath[PATHSIZE], actpath[PATHSIZE];
	char archname[NAMESIZE], entityname[NAMESIZE], mode[3]="at", errmesg[128];
	char defpath[PATHSIZE], idpath[PATHSIZE], endname[BUFSIZE], ArchPathName[PATHSIZE];
	FILE *dest;
	int value, i, hasTypes=0, hasNonComputables=0, hasAliases=0;
	unsigned ileicf=0;//wspolny licznik dla if, case i for
	unsigned ilebegin=0, ileproc=0, tokenscnt=0;

	SemGlobal->isArchitectureParsed=1;
	SemGlobal->defaultmode=3;/* default mode INOUT */
	fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt nazwy architektury
	getname(bufor, archname);
	TranslateIdent(archname, ArchPathName);
	fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt of
	fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt nazwy jednostki
	getname(bufor, entityname);
	fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt is
	SemGlobal->roottype=TYPEARCH;
	for(i=0 ; i<SemGlobal->ileencji ; i++)
	{
		if(stricmp(SemGlobal->entities[i].name, entityname)==0)
			break;
	}
	if(i!=SemGlobal->ileencji && SemGlobal->ileencji>0)
	{
		strcpy(SemGlobal->entities[i].archname, archname);
		SemGlobal->entities[i].isarch=1;
	}
	else
	{
		newentity();
		semwarn("No corresponding entity for architecture");
		SemGlobal->entities[i].isentity=0;//brak jednostki
		SemGlobal->entities[i].isarch=1;//jest architektura
		strcpy(SemGlobal->entities[i].name, entityname);
		strcpy(SemGlobal->entities[i].archname, archname);
		mkdir(destpath);
		sprintf(dpath, "%s\\%s_%d", SemGlobal->actroot, entityname, level);
		mkdir(dpath);
	}
	if(SemGlobal->archcnt==0)
	{
		SemGlobal->archcnt=1;
		SemGlobal->archs=(ARCH*) malloc (sizeof(ARCH));
	}
	else
	{
		int i;
		ARCH *tmparchs;
		tmparchs =(ARCH*) malloc ((SemGlobal->archcnt+1) * sizeof(ARCH));
		for(i=0 ; i<SemGlobal->archcnt ; i++)
		{
			tmparchs[i].archno=SemGlobal->archs[i].archno;
			strcpy(tmparchs[i].archname, SemGlobal->archs[i].archname);
			strcpy(tmparchs[i].entityname, SemGlobal->archs[i].entityname);
		}
		SemGlobal->archcnt++;
		free(SemGlobal->archs);
		SemGlobal->archs=tmparchs;
	}
	strcpy(SemGlobal->archs[SemGlobal->archcnt-1].archname, archname); 
	strcpy(SemGlobal->archs[SemGlobal->archcnt-1].entityname, entityname); 
	strcpy(SemGlobal->actEntityContext, entityname); 
	sprintf(valuef, "%s\\%s_%d\\%s.val", destpath, entityname, level, ArchPathName);
	if((dest=fopen(valuef, "wt")))
	{
		fprintf(dest, "# 0 Czy jest dostepny pakiet IEEE.std_logic_1164. 1-jest, 0-nie jest\n");
		fprintf(dest, "%d\n", SemGlobal->isIEEE1164avail);
		fprintf(dest, "# 1 Wartosc semantyczna rodzica aktualnej architektury.\n");
		fprintf(dest, "%s\n", entityname);
		fprintf(dest, "# 2 Dodatkowe nazwy dla sygnalow i zmiennych. Pierwsza linijka sygnaly, druga zmienna.\n");
		fprintf(dest, "NULL\n");
		fprintf(dest, "NULL\n");
	}
	else
		return -1;
	fprintf(dest, "# 5 Wartosci semantyczne.\n");
	fclose(dest);
	sprintf(actpath, "%s\\%s_%d\\%s", SemGlobal->actroot, entityname, level, ArchPathName);
	sprintf(dpath, "%s\\%s_%d", SemGlobal->actroot, entityname, level);
	sprintf(defpath, "%s\\%s_%d\\%s.def", SemGlobal->actroot, entityname, level, ArchPathName);
	sprintf(idpath, "%s\\%s_%d\\%s.id", SemGlobal->actroot, entityname, level, ArchPathName);
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, NULL);
		SemGlobal->defaultmode=3;/* default mode INOUT */
		SemGlobal->hasTypes=hasTypes;
		SemGlobal->hasNonComputables=hasNonComputables;
		strcpy(SemGlobal->actpath, actpath);
		//pierwszy begin - otwarcie pliku z kopia biezacej architektury
		if(value==BEGINV)
		{
			ilebegin=1;
			sprintf(dpath, "%s\\%s_%d\\%s", destpath, entityname, level, ArchPathName);
			sprintf(usepath, "%s\\%s_%d\\%s.val", destpath, entityname, level, ArchPathName);
			SemGlobal->indexno++;
			SemGlobal->entities[SemGlobal->ileencji-1].archno=SemGlobal->indexno;
			SemGlobal->archs[SemGlobal->archcnt-1].archno=SemGlobal->indexno;
			fprintf(SemGlobal->indexf, "%d 7 %s\t%s\n", SemGlobal->indexno, archname, dpath);
			if(SemGlobal->pathcnt)
			{
				if(dest=fopen(usepath, "at"))
					fprintf(dest, "# 3 Sciezki do pakietow uzywanych przez architekture\n");
				else
					return -1;
				for( i=0 ; i<SemGlobal->pathcnt ; i++)
					fprintf(dest, "%s\n", SemGlobal->usepaths[i].path);
				fclose(dest);
				free(SemGlobal->usepaths);
				SemGlobal->pathcnt=0;
			}
			if(SemGlobal->nfoundcnt)
			{
				for( i=0 ; i<SemGlobal->nfoundcnt ; i++)
				{
					if(!SemGlobal->usesnfound[i].issth)
					{
						strcpy(SemGlobal->usesnfound[i].path, usepath);
						SemGlobal->usesnfound[i].issth=1;
					}
				}
			}
			if(hasAliases)
			{
				if((dest=fopen(dpath, "at"))==NULL)
					return -1;
			}
			else
			{
				if((dest=fopen(dpath, "wt")))
					fprintf(dest, "11\tkeyword:\tbegin\n");
				else
					return -1;
			}
			break;
		}
		if(value==PROCV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d\\%s_%d", destpath, entityname, level-1, ArchPathName, level);
			strcpy(SemGlobal->actroot, archname);
			if(proparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t0\t4\t0\t0\t0\t0\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName);
			fclose(dest);
			level--;
			continue;
		}
		if(value==FUNCV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			SemGlobal->LastIsPrototype=0;
			sprintf(dpath, "%s\\%s_%d\\%s_%d", destpath, entityname, level-1, ArchPathName, level);
			strcpy(SemGlobal->actroot, archname);
			if(funcparse(dpath, level)==-1)
				return -1;
			if(!SemGlobal->LastIsPrototype)
			{
				dest=fopen(valuef, mode);
				fprintf(dest, "%d\t%s\t%d\t2\t%s\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName, SemGlobal->LastReturnValue, SemGlobal->LastReturnString);
				fclose(dest);
			}
			level--;
			continue;
		}
		if(value==TYPEV)
		{
			yylval.semval=tok_TYPE;
			dest=fopen(valuef, mode);
			ParseTypeDeclaration();
			hasTypes=1;
			fclose(dest);
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			hasNonComputables=SemGlobal->hasNonComputables;
			continue;
		}
		if(value==SUBTYPEV)
		{
			ParseSubTypeDeclaration();
			hasTypes=SemGlobal->hasTypes;
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			hasNonComputables=SemGlobal->hasNonComputables;
			continue;
		}
		if(value==CONSTANTV)
		{
			yylval.semval=tok_CONSTANT;
			dest=fopen(valuef, mode);
			ParseConstantDeclaration(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			//ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==SIGNALV)
		{
			yylval.semval=tok_SIGNAL;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			fclose(dest);
			continue;
		}
		if(value==VARV)
		{
			yylval.semval=tok_VARIABLE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			fclose(dest);
			continue;
		}
		if(value==SHAREDV)
		{
			yylex();
			yylval.semval=tok_VARIABLE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			fclose(dest);
			continue;
		}
		if(value==ALIASV)
		{
			yylval.semval=tok_ALIAS;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			sprintf(dpath, "%s\\%s_%d\\%s", destpath, entityname, level, ArchPathName);
			WriteAliasInfo(dpath, hasAliases);
			hasAliases=1;
			continue;
		}
		if(value==COMPONV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d\\%s_%d", destpath, entityname, level-1, ArchPathName, level);
			strcpy(SemGlobal->actroot, archname);
			if(componparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t0\t8\t0\t0\t0\t0\t0\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName);
			fclose(dest);
			level--;
			continue;
		}
		if(value==ASSERTV || value==ATTRIBV)
		{
			ReadToSemiC();
			continue;
		}
		sprintf(errmesg, ": %s", bufor);
		semerror(ERRUNSCONSTRUCT, errmesg);
	}
	if(hasTypes)
		enumsdef2enumsid(defpath, idpath);
	// TODO: wejscie do analizy procesu
	
	sprintf(dpath, "%s\\%s_%d\\%s_%d", destpath, entityname, level, ArchPathName, level+1);
	BodyParsing(dest, dpath, archname, level, valuef);
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		if(value==ARCHV)
		{
			fgets(bufor, BUFSIZE, SemGlobal->srcf);
			value=getvalue(bufor, dest);
			if(value==SEMICV)
				break;
			else
			{
				getname(bufor, endname);
				if(stricmp(endname, archname))
				{
					sprintf(bufor, "architecture: %s -> %s", archname, endname);
					semerror(ERRNAMESMISMATCH, bufor);
					return -1;
				}
				else
				{
					fgets(bufor, BUFSIZE, SemGlobal->srcf);
					value=getvalue(bufor, dest);
					if(value==SEMICV)
						break;
					else
					{
						sprintf(bufor, "architecture: %s", archname);
						semerror(ERRFORBIDCHAR, bufor);
					}
				}
			}
		}
		if(value==SEMICV)
			break;
		getname(bufor, endname);
		if(stricmp(endname, archname))
		{
			sprintf(bufor, "architecture: %s -> %s", archname, endname);
			semerror(ERRNAMESMISMATCH, bufor);
			return -1;
		}
		else
		{
			fgets(bufor, BUFSIZE, SemGlobal->srcf);
			value=getvalue(bufor, dest);
			if(value==SEMICV)
				break;
			else
			{
				sprintf(bufor, "architecture: %s", archname);
				semerror(ERRFORBIDCHAR, bufor);
				return -1;
			}
		}
	}
	fclose(dest);
	SemGlobal->actEntityContext[0]=0;
	SemGlobal->isArchitectureParsed=0;
	SemGlobal->isIEEE1164avail=0;
	return 0;
}//archparse


int GenerateParse(char *destpath, int level, char *gnname, char *lbufor)
{
	char bufor[BUFSIZE], tmpbuf[BUFSIZE], ProcessName[BUFSIZE], dpath[PATHSIZE], valuef[PATHSIZE], usepath[PATHSIZE], actpath[PATHSIZE];
	char componname[NAMESIZE], mode[3]="at", errmesg[128];
	char defpath[PATHSIZE], idpath[PATHSIZE], endname[BUFSIZE];
	FILE *dest, *vfile;
	int value, i, hasTypes=0, hasNonComputables=0, hasAliases=0, gnnumber, isRead=1, componvalue=0;
	unsigned ileicf=0;//wspolny licznik dla if, case i for
	unsigned ilebegin=0, ileproc=0, tokenscnt=0, defaultmode;
	defaultmode=SemGlobal->defaultmode;
	SemGlobal->defaultmode=3;/* default mode INOUT */
	mkdir(destpath);
	gnnumber=SemGlobal->indexno;
	sprintf(actpath, "%s\\%s", destpath, gnname);
	sprintf(valuef, "%s.val", actpath);
	sprintf(defpath, "%s.def", actpath);
	sprintf(idpath, "%s.id", actpath);

	if((dest=fopen(valuef, "wt")))
	{
		fprintf(dest, "# 0 Czy jest dostepny pakiet IEEE.std_logic_1164. 1-jest, 0-nie jest\n");
		fprintf(dest, "%d\n", SemGlobal->isIEEE1164avail);
		fprintf(dest, "# 1 Wartosc semantyczna rodzica aktualnej architektury.\n");
		fprintf(dest, "%s\n", SemGlobal->actroot);
		fprintf(dest, "# 2 Dodatkowe nazwy dla sygnalow i zmiennych. Pierwsza linijka sygnaly, druga zmienna.\n");
		fprintf(dest, "NULL\n");
		fprintf(dest, "NULL\n");
	}
	else
		return -1;
	fprintf(dest, "# 5 Wartosci semantyczne.\n");
	fclose(dest);
	sprintf(dpath, "%s\\%s", destpath, gnname);
	i=0;
	dest=fopen(dpath, "wt");
	fprintf(dest, "%s", lbufor);
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		fprintf(dest, "%s", bufor);//kopiuje wszystko az do generate
		if(value==GENERATEV)
			break;
	}
	fclose(dest);
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, NULL);
		SemGlobal->hasTypes=hasTypes;
		SemGlobal->hasNonComputables=hasNonComputables;
		strcpy(SemGlobal->actpath, actpath);
		if(!i)//Czy jest to poczatek ciala generate
		{
			i=1;
		}
		else
		{
			if(i=1)
				i=2;
		}
		//pierwszy begin - otwarcie pliku z kopia biezacej architektury
		if(value==BEGINV)
		{
			ilebegin=1;
			sprintf(dpath, "%s", actpath);
			sprintf(usepath, "%s.val", actpath);
			SemGlobal->indexno++;
			i=2;
			fprintf(SemGlobal->indexf, "%d 10 %s\t%s\n", SemGlobal->indexno, gnname, dpath);
			if(hasAliases)
			{
				if((dest=fopen(dpath, "at"))==NULL)
					return -1;
			}
			else
			{
				if((dest=fopen(dpath, "at")))
					fprintf(dest, "11\tkeyword:\tbegin\n");
				else
					return -1;
			}
			break;
		}
		if(value==PROCV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s_%d", actpath, level);
			strcpy(SemGlobal->actroot, gnname);
			if(proparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t0\t4\t0\t0\t0\t0\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName);
			fclose(dest);
			level--;
			continue;
		}
		if(value==FUNCV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s_%d", actpath, level);
			strcpy(SemGlobal->actroot, gnname);
			if(funcparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t%d\t2\t%s\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName, SemGlobal->LastReturnValue, SemGlobal->LastReturnString);
			fclose(dest);
			level--;
			continue;
		}
		if(value==TYPEV)
		{
			yylval.semval=tok_TYPE;
			dest=fopen(valuef, mode);
			//ParseDeclarativePart(dest);
			ParseTypeDeclaration();
			hasTypes=1;
			fclose(dest);
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			continue;
		}
		if(value==SUBTYPEV)
		{
			ParseSubTypeDeclaration();
			hasTypes=SemGlobal->hasTypes;
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			hasNonComputables=SemGlobal->hasNonComputables;
			continue;
		}
		if(value==CONSTANTV)
		{
			yylval.semval=tok_CONSTANT;
			dest=fopen(valuef, mode);
			ParseConstantDeclaration(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			//ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==SIGNALV)
		{
			yylval.semval=tok_SIGNAL;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			fclose(dest);
			continue;
		}
		if(value==VARV)
		{
			yylval.semval=tok_VARIABLE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			fclose(dest);
			continue;
		}
		if(value==SHAREDV)
		{
			yylex();
			yylval.semval=tok_VARIABLE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			fclose(dest);
			continue;
		}
		if(value==ALIASV)
		{
			yylval.semval=tok_ALIAS;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			sprintf(dpath, "%s_%d", actpath, level);
			WriteAliasInfo(dpath, 1);
			hasAliases=1;
			continue;
		}
		if(value==COMPONV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s_%d", actpath, level);
			strcpy(SemGlobal->actroot, gnname);
			if(componparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t0\t8\t0\t0\t0\t0\t0\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName, SemGlobal->LastFoundValue, SemGlobal->LastFoundName);
			fclose(dest);
			level--;
			continue;
		}
		if(value==ASSERTV || value==ATTRIBV)
		{
			ReadToSemiC();
			continue;
		}
		//generate nie musi sie zaczynac slowem kluczowym begin
		if(i==1)
		{
			ilebegin=1;
			sprintf(dpath, "%s", actpath);
			sprintf(usepath, "%s.val", actpath);
			SemGlobal->indexno++;
			strcpy(tmpbuf, bufor);
			fprintf(SemGlobal->indexf, "%d 10 %s\t%s\n", SemGlobal->indexno, gnname, dpath);
			if(hasAliases)
			{
				if((dest=fopen(dpath, "at"))==NULL)
					return -1;
			}
			else
			{
				if((dest=fopen(dpath, "at")))
					fprintf(dest, "11\tkeyword:\tbegin\n");
				else
					return -1;
			}
			break;
		}
	}
	if(hasTypes)
		enumsdef2enumsid(defpath, idpath);
	// TODO: wejscie do analizy procesu
	

	while(isRead)
	{
		int NameValue;
		if(i==1)
		{
			strcpy(bufor, tmpbuf);
			i=2;
			isRead=1;
		}
		else
		{
			if(fgets(bufor, BUFSIZE, SemGlobal->srcf))
				isRead=1;
			else
				isRead=0;
		}
		if(!isRead)
			break;
		value=getvalue(bufor, dest);
		NameValue=value;
		if(value>=IDENTV)
		{
			strcpy(ProcessName, bufor);//ewentualna nazwa procesu lub etykieta
			if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
			{
				fprintf(dest, "%s", ProcessName);
				semerror(ERRUNEXEOF, NULL);
				break;
			}
			value=getvalue(bufor, dest);
			if(value==COLONV)
			{
				if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
				{
					fprintf(dest, "%s", ProcessName);
					fprintf(dest, "113\tspecial_character:\t:\n");
					semerror(ERRUNEXEOF, NULL);
					break;
				}
				value=getvalue(bufor, dest);
				if(value==ASSERTV || value==ATTRIBV)
				{
					ReadToSemiC();
					if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
					{
						fprintf(dest, "%s", ProcessName);
						fprintf(dest, "113\tspecial_character:\t:\n");
						semerror(ERRUNEXEOF, NULL);
						break;
					}
					value=getvalue(bufor, dest);
				}
				if(value==BLOCKV)//blok z nazwa (nazwa bloku musi wystapic)
				{
					char prname[NAMESIZE];
					getname(ProcessName, prname);
					if(ileproc==0)
						ileproc=1;
					level++;
					sprintf(dpath, "%s_%d", actpath, level);
					strcpy(SemGlobal->actroot, gnname);
					if(BlockParse(dpath, level, prname)==-1)
						return -1;
					//fprintf(dest, "%d\tkeyword:\tprocess\n200\tidentifier:\t%s\n", PROCESSV, prname);
					fprintf(dest,
						"%d\tkeyword:\tblock\n%d\tidentifier:\t%s\n",
						BLOCKV, NameValue, prname);
					level--;
					continue;				   
				}
				if(value==FORV || value==IFV)//generate z etykieta
				{
					if(ileproc==0)
						ileproc=1;
					level++;
					sprintf(dpath, "%s_%d", actpath, level);
					strcpy(SemGlobal->actroot, gnname);
					if(NoNamedGenerateParse(dpath, level, bufor)==-1)
						return -1;
					fprintf(dest,
						"%d\tkeyword:\tgenerate\n%d\tidentifier:\t%s\n",
						GENERATEV, SemGlobal->LastLexemId+200, SemGlobal->LastFoundName);
					WriteLastLexemId(SemGlobal->LastFoundName);
					level--;
					continue;
				}
				if(value==PROCESSV)//proces z nazwa
				{
					char prname[NAMESIZE];
					int ModifiedValue;
					getname(ProcessName, prname);
					ModifiedValue=AddToProcessList(prname);
					if(ModifiedValue)
						NameValue=ModifiedValue;
					if(ileproc==0)
						ileproc=1;
					level++;
					sprintf(dpath, "%s_%d", actpath, level);
					strcpy(SemGlobal->actroot, gnname);
					if(ProcessParse(dpath, level, prname)==-1)
						return -1;
					if(*prname!='\\')
						fprintf(dest,
							"%d\tkeyword:\tprocess\n%d\tidentifier:\t%s\n",
							PROCESSV, NameValue, prname);
					else
					{
						fprintf(dest,
							"%d\tkeyword:\tprocess\n%d\tidentifier:\t%s\n",
							PROCESSV, SemGlobal->LastLexemId+200, SemGlobal->LastFoundName);
						WriteLastLexemId(SemGlobal->LastFoundName);
					}
					level--;
					continue;				   
				}
				else//prawdopodobnie nazwa componentu lub zwykla etykieta
				{
					fprintf(dest, "%s", ProcessName);
					fprintf(dest, "113\tspecial_character:\t:\n");
					value=getvalue(bufor, dest);
					if(value>200 || value==COMPONV)
					{
						if(value>200)
							componvalue=value;
						if(value==COMPONV)
						{
							if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
							{
								fprintf(dest, "%d", value);
								fprintf(dest, "\tidentifier:\t:%s\n", componname);
								semerror(ERRUNEXEOF, NULL);
								break;
							}
							value=getvalue(bufor, NULL);
							fprintf(dest, "17\tkeyword:\tcomponent\n");
						}
						AliasCheckAndReplace(bufor);//W przypadku kiedy zostanie zmodyfikowana nazwa jednostki
						//nalezy dolozyc old i new wartosci
						value=getvalue(bufor, NULL);
						getname(bufor, componname);
						if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
						{
							fprintf(dest, "%d", value);
							fprintf(dest, "\tidentifier:\t:%s\n", componname);
							semerror(ERRUNEXEOF, NULL);
							break;
						}
						fprintf(dest, "%d", value);
						fprintf(dest, "\tidentifier:\t%s\n", componname);
						value=getvalue(bufor, dest);
						if(value!=PORTV && value!=GENERICV)
						{
							fprintf(dest, "%s", bufor);
							continue;
						}
						else
							fprintf(dest, "%s", bufor);
						if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
						{
							semerror(ERRUNEXEOF, NULL);
							break;
						}
						value=getvalue(bufor, dest);
						if(value!=MAPV)
						{
							
							sprintf(errmesg, ": %s", bufor);
							semerror(ERRUNSCONSTRUCT, errmesg);
						}
						else
							fprintf(dest, "%s", bufor);
						CopyPortMap(dest, componname);
						NewComponentName(componname);
						if(componvalue)
						{
							vfile=fopen(valuef, "at");
							fprintf(vfile, "%d\t%s\t0\t6\t0\t0\t0\t0\n", componvalue, componname);
							fclose(vfile);
						}
						continue;
					}
				}
			}
			else
			{
				AliasCheckAndReplace(ProcessName);
				fprintf(dest, "%s", ProcessName);
				fprintf(dest, "%s", bufor);
				continue;
			}
		}
		if(value==PROCESSV)//process bez nazwy
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s_%d", actpath, level);
			strcpy(SemGlobal->actroot, gnname);
			if(NoNamedProcessParse(dpath, level)==-1)
				return -1;
			//fprintf(dest, "%d\tkeyword:\tprocess\n200\tidentifier:\t%s\n", PROCESSV, SemGlobal->LastFoundName);
			fprintf(dest,
				"%d\tkeyword:\tprocess\n%d\tidentifier:\t%s\n",
				PROCESSV, SemGlobal->LastLexemId+200, SemGlobal->LastFoundName);
			WriteLastLexemId(SemGlobal->LastFoundName);
			level--;
			continue;
		}
		if(value==FORV || value==IFV)//generate bez nazwy
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s_%d", actpath, level);
			strcpy(SemGlobal->actroot, gnname);
			if(NoNamedGenerateParse(dpath, level, bufor)==-1)
				return -1;
			fprintf(dest,
				"%d\tkeyword:\tgenerate\n%d\tidentifier:\t%s\n",
				GENERATEV, SemGlobal->LastLexemId+200, SemGlobal->LastFoundName);
			WriteLastLexemId(SemGlobal->LastFoundName);
			level--;
			continue;
		}
		if(value==ASSERTV || value==ATTRIBV)
		{
			ReadToSemiC();
			continue;
		}
		if(value==ENDV)
		{
			if(!ileicf)
				ilebegin--;
			else
			{
				fprintf(dest, "%s", bufor);
				fgets(bufor, BUFSIZE, SemGlobal->srcf);
				value=getvalue(bufor, dest);
				if(value==IFV || value==CASEV || value==FORV || value==LOOPV || value==SEMICV || value==GENERATEV)
					ileicf--;
				fprintf(dest, "%s", bufor);
				continue;
			}
		}
		if(ilebegin==0)
			break;
		if(value==BEGINV)
			ilebegin++;
		if(value==IFV || value==CASEV || value==FORV || value==LOOPV)
			ileicf++;
		fprintf(dest, "%s", bufor);
	}
	fprintf(dest, "%s", bufor);
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		fprintf(dest, "%s", bufor);
		if(value==GENERATEV)
		{
			fgets(bufor, BUFSIZE, SemGlobal->srcf);
			value=getvalue(bufor, dest);
			fprintf(dest, "%s", bufor);
			if(value==SEMICV)
				break;
			else
			{
				getname(bufor, endname);
				if(stricmp(endname, gnname) && strnicmp(gnname, "__gnid__", 8))
				{
					sprintf(bufor, "generate: %s -> %s", gnname, endname);
					semerror(ERRNAMESMISMATCH, bufor);
					return -1;
				}
				else
				{
					fgets(bufor, BUFSIZE, SemGlobal->srcf);
					value=getvalue(bufor, dest);
					fprintf(dest, "%s", bufor);
					if(value==SEMICV)
						break;
					else
					{
						sprintf(bufor, "generate: %s", gnname);
						semerror(ERRFORBIDCHAR, bufor);
					}
				}
			}
		}
		if(value==SEMICV)
			break;
		getname(bufor, endname);
		if(stricmp(endname, gnname))
		{
			sprintf(bufor, "generate: %s -> %s", gnname, endname);
			semerror(ERRNAMESMISMATCH, bufor);
			return -1;
		}
		else
		{
			fgets(bufor, BUFSIZE, SemGlobal->srcf);
			value=getvalue(bufor, dest);
			fprintf(dest, "%s", bufor);
			if(value==SEMICV)
				break;
			else
			{
				sprintf(bufor, "Ngenerate: %s", gnname);
				semerror(ERRFORBIDCHAR, bufor);
				return -1;
			}
		}
	}
	fclose(dest);
	strcpy(SemGlobal->LastFoundName, gnname);
	SemGlobal->LastFoundValue=gnnumber;
	SemGlobal->isIEEE1164avail=0;
	SemGlobal->defaultmode=defaultmode;/*return to previous value*/
	return 0;
}//GenerateParse



int configparse(void)
{
	char bufor[BUFSIZE], name[NAMESIZE];
	int cnt=0, i;
	if(SemGlobal->configcnt==0)
	{
		SemGlobal->configcnt=1;
		SemGlobal->configs=(CONFIG*) malloc (sizeof(CONFIG));
	}
	else
	{
		SemGlobal->configcnt++;
		SemGlobal->configs =
			(CONFIG*) realloc ((CONFIG*)SemGlobal->configs, sizeof(CONFIG)*SemGlobal->configcnt);
	}
	i=SemGlobal->configcnt-1;
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		if(getvalue(bufor, NULL))
		{
			switch(cnt)
			{
			case 0:
				getname(bufor, name);
				strcpy(SemGlobal->configs[i].configname, name);
				break;
			case 2:
				getname(bufor, name);
				strcpy(SemGlobal->configs[i].entityname, name);
				break;
			case 5:
				getname(bufor, name);
				strcpy(SemGlobal->configs[i].archname, name);
				break;
			}
			cnt++;
		}
		if(cnt>6 && getvalue(bufor, NULL)==SEMICV)//srednik po end for
			break;
	}
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		if(getvalue(bufor, NULL)==SEMICV)
			break;
	}
	return 0;
}//configparse




int packparse(char *destpath, int level)
{
	char bufor[BUFSIZE], dpath[PATHSIZE], valuef[PATHSIZE], actpath[PATHSIZE];
	char packname[NAMESIZE], endname[NAMESIZE], mode[3], errmesg[128];
	char defpath[PATHSIZE], idpath[PATHSIZE];
	FILE *dest;
	int value, hasTypes=0, hasNonComputables=0, i, isPackageBody;
	unsigned long IsCreated=0;
	mkdir(destpath);
	strcpy(mode, "wt");
	SemGlobal->roottype=TYPEPACK;
	strcpy(SemGlobal->actEntityContext, "__NULL__"); 
	//na poczatku pakietu znajduje sie jego nazwa, ktora bedzie nazwa pliku
	fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt nazwy pakietu lub slowa body
	value=getvalue(bufor, NULL);
	if(value==BODYV)
	{
		//!!!Dodatkowy mechanizm sprawdzania czy jest plik z wartosciami
		//semantycznymi dla pakietu
		isPackageBody=1;
		strcpy(mode, "at");
		fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt nazwy pakietu
		getname(bufor, packname);
		sprintf(dpath, "%s\\%s", destpath, packname);
		mkdir(dpath);
		sprintf(valuef, "%s\\%s.val", dpath, packname);
		strcpy(SemGlobal->rootsname, packname);
		fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt is
		while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
		{
			value=getvalue(bufor, NULL);
			if(value==USEV)
			{
				if(UseParse()==-1)
					return -1;
			}
			if(value==PROCV)
			{
				if(proparse(dpath, 1)==-1)
					return -1;
				dest=fopen(valuef, mode);
				fprintf(dest, "%d\t%s\t0\t4\t0\t0\t0\t0\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName);
				fclose(dest);
				continue;
			}
			if(value==FUNCV)
			{
				if(funcparse(dpath, 1)==-1)
					return -1;
				dest=fopen(valuef, mode);
				fprintf(dest, "%d\t%s\t%d\t2\t%s\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName, SemGlobal->LastReturnValue, SemGlobal->LastReturnString);
				fclose(dest);
				continue;
			}
			if(value==ENDV)
			{
				for(i=0 ; i<SemGlobal->protoscnt ; i++)
				{
					if(SemGlobal->fprotos[i].hasBody==0)
					{
						SemGlobal->indexno++;
						if(SemGlobal->IsSTLACompiled)
						{
							char syslibpath[PATHSIZE];
							switch(SemGlobal->IsSTLACompiled)
							{
							case 1:
								sprintf(syslibpath, "..\\syslib\\Ieee\\std_logic_arith\\%s", SemGlobal->fprotos[i].name);
								break;
							case 2:
								sprintf(syslibpath, "..\\syslib\\Ieee\\std_logic_signed\\%s", SemGlobal->fprotos[i].name);
								break;
							case 3:
								sprintf(syslibpath, "..\\syslib\\Ieee\\std_logic_unsigned\\%s", SemGlobal->fprotos[i].name);
								break;
							case 4:
								sprintf(syslibpath, "..\\syslib\\Ieee\\numeric_bit\\%s", SemGlobal->fprotos[i].name);
								break;
							}
							fprintf(SemGlobal->indexf, "%d 2 %s\t%s\n", SemGlobal->indexno, SemGlobal->fprotos[i].name, syslibpath);
						}
						else
							fprintf(SemGlobal->indexf, "%d %d %s\t%s\n", SemGlobal->indexno, SemGlobal->fprotos[i].TypeOf, SemGlobal->fprotos[i].name, SemGlobal->fprotos[i].fname);
					}
				}
				break;
			}
			if(value==TYPEV)
			{
				yylval.semval=tok_TYPE;
				dest=fopen(valuef, mode);
				//ParseDeclarativePart(dest);
				ParseTypeDeclaration();
				fclose(dest);
				hasTypes=1;
				IsCreated++;
				hasNonComputables=SemGlobal->hasNonComputables;
				if(hasTypes)
					enumsdef2enumsid(defpath, idpath);
				continue;
			}
			if(value==SUBTYPEV)
			{
				ParseSubTypeDeclaration();
				hasTypes=SemGlobal->hasTypes;
				if(hasTypes)
					enumsdef2enumsid(defpath, idpath);
				hasNonComputables=SemGlobal->hasNonComputables;
				continue;
			}
			if(value==CONSTANTV)
			{
				yylval.semval=tok_CONSTANT;
				dest=fopen(valuef, mode);
				ParseConstantDeclaration(dest);
				hasNonComputables=SemGlobal->hasNonComputables;
				//ParseDeclarativePart(dest);
				fclose(dest);
				IsCreated++;
				continue;
			}
			if(value==VARV)
			{
				yylval.semval=tok_VARIABLE;
				dest=fopen(valuef, mode);
				ParseDeclarativePart(dest);
				fclose(dest);
				hasNonComputables=SemGlobal->hasNonComputables;
				IsCreated++;
				continue;
			}
			if(value==SHAREDV)
			{
				yylex();
				yylval.semval=tok_VARIABLE;
				dest=fopen(valuef, mode);
				ParseDeclarativePart(dest);
				hasNonComputables=SemGlobal->hasNonComputables;
				fclose(dest);
				continue;
			}
			if(value==ASSERTV || value==ATTRIBV)
			{
				ReadToSemiC();
				continue;
			}
			sprintf(errmesg, ": %s", bufor);
			semerror(ERRUNSCONSTRUCT, errmesg);
		}
	}
	else
	{
		char cusepath[PATHSIZE];
		FILE *cusef;
		isPackageBody=0;
		SemGlobal->protoscnt=0;
		getname(bufor, packname);
#if __DEVEL__VER__ONLY__
		if(!stricmp("std_logic_arith", packname))
			SemGlobal->IsSTLACompiled=1;
		if(!stricmp("std_logic_signed", packname))
			SemGlobal->IsSTLACompiled=2;
		if(!stricmp("std_logic_unsigned", packname))
			SemGlobal->IsSTLACompiled=3;
		if(!stricmp("numeric_bit", packname))
			SemGlobal->IsSTLACompiled=4;
#endif
		sprintf(dpath, "%s\\%s", destpath, packname);
		mkdir(dpath);
		sprintf(cusepath, "%s\\clause.use", dpath);
		sprintf(defpath, "%s\\%s.def", dpath, packname);
		sprintf(idpath, "%s\\%s.id", dpath, packname);

		if(SemGlobal->pathcnt)
		{
			cusef=fopen(cusepath, "wt");
			if(cusef)
			{
				for( i=0 ; i<SemGlobal->pathcnt ; i++)
					fprintf(cusef, "%s\n", SemGlobal->usepaths[i].path);
				fclose(cusef);
				if(SemGlobal->pathcnt)
					free(SemGlobal->usepaths);
				SemGlobal->pathcnt=0;
			}
		}
		sprintf(valuef, "%s\\%s.val", dpath, packname);
		if(SemGlobal->isIEEE1164avail)
		{
			FILE *availf;
			char availp[PATHSIZE];
			sprintf(availp, "%s\\%s.avl", dpath, packname);
			availf=fopen(availp, "wt");
			fprintf(availf, "# 0 Czy jest dostepny pakiet IEEE.std_logic_1164. plik istnieje 1-jest, pliku brak-pakiet niedostepny\n");
			fprintf(availf, "%d\n", SemGlobal->isIEEE1164avail);
			fclose(availf);
		}
		strcpy(SemGlobal->rootsname, packname);
		fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt is
		sprintf(actpath, "%s\\%s", dpath, packname);
		while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
		{
			value=getvalue(bufor, NULL);
			SemGlobal->isInPackage=1;
			SemGlobal->hasTypes=hasTypes;
			SemGlobal->hasNonComputables=hasNonComputables;
			strcpy(SemGlobal->actpath, actpath);
			if(IsCreated==1)
				strcpy(mode, "at");
			//!!!!!!//parsowanie tylko czesci deklaracyjnej procedury lub funkcji
			if(value==PROCV)
			{
				if(ProcedurePrototypeParse(dpath, 1)==-1)
					return -1;
				dest=fopen(valuef, mode);
				fprintf(dest, "%d\t%s\t0\t4\t0\t0\t0\t0\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName);
				fclose(dest);
				IsCreated++;
				continue;
			}
			if(value==FUNCV)
			{
				if(FuncPrototypeParse(dpath, 1)==-1)
					return -1;
				dest=fopen(valuef, mode);
				fprintf(dest, "%d\t%s\t%d\t2\t%s\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName, SemGlobal->LastReturnValue, SemGlobal->LastReturnString);
				fclose(dest);
				IsCreated++;
				continue;
			}
			if(value==TYPEV)
			{
				yylval.semval=tok_TYPE;
				dest=fopen(valuef, mode);
				//ParseDeclarativePart(dest);
				ParseTypeDeclaration();
				fclose(dest);
				hasTypes=1;
				IsCreated++;
				hasNonComputables=SemGlobal->hasNonComputables;
				if(hasTypes)
					enumsdef2enumsid(defpath, idpath);
				continue;
			}
			if(value==SUBTYPEV)
			{
				ParseSubTypeDeclaration();
				hasTypes=SemGlobal->hasTypes;
				if(hasTypes)
					enumsdef2enumsid(defpath, idpath);
				hasNonComputables=SemGlobal->hasNonComputables;
				continue;
			}
			if(value==CONSTANTV)
			{
				yylval.semval=tok_CONSTANT;
				dest=fopen(valuef, mode);
				ParseConstantDeclaration(dest);
				hasNonComputables=SemGlobal->hasNonComputables;
				//ParseDeclarativePart(dest);
				fclose(dest);
				IsCreated++;
				continue;
			}
			if(value==SIGNALV)
			{
				yylval.semval=tok_SIGNAL;
				dest=fopen(valuef, mode);
				ParseDeclarativePart(dest);
				fclose(dest);
				hasNonComputables=SemGlobal->hasNonComputables;
				IsCreated++;
				continue;
			}
			if(value==VARV)
			{
				yylval.semval=tok_VARIABLE;
				dest=fopen(valuef, mode);
				ParseDeclarativePart(dest);
				fclose(dest);
				hasNonComputables=SemGlobal->hasNonComputables;
				IsCreated++;
				continue;
			}
			if(value==SHAREDV)
			{
				yylex();
				yylval.semval=tok_VARIABLE;
				dest=fopen(valuef, mode);
				ParseDeclarativePart(dest);
				hasNonComputables=SemGlobal->hasNonComputables;
				fclose(dest);
				continue;
			}
			if(value==USEV)
			{
				if(UseParse()==-1)
					return -1;
				if(SemGlobal->pathcnt)
				{
					cusef=fopen(cusepath, "at");
					if(cusef)
					{
						for( i=0 ; i<SemGlobal->pathcnt ; i++)
							fprintf(cusef, "%s\n", SemGlobal->usepaths[i].path);
						fclose(cusef);
						if(SemGlobal->pathcnt)
							free(SemGlobal->usepaths);
						SemGlobal->pathcnt=0;
					}
				}
				IsCreated++;
				continue;
			}
			if(value==COMPONV)
			{
				if(componparse(dpath, 1)==-1)
					return -1;
				IsCreated++;
				continue;
			}
			if(value==ASSERTV || value==ATTRIBV)
			{
				ReadToSemiC();
				continue;
			}
			if(value==ENDV)
			{
				SemGlobal->isInPackage=0;
				break;
			}
			sprintf(errmesg, ": %s", bufor);
			semerror(ERRUNSCONSTRUCT, errmesg);
		}
	}


	if(hasTypes)
		enumsdef2enumsid(defpath, idpath);
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, NULL);
		if(value==PACKV)
		{
			fgets(bufor, BUFSIZE, SemGlobal->srcf);
			value=getvalue(bufor, NULL);
			if(value==SEMICV)
				break;
			else
			{
				getname(bufor, endname);
				if(stricmp(endname, packname))
				{
					if(isPackageBody)
					{
						fgets(bufor, BUFSIZE, SemGlobal->srcf);
						value=getvalue(bufor, NULL);
						if(value==SEMICV)
							break;
						else
						{
							getname(bufor, endname);
							if(stricmp(endname, packname))
							{
								sprintf(bufor, "package: %s -> %s", packname, endname);
								semerror(ERRNAMESMISMATCH, bufor);
							}
							else
							{
								fgets(bufor, BUFSIZE, SemGlobal->srcf);
								value=getvalue(bufor, NULL);
								if(value==SEMICV)
									break;
								else
								{
									sprintf(bufor, "package: %s", packname);
									semerror(ERRFORBIDCHAR, bufor);
								}
							}
							sprintf(bufor, "package: %s", packname);
							semerror(ERRFORBIDCHAR, bufor);
						}
					}
					else
					{
						sprintf(bufor, "package: %s -> %s", packname, endname);
						semerror(ERRNAMESMISMATCH, bufor);
					}
					return -1;
				}
				else
				{
					fgets(bufor, BUFSIZE, SemGlobal->srcf);
					value=getvalue(bufor, NULL);
					if(value==SEMICV)
						break;
					else
					{
						sprintf(bufor, "package: %s", packname);
						semerror(ERRFORBIDCHAR, bufor);
					}
				}
			}
		}
		if(value==SEMICV)
			break;
		getname(bufor, endname);
		if(stricmp(endname, packname))
		{
			sprintf(bufor, "package: %s -> %s", packname, endname);
			semerror(ERRNAMESMISMATCH, bufor);
			return -1;
		}
		else
		{
			fgets(bufor, BUFSIZE, SemGlobal->srcf);
			value=getvalue(bufor, NULL);
			if(value==SEMICV)
				break;
			else
			{
				sprintf(bufor, "package: %s", packname);
				semerror(ERRFORBIDCHAR, bufor);
				return -1;
			}
		}
	}
	SemGlobal->isIEEE1164avail=0;
	return 0;
}//packparse



int proparse(char *destpath, int level)
{
	char bufor[BUFSIZE], c, dpath[PATHSIZE], valuef[PATHSIZE], actpath[PATHSIZE];
	char procname[NAMESIZE], mode[3]="at", errmesg[128];
	char defpath[PATHSIZE], lname[NAMESIZE], idpath[PATHSIZE];
	FILE *dest;
	int value, namevalue, LastWasFor=0;
	int i, j, hasTypes=0, hasNonComputables=0, hasAliases=0;
	int k, l, tvalue, isParams, lval;
	
	
	
	unsigned ilebegin=0, ileproc=0;
	FPROTO actualproto;		//wartosci semantyczne dla ostatnio znalezionego prototypu
	unsigned ileicf=0;//wspolny licznik dla if, case i for
	mkdir(destpath);
	//na poczatku procedury znajduje sie jej identyfikator, ktory bedzie nazwa pliku
	fgets(bufor, BUFSIZE, SemGlobal->srcf);
	c=bufor[0];
	i=0;
	while(c!=':')
	{
		i++;
		c=bufor[i];
	}
	i++;
	c=bufor[i];
	while(!isalpha(c))
	{
		i++;
		c=bufor[i];
	}
	procname[0]=c;
	j=0;
	while(c!='\n')
	{
		j++;
		i++;
		c=bufor[i];
		procname[j]=c;
	}
	procname[j]=0;
	namevalue=getvalue(bufor, NULL);
	sprintf(defpath, "%s\\%s.def", destpath, procname);
	sprintf(idpath, "%s\\%s.id", destpath, procname);
	sprintf(valuef, "%s\\%s.val", destpath, procname);;
	if((dest=fopen(valuef, "wt")))
	{
 		fprintf(dest, "# 0 Czy jest dostepny pakiet IEEE.std_logic_1164. 1-jest, 0-nie jest\n");
		fprintf(dest, "%d\n", SemGlobal->isIEEE1164avail);
		fprintf(dest, "# 1 Wartosc rodzica aktualnej procedury.\n");
		fprintf(dest, "%s\n", SemGlobal->actroot);
		fprintf(dest, "# 2 Dodatkowe nazwy dla sygnalow i zmiennych. Pierwsza linijka sygnaly, druga zmienna.\n");
		SemGlobal->varlevel++;
		if(SemGlobal->roottype==TYPEARCH)
		{
			fprintf(dest, "NULL\n");
			if(SemGlobal->varlevel>1)
				fprintf(dest, "%d\n", SemGlobal->varlevel);
			else
				fprintf(dest, "NULL\n");
		}
		if(SemGlobal->roottype==TYPEPACK)
		{
			fprintf(dest, "%s\n", SemGlobal->rootsname);
			if(SemGlobal->varlevel>1)
				fprintf(dest, "%s\n", SemGlobal->rootvname);
			else
				fprintf(dest, "NULL\n");
		}
	}
	else
		return -1;
	fprintf(dest, "# 4 Wartosci semantyczne dla listy parametrow formalnych.\n");
	ParseParameterList(dest);
	fprintf(dest, "# 5 Wartosci semantyczne.\n");
	fclose(dest);

	/* sprawdzenie czy dla danej procedury zdefioniowano prototyp i oznaczenie
	 * tego prototypu jako posiadajacego odpowiednie cialo */
	actualproto.parcnt=0;
	strcpy(actualproto.name, procname);
	dest=fopen(valuef, "rt");
	isParams=0;
	while(fgets(bufor, BUFSIZE, dest))
	{
		if(bufor[0]=='#')
		{
			isParams=0;
			if(bufor[2]=='4')
				isParams=1;
			continue;
		}
		if(isParams==1)
		{
			sscanf(bufor, "%d\t%s\t%d", &lval, lname, &tvalue);
			if(actualproto.parcnt)
				actualproto.params=(int*)
					(int*)realloc(actualproto.params,
					(1+actualproto.parcnt)*sizeof(int));
			else
				actualproto.params=(int*)
					malloc(sizeof(int));
			actualproto.params[actualproto.parcnt]=tvalue;
			actualproto.parcnt++;
		}
	}
	fclose(dest);
	for(k=0 ; k<SemGlobal->protoscnt ; k++)
	{
		if(SemGlobal->fprotos[k].hasBody==0)
			if(!stricmp(SemGlobal->fprotos[k].name, actualproto.name))
				if(SemGlobal->fprotos[k].parcnt==actualproto.parcnt)
				{
					for(l=0 ; l<actualproto.parcnt ; l++)
					{
						if(SemGlobal->fprotos[k].params[l]!=actualproto.params[l])
							break;
					}
					if(l==actualproto.parcnt)
						SemGlobal->fprotos[k].hasBody=1;
					
				}
	}
	if(actualproto.parcnt)
		free(actualproto.params);

	
	
	sprintf(actpath, "%s\\%s", destpath, procname);
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, NULL);
		SemGlobal->hasTypes=hasTypes;
		SemGlobal->hasNonComputables=hasNonComputables;
		strcpy(SemGlobal->actpath, actpath);
		//pierwszy begin - otwarcie pliku z kopia biezacej procedury
		if(value==BEGINV)
		{
			ilebegin=1;
			sprintf(dpath, "%s\\%s", destpath, procname);
			SemGlobal->indexno++;
			fprintf(SemGlobal->indexf, "%d 4 %s\t%s\n", SemGlobal->indexno, procname, dpath);
			if(hasAliases)
			{
				if((dest=fopen(dpath, "at"))==NULL)
					return -1;
			}
			else
			{
				if((dest=fopen(dpath, "wt")))
					fprintf(dest, "11\tkeyword:\tbegin\n");
				else
					return -1;
			}
			break;
		}
		if(value==PROCV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d", destpath, procname, level);
			strcpy(SemGlobal->actroot, procname);
			if(proparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t0\t4\t0\t0\t0\t0\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName);
			fclose(dest);
			continue;
		}
		if(value==FUNCV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d", destpath, procname, level);
			strcpy(SemGlobal->actroot, procname);
			if(funcparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t%d\t2\t%s\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName, SemGlobal->LastReturnValue, SemGlobal->LastReturnString);
			fclose(dest);
			continue;
		}
		if(value==TYPEV)
		{
			yylval.semval=tok_TYPE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasTypes=1;
			fclose(dest);
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			continue;
		}
		if(value==SUBTYPEV)
		{
			ParseSubTypeDeclaration();
			hasTypes=SemGlobal->hasTypes;
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			continue;
		}
		if(value==CONSTANTV)
		{
			yylval.semval=tok_CONSTANT;
			dest=fopen(valuef, mode);
			ParseConstantDeclaration(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			//ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==SIGNALV)
		{
			yylval.semval=tok_SIGNAL;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==VARV)
		{
			yylval.semval=tok_VARIABLE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==SHAREDV)
		{
			yylex();
			yylval.semval=tok_VARIABLE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==ALIASV)
		{
			yylval.semval=tok_ALIAS;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			sprintf(dpath, "%s\\%s", destpath, procname);
			WriteAliasInfo(dpath, hasAliases);
			hasAliases=1;
			continue;
		}
		if(value==USEV)
		{
			if(UseParseInSubProgram(valuef)==-1)
				return -1;
			continue;
		}
		if(value==ASSERTV || value==ATTRIBV)
		{
			ReadToSemiC();
			continue;
		}
		sprintf(errmesg, ": %s", bufor);
		semerror(ERRUNSCONSTRUCT, errmesg);
	}
	
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		if(value==ENDV)
		{
			LastWasFor=0;
			if(!ileicf)
				ilebegin--;
			else
			{
				fprintf(dest, "%s", bufor);
				fgets(bufor, BUFSIZE, SemGlobal->srcf);
				value=getvalue(bufor, dest);
				if(value==IFV || value==CASEV || value==FORV || value==SEMICV || value==LOOPV || value==GENERATEV)
					ileicf--;
				fprintf(dest, "%s", bufor);
				continue;
			}
		}
		if(ilebegin==0)
			break;
		if(value==BEGINV)
		{
			LastWasFor=0;
			ilebegin++;
		}
		if(value==IFV || value==CASEV || value==FORV || value==LOOPV)
		{
			if(LastWasFor && value==LOOPV)
				ileicf--;
			LastWasFor=0;
			ileicf++;
		}
		if(value==FORV)
			LastWasFor=1;
		AliasCheckAndReplace(bufor);
		fprintf(dest, "%s", bufor);
	}
	fprintf(dest, "%s", bufor);
	CopyToSemicolon(dest);
	fclose(dest);
	if(hasTypes)
		enumsdef2enumsid(defpath, idpath);
	strcpy(SemGlobal->LastFoundName, procname);
	SemGlobal->LastFoundValue=namevalue;
	SemGlobal->varlevel--;
	return 0;
}//proparse



void AddResolutionFunction(int toknum)
{
	FILE *tfile;
	char path[PATHSIZE];
	sprintf(path, "%s.tdf", SemGlobal->actpath);
	if(SemGlobal->hasTypes)
		tfile=fopen(path, "at");
	else
		tfile=fopen(path, "wt");
	if(!tfile)
		return;
	fprintf(tfile, "%d\tresolvedbit\t0\t0\t0\t0\t0\t0\n", toknum);
	fclose(tfile);
}//AddResolutionFunction

void NewFunctionProto(void)
{
	if(SemGlobal->protoscnt)
		SemGlobal->fprotos=(FPROTO*)realloc((FPROTO*)SemGlobal->fprotos, (1+SemGlobal->protoscnt) * sizeof(FPROTO));
	else
		SemGlobal->fprotos=(FPROTO*)malloc(sizeof(FPROTO));
	SemGlobal->fprotos[SemGlobal->protoscnt].parcnt=0;
	SemGlobal->fprotos[SemGlobal->protoscnt].TypeOf=2;
	SemGlobal->fprotos[SemGlobal->protoscnt].hasBody=0;
	SemGlobal->fprotos[SemGlobal->protoscnt].ResolvingType=0;
}


int funcparse(char *destpath, int level)
{
	char bufor[BUFSIZE], c, dpath[PATHSIZE], valuef[PATHSIZE], actpath[PATHSIZE];
	char funcname[NAMESIZE], lname[NAMESIZE], rfuncname[NAMESIZE], mode[3]="at", errmesg[128];
	char ReturnTypeName[NAMESIZE], ProcessName[NAMESIZE], ReturnNameString[BUFSIZE], *token;
	char defpath[PATHSIZE], idpath[PATHSIZE];
	char FuncPathName[PATHSIZE], RFuncPathName[PATHSIZE];
	FILE *dest;
	int k, l, value, namevalue, ReturnValue, FileExist=0, isSpecialName=0, isIsRead, isSomethingRead=0, LastWasFor=0;
	unsigned i, j, hasTypes=0, hasNonComputables=0, hasAliases=0;
	unsigned ilebegin=0, ileproc=0;
	unsigned ileicf=0;//wspolny licznik dla if, case i for
	int tvalue, isParams, lval, IsOnlyPrototype=0, defaultmode;
	FPROTO actualproto;			//wartosci semantyczne dla ostatnio znalezionego prototypu
	defaultmode=SemGlobal->defaultmode;
	SemGlobal->defaultmode=1;/* default mode IN */
	mkdir(destpath);
	//na poczatku funkcji znajduje sie jej identyfikator, ktory bedzie nazwa pliku
	fgets(bufor, BUFSIZE, SemGlobal->srcf);
	c=bufor[0];
	i=0;
	while(c!=':')
	{
		i++;
		c=bufor[i];
	}
	i++;
	c=bufor[i];
	while(!(isalpha(c) || c=='"' || c=='\\'))
	{
		i++;
		c=bufor[i];
	}
	if(c=='"')
	{
		j=-1;
		isSpecialName=1;
		c=bufor[i+1];
	}
	else
	{
		funcname[0]=c;
		j=0;
	}
	while(c!='\n' && c!='"')
	{
		j++;
		i++;
		c=bufor[i];
		funcname[j]=c;
	}
	funcname[j]=0;
	namevalue=getvalue(bufor, NULL);
	TranslateIdent(funcname, FuncPathName);
	sprintf(valuef, "%s\\%s.val", destpath, FuncPathName);;
	if((dest=fopen(valuef, "rt")) || isSpecialName)
	{
		if(dest)
			fclose(dest);
		FileExist=1;
		SemGlobal->LastFunctionNo++;
		sprintf(rfuncname, "__%d__", SemGlobal->LastFunctionNo);
	}
	else
		strcpy(rfuncname, funcname);
	TranslateIdent(rfuncname, RFuncPathName);
	sprintf(valuef, "%s\\%s.val", destpath, RFuncPathName);;
	sprintf(defpath, "%s\\%s.def", destpath, RFuncPathName);
	sprintf(idpath, "%s\\%s.id", destpath, RFuncPathName);
	if((dest=fopen(valuef, "wt")))
	{
		fprintf(dest, "# 0 Czy jest dostepny pakiet IEEE.std_logic_1164. 1-jest, 0-nie jest\n");
		fprintf(dest, "%d\n", SemGlobal->isIEEE1164avail);
		fprintf(dest, "# 1 Wartosc rodzica aktualnej funkcji.\n");
		fprintf(dest, "%s\n", SemGlobal->actroot);
		fprintf(dest, "# 2 Dodatkowe nazwy dla sygnalow i zmiennych. Pierwsza linijka sygnaly, druga zmienna.\n");
		SemGlobal->varlevel++;
		if(SemGlobal->roottype==TYPEARCH)
		{
			fprintf(dest, "NULL\n");
			if(SemGlobal->varlevel>1)
				fprintf(dest, "%d\n", SemGlobal->varlevel);
			else
				fprintf(dest, "NULL\n");
		}
		if(SemGlobal->roottype==TYPEPACK)
		{
			fprintf(dest, "%s\n", SemGlobal->rootsname);
			if(SemGlobal->varlevel>1)
				fprintf(dest, "%s\n", SemGlobal->rootvname);
			else
				fprintf(dest, "NULL\n");
			if(SemGlobal->IsSTLACompiled)
			{
				switch(SemGlobal->IsSTLACompiled)
				{
				case 1:
					fprintf(dest, "# 3 Sciezki do pakietow uzywanych przez funkcje\n..\\syslib\\Ieee\\std_logic_arith\\std_logic_arith\n");
					break;
				case 2:
					fprintf(dest, "# 3 Sciezki do pakietow uzywanych przez funkcje\n..\\syslib\\Ieee\\std_logic_signed\\std_logic_signed\n");
					break;
				case 3:
					fprintf(dest, "# 3 Sciezki do pakietow uzywanych przez funkcje\n..\\syslib\\Ieee\\std_logic_unsigned\\std_logic_unsigned\n");
					break;
				case 4:
					fprintf(dest, "# 3 Sciezki do pakietow uzywanych przez funkcje\n..\\syslib\\Ieee\\std_logic_unsigned\\numeric_bit\n");
					break;
				}
			}
			else
			{
				fprintf(dest, "# 3 Sciezki do pakietow uzywanych przez funkcje\nlibrary\\%s\\%s\n", SemGlobal->rootsname, SemGlobal->rootsname);
			}
		}
	}
	else
		return -1;
	fprintf(dest, "# 4 Wartosci semantyczne dla listy parametrow formalnych.\n");
	ParseFunctionParameterList(dest);
	if(yylval.semval==tok_RETURN)
		yylex();
	//read type description
	fprintf(dest, "# 6 Wartosci semantyczne opisujace typ zwracany przez funkcje.\n");
	fgets(bufor, BUFSIZE, SemGlobal->srcf);
	ReturnValue=getvalue(bufor, NULL);
	if(ReturnValue==ISV)
	{
		isIsRead=1;
		strcpy(ReturnTypeName, yylval.tokstr);
		ReturnValue=yylval.toknum;
	}
	else
	{
		isIsRead=0;
		getname(bufor, ReturnTypeName);
	}
	if(ReturnValue==SEMICV)
	{
		ReturnValue=yylval.toknum;
		strcpy(ReturnTypeName, yylval.tokstr);
		IsOnlyPrototype=1;
	}
	SemGlobal->TypeCollected=2;
	if(FindTypeInformation(ReturnValue, ReturnTypeName, ReturnNameString)==0)
	{
		if(yylval.toknum<200)
			isSomethingRead=1;
	}
	SemGlobal->TypeCollected=0;
	fprintf(dest, "%d\t%s\t%s\n", ReturnValue, ReturnTypeName, ReturnNameString);
	fprintf(dest, "# 5 Wartosci semantyczne.\n");
	fclose(dest);
	if(IsOnlyPrototype)
	{
		strcpy(SemGlobal->LastReturnString, ReturnNameString);
		strcpy(SemGlobal->LastFoundName, funcname);
		SemGlobal->LastFoundValue=namevalue;
		SemGlobal->LastReturnValue=ReturnValue;
		SemGlobal->varlevel--;
		SemGlobal->defaultmode=3;/* default mode INOUT */

		sprintf(dpath, "%s\\%s", destpath, RFuncPathName);
		if((dest=fopen(dpath, "wt")))
		{
			fprintf(dest, "11\tkeyword:\tbegin\n");
			fprintf(dest, "%d\tkeyword:\tend\n", ENDV);
		}
		else
			return -1;
		SemGlobal->indexno++;
		//fprintf(SemGlobal->indexf, "%d 2 %s\t%s\n", SemGlobal->indexno, funcname, dpath);
		NewFunctionProto();

		dest=fopen(valuef, "rt");
		isParams=0;
		while(fgets(bufor, BUFSIZE, dest))
		{
			if(bufor[0]=='#')
			{
				isParams=0;
				if(bufor[2]=='4')
					isParams=1;
				if(bufor[2]=='6')
					isParams=2;
				continue;
			}
			if(isParams==1)
			{
				sscanf(bufor, "%d\t%s\t%d", &lval, lname, &tvalue);
				if(SemGlobal->fprotos[SemGlobal->protoscnt].parcnt)
					SemGlobal->fprotos[SemGlobal->protoscnt].params=(int*)
					(int*)realloc(SemGlobal->fprotos[SemGlobal->protoscnt].params,
					(1+SemGlobal->fprotos[SemGlobal->protoscnt].parcnt)*sizeof(int));
				else
					SemGlobal->fprotos[SemGlobal->protoscnt].params=(int*)
					malloc(sizeof(int));
				SemGlobal->fprotos[SemGlobal->protoscnt].params[SemGlobal->fprotos[SemGlobal->protoscnt].parcnt]=tvalue;
				SemGlobal->fprotos[SemGlobal->protoscnt].parcnt++;
			}
			if(isParams==2)
			{
				tvalue=getvalue(bufor, NULL);
				SemGlobal->fprotos[SemGlobal->protoscnt].retType=tvalue;
			}
		}
		fclose(dest);
		
		strcpy(SemGlobal->fprotos[SemGlobal->protoscnt].name, funcname);
		sprintf(SemGlobal->fprotos[SemGlobal->protoscnt].fname, "%s\\%s", destpath, rfuncname);
		SemGlobal->protoscnt++;
		SemGlobal->LastIsPrototype=1;
		return 0;
	}

	/* sprawdzenie czy dla danej funkcji zdefioniowano prototyp i oznaczenie
	 * tego prototypu jako posiadajacego odpowiednie cialo */
	actualproto.parcnt=0;
	strcpy(actualproto.name, funcname);
	dest=fopen(valuef, "rt");
	isParams=0;
	while(fgets(bufor, BUFSIZE, dest))
	{
		if(bufor[0]=='#')
		{
			isParams=0;
			if(bufor[2]=='4')
				isParams=1;
			if(bufor[2]=='6')
				isParams=2;
			continue;
		}
		if(isParams==1)
		{
			sscanf(bufor, "%d\t%s\t%d", &lval, lname, &tvalue);
			if(actualproto.parcnt)
				actualproto.params=(int*)
					(int*)realloc(actualproto.params,
					(1+actualproto.parcnt)*sizeof(int));
			else
				actualproto.params=(int*)
					malloc(sizeof(int));
			actualproto.params[actualproto.parcnt]=tvalue;
			actualproto.parcnt++;
		}
		if(isParams==2)
		{
			tvalue=getvalue(bufor, NULL);
			actualproto.retType=tvalue;
		}
	}
	fclose(dest);
	for(k=0 ; k<SemGlobal->protoscnt ; k++)
	{
		if(SemGlobal->fprotos[k].hasBody==0)
			if(!stricmp(SemGlobal->fprotos[k].name, actualproto.name))
				if(SemGlobal->fprotos[k].retType==actualproto.retType)
					if(SemGlobal->fprotos[k].parcnt==actualproto.parcnt)
					{
						for(l=0 ; l<actualproto.parcnt ; l++)
						{
							if(SemGlobal->fprotos[k].params[l]!=actualproto.params[l])
								break;
						}
						if(l==actualproto.parcnt)
						{
							SemGlobal->fprotos[k].hasBody=1;
						}
					}
	}
	if(actualproto.parcnt)
		free(actualproto.params);
	SemGlobal->defaultmode=3;/* default mode INOUT */
	sprintf(actpath, "%s\\%s", destpath, RFuncPathName);
	while(1)
	{
		if(isIsRead)
		{
			isIsRead=0;
			continue;
		}
		if(isSomethingRead)
		{
			value=yylval.toknum;
			isSomethingRead=0;
		}
		else
		{
			if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
				break;
			value=getValueForResolved(bufor, NULL);
			if(SemGlobal->LastResolvedType)
			{
				int i;
				for(i=0 ; i<SemGlobal->ResolvedCnt ; i++)
				{
					if(!stricmp(SemGlobal->ListOfResolved[i].FunName, actualproto.name))
						SemGlobal->ListOfResolved[i].ResMethod=SemGlobal->LastResolvedType;
				}
				SemGlobal->LastResolvedType=0;
			}
		}
		SemGlobal->hasTypes=hasTypes;
		SemGlobal->hasNonComputables=hasNonComputables;
		strcpy(SemGlobal->actpath, actpath);
		//pierwszy begin - otwarcie pliku z kopia biezacej funkcji
		if(value==BEGINV)
		{
			ilebegin=1;
			sprintf(dpath, "%s\\%s", destpath, RFuncPathName);
			SemGlobal->indexno++;
			if(SemGlobal->IsSTLACompiled)
			{
				char syslibpath[PATHSIZE];
				switch(SemGlobal->IsSTLACompiled)
				{
				case 1:
					sprintf(syslibpath, "..\\syslib\\Ieee\\std_logic_arith\\%s", RFuncPathName);
					break;
				case 2:
					sprintf(syslibpath, "..\\syslib\\Ieee\\std_logic_signed\\%s", RFuncPathName);
					break;
				case 3:
					sprintf(syslibpath, "..\\syslib\\Ieee\\std_logic_unsigned\\%s", RFuncPathName);
					break;
				case 4:
					sprintf(syslibpath, "..\\syslib\\Ieee\\numeric_bit\\%s", RFuncPathName);
					break;
				}
				fprintf(SemGlobal->indexf, "%d 2 %s\t%s\n", SemGlobal->indexno, funcname, syslibpath);
			}
			else
				fprintf(SemGlobal->indexf, "%d 2 %s\t%s\n", SemGlobal->indexno, funcname, dpath);
			if(hasAliases)
			{
				if((dest=fopen(dpath, "at"))==NULL)
					return -1;
			}
			else
			{
				if((dest=fopen(dpath, "wt")))
					fprintf(dest, "11\tkeyword:\tbegin\n");
				else
					return -1;
			}
			break;
		}
		if(value==PROCV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d", destpath, RFuncPathName, level);
			strcpy(SemGlobal->actroot, funcname);
			if(proparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t0\t4\t0\t0\t0\t0\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName);
			fclose(dest);
			continue;
		}
		if(value==FUNCV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d", destpath, RFuncPathName, level);
			strcpy(SemGlobal->actroot, funcname);
			if(funcparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t%d\t2\t%s\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName, SemGlobal->LastReturnValue, SemGlobal->LastReturnString);
			fclose(dest);
			continue;
		}
		if(value==TYPEV)
		{
			yylval.semval=tok_TYPE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasTypes=1;
			fclose(dest);
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			continue;
		}
		if(value==SUBTYPEV)
		{
			ParseSubTypeDeclaration();
			hasTypes=SemGlobal->hasTypes;
			hasNonComputables=SemGlobal->hasNonComputables;
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			continue;
		}
		if(value==CONSTANTV)
		{
			yylval.semval=tok_CONSTANT;
			dest=fopen(valuef, mode);
			ParseConstantDeclaration(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			//ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==SIGNALV)
		{
			yylval.semval=tok_SIGNAL;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			fclose(dest);
			continue;
		}
		if(value==VARV)
		{
			yylval.semval=tok_VARIABLE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			fclose(dest);
			continue;
		}
		if(value==SHAREDV)
		{
			yylex();
			yylval.semval=tok_VARIABLE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==ALIASV)
		{
			yylval.semval=tok_ALIAS;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			sprintf(dpath, "%s\\%s", destpath, RFuncPathName);
			WriteAliasInfo(dpath, hasAliases);
			hasAliases=1;
			continue;
		}
		if(value==USEV)
		{
			if(UseParseInSubProgram(valuef)==-1)
				return -1;
			continue;
		}
		if(value==ASSERTV || value==ATTRIBV)
		{
			ReadToSemiC();
			continue;
		}
		if(value==ISV)
			continue;
		sprintf(errmesg, ": %s", bufor);
		semerror(ERRUNSCONSTRUCT, errmesg);
	}
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		if(value==ENDV)
		{
			LastWasFor=0;
			if(!ileicf)
				ilebegin--;
			else
			{
				fprintf(dest, "%s", bufor);
				fgets(bufor, BUFSIZE, SemGlobal->srcf);
				value=getvalue(bufor, dest);
				if(value==IFV || value==CASEV || value==FORV || value==SEMICV || value==LOOPV || value==GENERATEV)
					ileicf--;
				fprintf(dest, "%s", bufor);
				continue;
			}
		}
		if(ilebegin==0)
			break;
		if(value==BEGINV)
		{
			LastWasFor=0;
			ilebegin++;
		}
		if(value==IFV || value==CASEV || value==FORV || value==LOOPV)
		{
			if(LastWasFor && value==LOOPV)
				ileicf--;
			LastWasFor=0;
			ileicf++;
		}




		if(value>=IDENTV)
		{
			char RealName[NAMESIZE];
			strcpy(ProcessName, bufor);//ewentualna nazwa procesu lub etykieta
			getname(ProcessName, RealName);
			if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
			{
				fprintf(dest, "%s", ProcessName);
				semerror(ERRUNEXEOF, NULL);
				break;
			}
			value=getvalue(bufor, dest);
			if(value==DOTV && !stricmp(RealName, "work"))
			{
				int isIdent=1;
				while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
				{
					value=getvalue(bufor, dest);
					if(value==DOTV || value>=IDENTV)
					{
						if(value==DOTV)
							isIdent=0;
						else
							isIdent=1;
					}
					else
					{
						int i;
						FILE *vfile;
						char TmpLexemStr[SBSIZE], RealFunctionName[SBSIZE], RealPath[PATHSIZE], PathPart[SBSIZE];
						//Dopisac wpisywanie leksemu tymczasowego
						sprintf(TmpLexemStr, "__pbtmpid__%d", SemGlobal->LastLexemId+200);
						WriteLastLexemId(TmpLexemStr);
						fprintf(dest, "%d\tidentifier:\t%s\n", SemGlobal->LastLexemId+199, TmpLexemStr);
						getname(SemGlobal->ListOfYY[SemGlobal->LexemsInListCnt-1].tokstr, RealFunctionName);
						RealPath[0]=0;
						getname(ProcessName, PathPart);
						if(!stricmp(PathPart, "work"))
							strcpy(RealPath, "library");
						else
							strcpy(RealPath, "..\\syslib");
						for(i=0 ; i<SemGlobal->LexemsInListCnt ; i++)
						{
							if(!(i%2) && i<SemGlobal->LexemsInListCnt-1)
							{
								getname(SemGlobal->ListOfYY[i].tokstr, PathPart);
								strcat(RealPath, "\\");
								strcat(RealPath, PathPart);
							}
						}
						if(value==LEFTBRV)
						{
							FreeListOfYY();
							fprintf(dest, "%s", bufor);
							vfile=fopen(valuef, "at");
							fprintf(vfile, "%d\t%s\t0\t21\t0\t0\t0\t0\t%s\n", SemGlobal->LastLexemId+199, RealFunctionName, RealPath);
							fclose(vfile);
						}
						else
						{
							char line[BUFSIZE];
							strcat(RealPath, "\\");
							strcat(RealPath, PathPart);
							strcat(RealPath, ".val");
							FindLexemInPackage(RealPath, line, SemGlobal->ListOfYY[i-1].tokstr);
							FreeListOfYY();
							if(*line)
							{
								char *ptr;
								ptr=line;
								while(*ptr!='\t')
									ptr++;
								ptr++;
								while(*ptr!='\t')
									ptr++;
								fprintf(dest, "%s", bufor);
								vfile=fopen(valuef, "at");
								fprintf(vfile, "%d\t%s%s", SemGlobal->LastLexemId+199, TmpLexemStr, ptr);
								fclose(vfile);
							}
							else
								semerror(ERRSEMVALNFIP, NULL);
						}
						break;
					}
					yylval.toknum=value;
					yylval.semval=value;
					strcpy(yylval.tokstr, bufor);
					NextYY();
				}
				continue;
			}
			if(value==COLONV)
			{
				if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
				{
					fprintf(dest, "%s", ProcessName);
					fprintf(dest, "113\tspecial_character:\t:\n");
					semerror(ERRUNEXEOF, NULL);
					break;
				}
				value=getvalue(bufor, dest);
				if(value==ASSERTV || value==ATTRIBV)
				{
					ReadToSemiC();
					if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
					{
						fprintf(dest, "%s", ProcessName);
						fprintf(dest, "113\tspecial_character:\t:\n");
						semerror(ERRUNEXEOF, NULL);
						break;
					}
					value=getvalue(bufor, dest);
				}
				else//prawdopodobnie nazwa componentu lub zwykla etykieta
				{
					char componname[NAMESIZE];
					fprintf(dest, "%s", ProcessName);
					fprintf(dest, "113\tspecial_character:\t:\n");
					value=getvalue(bufor, dest);
					if(value>200)
					{
						getname(bufor, componname);
						if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
						{
							fprintf(dest, "%d", value);
							fprintf(dest, "\tidentifier:\t:%s\n", componname);
							semerror(ERRUNEXEOF, NULL);
							break;
						}
						fprintf(dest, "%d", value);
						fprintf(dest, "\tidentifier:\t%s\n", componname);
						value=getvalue(bufor, dest);
						fprintf(dest, "%s", bufor);
						if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
						{
							semerror(ERRUNEXEOF, NULL);
							break;
						}
						fprintf(dest, "%s", bufor);
						value=getvalue(bufor, dest);
						CopyToSemicolon(dest);
						NewComponentName(componname);
						continue;
					}
					if(value==IFV || value==CASEV || value==LOOPV)
					{
						if(LastWasFor && value==LOOPV)
							ileicf--;
						LastWasFor=0;
						ileicf++;
					}
					if(value==FORV)
					{
						LastWasFor=1;
						ileicf++;
					}
				}
			}
			else
			{
				if(value==IFV || value==CASEV || value==FORV || value==LOOPV)
				{
					if(LastWasFor && value==LOOPV)
						ileicf--;
					LastWasFor=0;
					ileicf++;
				}
				AliasCheckAndReplace(ProcessName);
				fprintf(dest, "%s", ProcessName);
				fprintf(dest, "%s", bufor);
				continue;
			}
		}











		if(value==FORV)
			LastWasFor=1;
		fprintf(dest, "%s", bufor);
	}
	fprintf(dest, "%s", bufor);
	CopyToSemicolon(dest);
	fclose(dest);
	token=ReturnNameString;
	for(i=0 ; *token!='\t' ; i++)
		token=ReturnNameString+i;
	token=ReturnNameString+i;
	for( ; *token!='\t' ; i++)
		token=ReturnNameString+i;
	token=ReturnNameString+i;
	if(hasTypes)
		enumsdef2enumsid(defpath, idpath);
	strcpy(SemGlobal->LastReturnString, token);
	strcpy(SemGlobal->LastFoundName, funcname);
	SemGlobal->LastFoundValue=namevalue;
	SemGlobal->LastReturnValue=ReturnValue;
	SemGlobal->varlevel--;
	SemGlobal->defaultmode=defaultmode;/*return to previous value*/
	return 0;
}//funcparse



int FuncPrototypeParse(char *destpath, int level)
{
	char bufor[BUFSIZE], c, valuef[PATHSIZE];
	char funcname[NAMESIZE], rfuncname[NAMESIZE], mode[3]="at";
	char ReturnTypeName[NAMESIZE], lname[NAMESIZE], ReturnNameString[BUFSIZE], *token;
	FILE *dest;
	int namevalue, ReturnValue, FileExist=0, isSpecialName=0;
	unsigned i, j, hasTypes=0;
	unsigned ilebegin=0, ileproc=0;
	unsigned ileicf=0;//wspolny licznik dla if, case i for
	int tvalue, isParams, lval, defaultmode;
	defaultmode=SemGlobal->defaultmode;
	SemGlobal->defaultmode=1;/* default mode IN */
	mkdir(destpath);
	//na poczatku funkcji znajduje sie jej identyfikator, ktory bedzie nazwa pliku
	fgets(bufor, BUFSIZE, SemGlobal->srcf);
	c=bufor[0];
	i=0;
	while(c!=':')
	{
		i++;
		c=bufor[i];
	}
	i++;
	c=bufor[i];
	while(!(isalpha(c) || c=='"'))
	{
		i++;
		c=bufor[i];
	}
	if(c=='"')
	{
		j=-1;
		isSpecialName=1;
		c=bufor[i+1];
	}
	else
	{
		funcname[0]=c;
		j=0;
	}
	while(c!='\n' && c!='"')
	{
		j++;
		i++;
		c=bufor[i];
		funcname[j]=c;
	}
	funcname[j]=0;
	namevalue=getvalue(bufor, NULL);
	sprintf(valuef, "%s\\%s.val", destpath, funcname);;
	if((dest=fopen(valuef, "rt")) || isSpecialName)
	{
		if(dest)
			fclose(dest);
		FileExist=1;
		SemGlobal->LastFunctionNo++;
		sprintf(rfuncname, "__%d__", SemGlobal->LastFunctionNo);
	}
	else
		strcpy(rfuncname, funcname);
//	if(!stricmp(funcname, "resolvedbit"))
//		AddResolutionFunction(namevalue);
	sprintf(valuef, "%s\\%s.val", destpath, rfuncname);;
	if((dest=fopen(valuef, "wt")))
	{
		fprintf(dest, "# 0 Czy jest dostepny pakiet IEEE.std_logic_1164. 1-jest, 0-nie jest\n");
		fprintf(dest, "%d\n", SemGlobal->isIEEE1164avail);
		fprintf(dest, "# 1 Wartosc rodzica aktualnej funkcji.\n");
		fprintf(dest, "%s\n", SemGlobal->actroot);
		fprintf(dest, "# 2 Dodatkowe nazwy dla sygnalow i zmiennych. Pierwsza linijka sygnaly, druga zmienna.\n");
		SemGlobal->varlevel++;
		if(SemGlobal->roottype==TYPEARCH)
		{
			fprintf(dest, "NULL\n");
			if(SemGlobal->varlevel>1)
				fprintf(dest, "%d\n", SemGlobal->varlevel);
			else
				fprintf(dest, "NULL\n");
		}
		if(SemGlobal->roottype==TYPEPACK)
		{
			fprintf(dest, "%s\n", SemGlobal->rootsname);
			if(SemGlobal->varlevel>1)
				fprintf(dest, "%s\n", SemGlobal->rootvname);
			else
				fprintf(dest, "NULL\n");
			if(SemGlobal->IsSTLACompiled)
			{
				switch(SemGlobal->IsSTLACompiled)
				{
				case 1:
					fprintf(dest, "# 3 Sciezki do pakietow uzywanych przez funkcje\n..\\syslib\\Ieee\\std_logic_arith\\std_logic_arith\n");
					break;
				case 2:
					fprintf(dest, "# 3 Sciezki do pakietow uzywanych przez funkcje\n..\\syslib\\Ieee\\std_logic_signed\\std_logic_signed\n");
					break;
				case 3:
					fprintf(dest, "# 3 Sciezki do pakietow uzywanych przez funkcje\n..\\syslib\\Ieee\\std_logic_unsigned\\std_logic_unsigned\n");
					break;
				case 4:
					fprintf(dest, "# 3 Sciezki do pakietow uzywanych przez funkcje\n..\\syslib\\Ieee\\std_logic_unsigned\\numeric_bit\n");
					break;
				}
			}
			else
				fprintf(dest, "# 3 Sciezki do pakietow uzywanych przez funkcje\nlibrary\\%s\\%s\n", SemGlobal->rootsname, SemGlobal->rootsname);

		}
	}
	else
		return -1;
	fprintf(dest, "# 4 Wartosci semantyczne dla listy parametrow formalnych.\n");
	ParseFunctionParameterList(dest);
	//read type description
	fprintf(dest, "# 6 Wartosci semantyczne opisujace typ zwracany przez funkcje.\n");
	fgets(bufor, BUFSIZE, SemGlobal->srcf);
	ReturnValue=getvalue(bufor, NULL);
	getname(bufor, ReturnTypeName);
	SemGlobal->TypeCollected=2;
	FindTypeInformation(ReturnValue, ReturnTypeName, ReturnNameString);
	SemGlobal->TypeCollected=0;
	fprintf(dest, "%d\t%s\t%s\n", ReturnValue, ReturnTypeName, ReturnNameString);
	fclose(dest);
	/* Zapisanie wartosci semantycznych dotyczacych prototypu do tablicy
	 * w celu pozniejszego ich porownywania ze znalezionymi cialami funkcji
	 */
	NewFunctionProto();
	dest=fopen(valuef, "rt");
		isParams=0;
	while(fgets(bufor, BUFSIZE, dest))
	{
		if(bufor[0]=='#')
		{
			isParams=0;
			if(bufor[2]=='4')
				isParams=1;
			if(bufor[2]=='6')
				isParams=2;
			continue;
		}
		if(isParams==1)
		{
			sscanf(bufor, "%d\t%s\t%d", &lval, lname, &tvalue);
			if(SemGlobal->fprotos[SemGlobal->protoscnt].parcnt)
				SemGlobal->fprotos[SemGlobal->protoscnt].params=(int*)
					(int*)realloc(SemGlobal->fprotos[SemGlobal->protoscnt].params,
					(1+SemGlobal->fprotos[SemGlobal->protoscnt].parcnt)*sizeof(int));
			else
				SemGlobal->fprotos[SemGlobal->protoscnt].params=(int*)
					malloc(sizeof(int));
			SemGlobal->fprotos[SemGlobal->protoscnt].params[SemGlobal->fprotos[SemGlobal->protoscnt].parcnt]=tvalue;
			SemGlobal->fprotos[SemGlobal->protoscnt].parcnt++;
		}
		if(isParams==2)
		{
			tvalue=getvalue(bufor, NULL);
			SemGlobal->fprotos[SemGlobal->protoscnt].retType=tvalue;
		}
	}
	fclose(dest);
	if(yylval.toknum!=SEMICV)
		ReadToSemiC();
	token=ReturnNameString;
	for(i=0 ; *token!='\t' ; i++)
		token=ReturnNameString+i;
	token=ReturnNameString+i;
	for( ; *token!='\t' ; i++)
		token=ReturnNameString+i;
	token=ReturnNameString+i;
	strcpy(SemGlobal->LastReturnString, token);
	strcpy(SemGlobal->LastFoundName, funcname);
	strcpy(SemGlobal->fprotos[SemGlobal->protoscnt].name, funcname);
	sprintf(SemGlobal->fprotos[SemGlobal->protoscnt].fname, "%s\\%s", destpath, rfuncname);
	SemGlobal->protoscnt++;
	SemGlobal->LastFoundValue=namevalue;
	SemGlobal->LastReturnValue=ReturnValue;
	SemGlobal->defaultmode=defaultmode;/*return to previous value*/
	SemGlobal->varlevel--;
	return 0;
}//FuncPrototypeParse

int ProcedurePrototypeParse(char *destpath, int level)
{
	char bufor[BUFSIZE], c, valuef[PATHSIZE];
	char funcname[NAMESIZE], rfuncname[NAMESIZE], mode[3]="at";
	char lname[NAMESIZE];
	FILE *dest;
	int namevalue, FileExist=0, isSpecialName=0;
	unsigned i, j, hasTypes=0;
	unsigned ilebegin=0, ileproc=0;
	unsigned ileicf=0;//wspolny licznik dla if, case i for
	int tvalue, isParams, lval;
	mkdir(destpath);
	//na poczatku funkcji znajduje sie jej identyfikator, ktory bedzie nazwa pliku
	fgets(bufor, BUFSIZE, SemGlobal->srcf);
	c=bufor[0];
	i=0;
	while(c!=':')
	{
		i++;
		c=bufor[i];
	}
	i++;
	c=bufor[i];
	while(!(isalpha(c) || c=='"'))
	{
		i++;
		c=bufor[i];
	}
	if(c=='"')
	{
		j=-1;
		isSpecialName=1;
		c=bufor[i+1];
	}
	else
	{
		funcname[0]=c;
		j=0;
	}
	while(c!='\n' && c!='"')
	{
		j++;
		i++;
		c=bufor[i];
		funcname[j]=c;
	}
	funcname[j]=0;
	namevalue=getvalue(bufor, NULL);
	sprintf(valuef, "%s\\%s.val", destpath, funcname);;
	if((dest=fopen(valuef, "rt")) || isSpecialName)
	{
		if(dest)
			fclose(dest);
		FileExist=1;
		SemGlobal->LastFunctionNo++;
		sprintf(rfuncname, "__%d__", SemGlobal->LastFunctionNo);
	}
	else
		strcpy(rfuncname, funcname);
	sprintf(valuef, "%s\\%s.val", destpath, rfuncname);;
	if((dest=fopen(valuef, "wt")))
	{
		fprintf(dest, "# 0 Czy jest dostepny pakiet IEEE.std_logic_1164. 1-jest, 0-nie jest\n");
		fprintf(dest, "%d\n", SemGlobal->isIEEE1164avail);
		fprintf(dest, "# 1 Wartosc rodzica aktualnej funkcji.\n");
		fprintf(dest, "%s\n", SemGlobal->actroot);
		fprintf(dest, "# 2 Dodatkowe nazwy dla sygnalow i zmiennych. Pierwsza linijka sygnaly, druga zmienna.\n");
		SemGlobal->varlevel++;
		if(SemGlobal->roottype==TYPEARCH)
		{
			fprintf(dest, "NULL\n");
			if(SemGlobal->varlevel>1)
				fprintf(dest, "%d\n", SemGlobal->varlevel);
			else
				fprintf(dest, "NULL\n");
		}
		if(SemGlobal->roottype==TYPEPACK)
		{
			fprintf(dest, "%s\n", SemGlobal->rootsname);
			if(SemGlobal->varlevel>1)
				fprintf(dest, "%s\n", SemGlobal->rootvname);
			else
				fprintf(dest, "NULL\n");
		}
	}
	else
		return -1;
	fprintf(dest, "# 4 Wartosci semantyczne dla listy parametrow formalnych.\n");
	ParseProcedureParameterList(dest);
	//read type description
	fprintf(dest, "# 6 Wartosci semantyczne opisujace typ zwracany przez funkcje.\n");
	fclose(dest);
	/* Zapisanie wartosci semantycznych dotyczacych prototypu do tablicy
	 * w celu pozniejszego ich porownywania ze znalezionymi cialami funkcji
	 */
	if(SemGlobal->protoscnt)
		SemGlobal->fprotos=(FPROTO*)realloc((FPROTO*)SemGlobal->fprotos, (1+SemGlobal->protoscnt) * sizeof(FPROTO));
	else
		SemGlobal->fprotos=(FPROTO*)malloc(sizeof(FPROTO));
	SemGlobal->fprotos[SemGlobal->protoscnt].parcnt=0;
	SemGlobal->fprotos[SemGlobal->protoscnt].TypeOf=4;
	SemGlobal->fprotos[SemGlobal->protoscnt].hasBody=0;
	dest=fopen(valuef, "rt");
	isParams=0;
	while(fgets(bufor, BUFSIZE, dest))
	{
		if(bufor[0]=='#')
		{
			isParams=0;
			if(bufor[2]=='4')
				isParams=1;
			if(bufor[2]=='6')
				isParams=2;
			continue;
		}
		if(isParams==1)
		{
			sscanf(bufor, "%d\t%s\t%d", &lval, lname, &tvalue);
			if(SemGlobal->fprotos[SemGlobal->protoscnt].parcnt)
				SemGlobal->fprotos[SemGlobal->protoscnt].params=(int*)
					realloc((int*)SemGlobal->fprotos[SemGlobal->protoscnt].params,
					(1+SemGlobal->fprotos[SemGlobal->protoscnt].parcnt)*sizeof(int));
			else
				SemGlobal->fprotos[SemGlobal->protoscnt].params=(int*)
					malloc(sizeof(int));
			SemGlobal->fprotos[SemGlobal->protoscnt].params[SemGlobal->fprotos[SemGlobal->protoscnt].parcnt]=tvalue;
			SemGlobal->fprotos[SemGlobal->protoscnt].parcnt++;
		}
		if(isParams==2)
		{
			tvalue=getvalue(bufor, NULL);
			SemGlobal->fprotos[SemGlobal->protoscnt].retType=tvalue;
		}
	}
	fclose(dest);
	if(yylval.toknum!=SEMICV)
		ReadToSemiC();
	strcpy(SemGlobal->LastFoundName, funcname);
	strcpy(SemGlobal->fprotos[SemGlobal->protoscnt].name, funcname);
	sprintf(SemGlobal->fprotos[SemGlobal->protoscnt].fname, "%s\\%s", destpath, rfuncname);
	SemGlobal->protoscnt++;
	SemGlobal->LastFoundValue=namevalue;
	SemGlobal->varlevel--;
	return 0;
}//ProcedurePrototypeParse


int NoNamedProcessParse(char *destpath, int level)
{
	static unsigned long processcnt=0;
	char prname[32];
	sprintf(prname, "__prid__%ld", processcnt);
	processcnt++;
	return ProcessParse(destpath, level, prname);
}//NoNamedProcessParse

int NoNamedGenerateParse(char *destpath, int level, char *bufor)
{
	static unsigned long generatecnt=0;
	char generatename[32];
	sprintf(generatename, "__gnid__%ld", generatecnt);
	generatecnt++;
	return GenerateParse(destpath, level, generatename, bufor);
}//NoNamedGenerateParse


int CheckAndAddProcessName(char *prname)
{
	return 0;
}

int ProcessParse(char *destpath, int level, char *prname)
{
	char bufor[BUFSIZE], dpath[PATHSIZE], valuef[PATHSIZE], actpath[PATHSIZE];
	char funcname[NAMESIZE], mode[3]="at", errmesg[128];
	char defpath[PATHSIZE], idpath[PATHSIZE], ProcessName[BUFSIZE];
	FILE *dest;
	int value, hasTypes=0, hasNonComputables=0, prnumber, hasAliases=0, LastWasFor=0;
	unsigned ilebegin=0, ileproc=0;
	unsigned ileicf=0;//wspolny licznik dla if, case i for
	if(*prname=='\\')
	{
		NoNamedProcessParse(destpath, level);
		return 0;
	}
	mkdir(destpath);
	strcpy(funcname, prname);
	prnumber=SemGlobal->indexno;
	sprintf(valuef, "%s\\%s.val", destpath, funcname);;
	sprintf(defpath, "%s\\%s.def", destpath, funcname);
	sprintf(idpath, "%s\\%s.id", destpath, funcname);
	if((dest=fopen(valuef, "wt")))
	{
		fprintf(dest, "# 0 Czy jest dostepny pakiet IEEE.std_logic_1164. 1-jest, 0-nie jest\n");
		fprintf(dest, "%d\n", SemGlobal->isIEEE1164avail);
		fprintf(dest, "# 1 Wartosc rodzica aktualnej funkcji.\n");
		fprintf(dest, "%s\n", SemGlobal->actroot);
		fprintf(dest, "# 2 Dodatkowe nazwy dla sygnalow i zmiennych. Pierwsza linijka sygnaly, druga zmienna.\n");
		SemGlobal->varlevel++;
		if(SemGlobal->roottype==TYPEARCH)
		{
			fprintf(dest, "NULL\n");
			if(SemGlobal->varlevel>1)
				fprintf(dest, "%d\n", SemGlobal->varlevel);
			else
				fprintf(dest, "NULL\n");
		}
		if(SemGlobal->roottype==TYPEPACK)
		{
			fprintf(dest, "%s\n", SemGlobal->rootsname);
			if(SemGlobal->varlevel>1)
				fprintf(dest, "%s\n", SemGlobal->rootvname);
			else
				fprintf(dest, "NULL\n");
		}
	}
	else
		return -1;
	fprintf(dest, "# 5 Wartosci semantyczne.\n");
	fclose(dest);
	sprintf(actpath, "%s\\%s", destpath, funcname);
	strcpy(SemGlobal->actProcessContext, funcname);
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, NULL);
		strcpy(SemGlobal->actpath, actpath);
		SemGlobal->hasTypes=hasTypes;
		SemGlobal->hasNonComputables=hasNonComputables;
		//obsluga sensitivity list
		if(ilebegin==0)
		{
			if(value==LEFTBRV)
			{
				int ParCnt=1;
				while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
				{
					value=getvalue(bufor, NULL);
					if(value==RIGHTBRV)
					{
						if(ParCnt==1)
							break;
						else
							ParCnt--;
						//AliasCheckAndReplace(name);
					}
					if(value==LEFTBRV)
						ParCnt++;
				}
				continue;
			}
			if(value==ISV)
				continue;
		}
		//pierwszy begin - otwarcie pliku z kopia biezacego procesu
		if(value==BEGINV)
		{
			ilebegin=1;
			sprintf(dpath, "%s\\%s", destpath, funcname);
			SemGlobal->indexno++;
			fprintf(SemGlobal->indexf, "%d 3 %s\t%s\n", SemGlobal->indexno, funcname, dpath);
			if(hasAliases)
			{
				if((dest=fopen(dpath, "at"))==NULL)
					return -1;
			}
			else
			{
				if((dest=fopen(dpath, "wt")))
					fprintf(dest, "11\tkeyword:\tbegin\n");
				else
					return -1;
			}
			break;
		}
		if(value==PROCV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d", destpath, funcname, level);
			strcpy(SemGlobal->actroot, funcname);
			if(proparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t0\t4\t0\t0\t0\t0\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName);
			fclose(dest);
			continue;
		}
		if(value==FUNCV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d", destpath, funcname, level);
			strcpy(SemGlobal->actroot, funcname);
			if(funcparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t%d\t2\t%s\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName, SemGlobal->LastReturnValue, SemGlobal->LastReturnString);
			fclose(dest);
			continue;
		}
		if(value==TYPEV)
		{
			yylval.semval=tok_TYPE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasTypes=1;
			fclose(dest);
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			continue;
		}
		if(value==SUBTYPEV)
		{
			ParseSubTypeDeclaration();
			hasTypes=SemGlobal->hasTypes;
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			continue;
		}
		if(value==CONSTANTV)
		{
			yylval.semval=tok_CONSTANT;
			dest=fopen(valuef, mode);
			ParseConstantDeclaration(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			//ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==SIGNALV)
		{
			yylval.semval=tok_SIGNAL;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			fclose(dest);
			continue;
		}
		if(value==VARV)
		{
			yylval.semval=tok_VARIABLE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			fclose(dest);
			continue;
		}
		if(value==SHAREDV)
		{
			yylex();
			yylval.semval=tok_VARIABLE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			fclose(dest);
			continue;
		}
		if(value==ALIASV)
		{
			yylval.semval=tok_ALIAS;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			sprintf(dpath, "%s\\%s", destpath, funcname);
			WriteAliasInfo(dpath, hasAliases);
			hasAliases=1;
			continue;
		}
		if(value==COMPONV)
		{
			if(componparse(dpath, 1)==-1)
				return -1;
			continue;
		}
		if(value==ASSERTV || value==ATTRIBV)
		{
			ReadToSemiC();
			continue;
		}
		sprintf(errmesg, ": %s", bufor);
		semerror(ERRUNSCONSTRUCT, errmesg);
	}
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		if(value==ENDV)
		{
			LastWasFor=0;
			if(!ileicf)
				ilebegin--;
			else
			{
				fprintf(dest, "%s", bufor);
				fgets(bufor, BUFSIZE, SemGlobal->srcf);
				value=getvalue(bufor, dest);
				if(value==IFV || value==CASEV || value==FORV || value==LOOPV || value==SEMICV || value==GENERATEV)
					ileicf--;
				fprintf(dest, "%s", bufor);
				continue;
			}
		}
		if(ilebegin==0)
			break;
		if(value==BEGINV)
		{
			LastWasFor=0;
			ilebegin++;
		}
		if(value==IFV || value==CASEV || value==FORV || value==WHILEV || value==LOOPV)
		{
			if(LastWasFor && value==LOOPV)
				ileicf--;
			LastWasFor=0;
			ileicf++;
		}
		if(value==FORV || value==WHILEV)
			LastWasFor=1;
		if(value==ASSERTV || value==ATTRIBV)
		{
			ReadToSemiC();
			continue;
		}

		
		

		if(value>=IDENTV)
		{
			char RealName[BUFSIZE];
			strcpy(ProcessName, bufor);//ewentualna nazwa procesu lub etykieta
			getname(ProcessName, RealName);
			if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
			{
				fprintf(dest, "%s", ProcessName);
				semerror(ERRUNEXEOF, NULL);
				break;
			}
			value=getvalue(bufor, dest);
			if(value==DOTV && !stricmp(RealName, "work"))
			{
				int isIdent=1;
				while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
				{
					value=getvalue(bufor, dest);
					if(value==DOTV || value>=IDENTV)
					{
						if(value==DOTV)
							isIdent=0;
						else
							isIdent=1;
					}
					else
					{
						int i;
						FILE *vfile;
						char TmpLexemStr[SBSIZE], RealFunctionName[SBSIZE], RealPath[PATHSIZE], PathPart[SBSIZE];
						//Dopisac wpisywanie leksemu tymczasowego
						sprintf(TmpLexemStr, "__pbtmpid__%d", SemGlobal->LastLexemId+200);
						WriteLastLexemId(TmpLexemStr);
						fprintf(dest, "%d\tidentifier:\t%s\n", SemGlobal->LastLexemId+199, TmpLexemStr);
						getname(SemGlobal->ListOfYY[SemGlobal->LexemsInListCnt-1].tokstr, RealFunctionName);
						RealPath[0]=0;
						getname(ProcessName, PathPart);
						if(!stricmp(PathPart, "work"))
							strcpy(RealPath, "library");
						else
							strcpy(RealPath, "..\\syslib");
						for(i=0 ; i<SemGlobal->LexemsInListCnt ; i++)
						{
							if(!(i%2) && i<SemGlobal->LexemsInListCnt-1)
							{
								getname(SemGlobal->ListOfYY[i].tokstr, PathPart);
								strcat(RealPath, "\\");
								strcat(RealPath, PathPart);
							}
						}
						if(value==LEFTBRV)
						{
							FreeListOfYY();
							fprintf(dest, "%s", bufor);
							vfile=fopen(valuef, "at");
							fprintf(vfile, "%d\t%s\t0\t21\t0\t0\t0\t0\t%s\n", SemGlobal->LastLexemId+199, RealFunctionName, RealPath);
							fclose(vfile);
						}
						else
						{
							char line[BUFSIZE];
							strcat(RealPath, "\\");
							strcat(RealPath, PathPart);
							strcat(RealPath, ".val");
							FindLexemInPackage(RealPath, line, SemGlobal->ListOfYY[i-1].tokstr);
							FreeListOfYY();
							if(*line)
							{
								char *ptr;
								ptr=line;
								while(*ptr!='\t')
									ptr++;
								ptr++;
								while(*ptr!='\t')
									ptr++;
								fprintf(dest, "%s", bufor);
								vfile=fopen(valuef, "at");
								fprintf(vfile, "%d\t%s%s", SemGlobal->LastLexemId+199, TmpLexemStr, ptr);
								fclose(vfile);
							}
							else
								semerror(ERRSEMVALNFIP, NULL);
						}
						break;
					}
					yylval.toknum=value;
					yylval.semval=value;
					strcpy(yylval.tokstr, bufor);
					NextYY();
				}
				continue;
			}
			if(value==COLONV)
			{
				if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
				{
					fprintf(dest, "%s", ProcessName);
					fprintf(dest, "113\tspecial_character:\t:\n");
					semerror(ERRUNEXEOF, NULL);
					break;
				}
				value=getvalue(bufor, dest);
				if(value==ASSERTV || value==ATTRIBV)
				{
					ReadToSemiC();
					continue;
				}
				else//prawdopodobnie nazwa componentu lub zwykla etykieta
				{
					char componname[NAMESIZE];
					fprintf(dest, "%s", ProcessName);
					fprintf(dest, "113\tspecial_character:\t:\n");
					value=getvalue(bufor, dest);
					if(value>200)
					{
						getname(bufor, componname);
						if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
						{
							fprintf(dest, "%d", value);
							fprintf(dest, "\tidentifier:\t:%s\n", componname);
							semerror(ERRUNEXEOF, NULL);
							break;
						}
						fprintf(dest, "%d", value);
						fprintf(dest, "\tidentifier:\t%s\n", componname);
						value=getvalue(bufor, dest);
						fprintf(dest, "%s", bufor);
//						if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
//						{
//							semerror(ERRUNEXEOF, NULL);
//							break;
//						}
//						value=getvalue(bufor, dest);
						CopyToSemicolon(dest);
						NewComponentName(componname);
						continue;
					}
					if(value==IFV || value==CASEV || value==LOOPV)
					{
						if(LastWasFor && value==LOOPV)
							ileicf--;
						LastWasFor=0;
						ileicf++;
					}
					if(value==FORV || value==WHILEV)
					{
						LastWasFor=1;
						ileicf++;
					}
				}
			}
			else
			{
				AliasCheckAndReplace(ProcessName);
				fprintf(dest, "%s", ProcessName);
				fprintf(dest, "%s", bufor);
				continue;
			}
		}


		
		
		
		
		AliasCheckAndReplace(bufor);
		fprintf(dest, "%s", bufor);
	}
	fprintf(dest, "%s", bufor);
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		if(value==SEMICV)
		{
			fprintf(dest, "%s", bufor);
			break;
		}
		fprintf(dest, "%s", bufor);
	}
	fclose(dest);
  if(hasTypes)
    enumsdef2enumsid(defpath, idpath);

	strcpy(SemGlobal->LastFoundName, prname);
	SemGlobal->actProcessContext[0]=0;
	SemGlobal->LastFoundValue=prnumber;
	//SemGlobal->LastFunctionNo=prnumber;
	SemGlobal->varlevel--;
	return 0;
}//ProcessParse


int FindGuardInVar()
{
//	FILE *alufile;
//	char bufor[BUFSIZE];
	char name[NAMESIZE];
//	int LexemNo;
//	if((alufile=fopen("alu.var", "rt"))!=NULL)
//	{
//		while(fgets(bufor, BUFSIZE, alufile))
//		{
//			sscanf(bufor, "%d\t%s", &LexemNo, name);
//			if(!stricmp(name, "guard"))
//			{
//				fclose(alufile);
//				return LexemNo+200; 
//			}
//		}
//		fclose(alufile);
//	}
	SemGlobal->GuardCnt++;
	sprintf(name, "Guard%d", SemGlobal->GuardCnt);
	WriteLastLexemId(name);
	return SemGlobal->LastLexemId+199;
}

int BlockParse(char *destpath, int level, char *prname)
{
	char bufor[BUFSIZE], dpath[PATHSIZE], valuef[PATHSIZE], actpath[PATHSIZE];
	char funcname[NAMESIZE], ProcessName[BUFSIZE], mode[3]="at", errmesg[128];
	char defpath[PATHSIZE], idpath[PATHSIZE], componname[NAMESIZE];
	int isGuarded=0, GuardLexem;
	FILE *dest;
	int value, hasTypes=0, hasNonComputables=0, prnumber, hasAliases=0;
	unsigned ilebegin=0, ileproc=0;
	unsigned ileicf=0;//wspolny licznik dla if, case i for
	mkdir(destpath);
	strcpy(funcname, prname);
	prnumber=SemGlobal->indexno;
	sprintf(valuef, "%s\\%s.val", destpath, funcname);;
	sprintf(defpath, "%s\\%s.def", destpath, funcname);
	sprintf(idpath, "%s\\%s.id", destpath, funcname);
	if((dest=fopen(valuef, "wt")))
	{
		fprintf(dest, "# 0 Czy jest dostepny pakiet IEEE.std_logic_1164. 1-jest, 0-nie jest\n");
		fprintf(dest, "%d\n", SemGlobal->isIEEE1164avail);
		fprintf(dest, "# 1 Wartosc rodzica aktualnej funkcji.\n");
		fprintf(dest, "%s\n", SemGlobal->actroot);
		fprintf(dest, "# 2 Dodatkowe nazwy dla sygnalow i zmiennych. Pierwsza linijka sygnaly, druga zmienna.\n");
		SemGlobal->varlevel++;
		if(SemGlobal->roottype==TYPEARCH)
		{
			fprintf(dest, "NULL\n");
			if(SemGlobal->varlevel>1)
				fprintf(dest, "%d\n", SemGlobal->varlevel);
			else
				fprintf(dest, "NULL\n");
		}
		if(SemGlobal->roottype==TYPEPACK)
		{
			fprintf(dest, "%s\n", SemGlobal->rootsname);
			if(SemGlobal->varlevel>1)
				fprintf(dest, "%s\n", SemGlobal->rootvname);
			else
				fprintf(dest, "NULL\n");
		}
	}
	else
		return -1;
	fprintf(dest, "# 5 Wartosci semantyczne.\n");
	fclose(dest);
	sprintf(actpath, "%s\\%s", destpath, funcname);
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, NULL);
		strcpy(SemGlobal->actpath, actpath);
		SemGlobal->hasTypes=hasTypes;
		SemGlobal->hasNonComputables=hasNonComputables;
		//obsluga sensitivity list
		if(ilebegin==0)
		{
			if(value==LEFTBRV)
			{
				SemGlobal->ListOfYY=NULL;
				while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
				{
					value=getvalue(bufor, NULL);
					if(value==RIGHTBRV)
					{
						isGuarded=1;
						if((dest=fopen(valuef, "at")))
						{
							GuardLexem=FindGuardInVar();
							fprintf(dest,"%d\tGuard%d\t204\t1\t3\t1\t0\t1\t0\n", GuardLexem, SemGlobal->GuardCnt);
							fclose(dest);
						}
						break;
					}
					yylval.toknum=value;
					yylval.semval=0;
					strcpy(yylval.tokstr, bufor);
					NextYY();
				}
				continue;
			}
			if(value==ISV)
				continue;
		}
		//pierwszy begin - otwarcie pliku z kopia biezacego procesu
		if(value==BEGINV)
		{
			ilebegin=1;
			sprintf(dpath, "%s\\%s", destpath, funcname);
			SemGlobal->indexno++;
			fprintf(SemGlobal->indexf, "%d 11 %s\t%s\n", SemGlobal->indexno, funcname, dpath);
			if(hasAliases)
			{
				if((dest=fopen(dpath, "at"))==NULL)
					return -1;
			}
			else
			{
				if((dest=fopen(dpath, "wt")))
					fprintf(dest, "%s", bufor);
				else
					return -1;
			}
			if(isGuarded)
			{
				int i;
				fprintf(dest, "%d\tidentifier:\tGuard%d\n", GuardLexem, SemGlobal->GuardCnt);
				fprintf(dest, "171\tdouble_char_symbol:\t<=\n");
				for(i=0 ; i<SemGlobal->LexemsInListCnt ; i++)
					fprintf(dest, "%s", SemGlobal->ListOfYY[i].tokstr);
				fprintf(dest, "114\tspecial_character:\t;\n");
				FreeListOfYY();
			}
			break;
		}
		if(value==PROCV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d", destpath, funcname, level);
			strcpy(SemGlobal->actroot, funcname);
			if(proparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t0\t4\t0\t0\t0\t0\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName);
			fclose(dest);
			continue;
		}
		if(value==FUNCV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d", destpath, funcname, level);
			strcpy(SemGlobal->actroot, funcname);
			if(funcparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t%d\t2\t%s\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName, SemGlobal->LastReturnValue, SemGlobal->LastReturnString);
			fclose(dest);
			continue;
		}
		if(value==TYPEV)
		{
			yylval.semval=tok_TYPE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			hasTypes=1;
			fclose(dest);
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			continue;
		}
		if(value==SUBTYPEV)
		{
			ParseSubTypeDeclaration();
			hasTypes=SemGlobal->hasTypes;
			if(hasTypes)
				enumsdef2enumsid(defpath, idpath);
			continue;
		}
		if(value==CONSTANTV)
		{
			yylval.semval=tok_CONSTANT;
			dest=fopen(valuef, mode);
			ParseConstantDeclaration(dest);
			hasNonComputables=SemGlobal->hasNonComputables;
			//ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==SIGNALV)
		{
			yylval.semval=tok_SIGNAL;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==VARV)
		{
			yylval.semval=tok_VARIABLE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==SHAREDV)
		{
			yylex();
			yylval.semval=tok_VARIABLE;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			continue;
		}
		if(value==ALIASV)
		{
			yylval.semval=tok_ALIAS;
			dest=fopen(valuef, mode);
			ParseDeclarativePart(dest);
			fclose(dest);
			sprintf(dpath, "%s\\%s", destpath, funcname);
			WriteAliasInfo(dpath, hasAliases);
			hasAliases=1;
			continue;
		}
		if(value==COMPONV)
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d", destpath, funcname, level);
			strcpy(SemGlobal->actroot, funcname);
			if(componparse(dpath, level)==-1)
				return -1;
			dest=fopen(valuef, mode);
			fprintf(dest, "%d\t%s\t0\t8\t0\t0\t0\t0\t0\n", SemGlobal->LastFoundValue, SemGlobal->LastFoundName, SemGlobal->LastFoundValue, SemGlobal->LastFoundName);
			fclose(dest);
			level--;
			continue;
		}
		if(value==USEV)
		{
			if(UseParseInSubProgram(valuef)==-1)
				return -1;
			continue;
		}
		if(value==ASSERTV || value==ATTRIBV)
		{
			ReadToSemiC();
			continue;
		}
		sprintf(errmesg, ": %s", bufor);
		semerror(ERRUNSCONSTRUCT, errmesg);
	}
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		int NameValue;
		value=getvalue(bufor, dest);
		NameValue=value;
		if(value>=IDENTV)
		{
			char RealName[NAMESIZE];
			strcpy(ProcessName, bufor);//ewentualna nazwa procesu lub etykieta
			getname(ProcessName, RealName);
			if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
			{
				fprintf(dest, "%s", ProcessName);
				semerror(ERRUNEXEOF, NULL);
				break;
			}
			value=getvalue(bufor, dest);
			if(value==DOTV && !stricmp(RealName, "work"))
			{
				int isIdent=1;
				while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
				{
					value=getvalue(bufor, dest);
					if(value==DOTV || value>=IDENTV)
					{
						if(value==DOTV)
							isIdent=0;
						else
							isIdent=1;
					}
					else
					{
						int i;
						if(value==LEFTBRV)
						{
							FILE *vfile;
							char TmpLexemStr[SBSIZE], RealFunctionName[SBSIZE], RealPath[PATHSIZE], PathPart[SBSIZE];
							//Dopisac wpisywanie leksemu tymczasowego
							sprintf(TmpLexemStr, "__pbtmpid__%d", SemGlobal->LastLexemId+200);
							WriteLastLexemId(TmpLexemStr);
							fprintf(dest, "%d\tidentifier:\t%s\n", SemGlobal->LastLexemId+199, TmpLexemStr);
							getname(SemGlobal->ListOfYY[SemGlobal->LexemsInListCnt-1].tokstr, RealFunctionName);
							RealPath[0]=0;
							getname(ProcessName, PathPart);
							if(!stricmp(PathPart, "work"))
								strcpy(RealPath, "library");
							else
								strcpy(RealPath, "..\\syslib");
							for(i=0 ; i<SemGlobal->LexemsInListCnt ; i++)
							{
								if(!(i%2) && i<SemGlobal->LexemsInListCnt-1)
								{
									getname(SemGlobal->ListOfYY[i].tokstr, PathPart);
									strcat(RealPath, "\\");
									strcat(RealPath, PathPart);
								}
							}
							FreeListOfYY();
							fprintf(dest, "%s", bufor);
							vfile=fopen(valuef, "at");
							fprintf(vfile, "%d\t%s\t0\t21\t0\t0\t0\t0\t%s\n", SemGlobal->LastLexemId+199, RealFunctionName, RealPath);
							fclose(vfile);
							break;
						}
						else
						{
							fprintf(dest, "%s", ProcessName);
							fprintf(dest, "111\tspecial_character:\t.\n");
							for(i=0 ; i<SemGlobal->LexemsInListCnt ; i++)
							{
								fprintf(dest, "%s", SemGlobal->ListOfYY[i].tokstr);
							}
							FreeListOfYY();
							break;
						}
					}
					yylval.toknum=value;
					yylval.semval=value;
					strcpy(yylval.tokstr, bufor);
					NextYY();
				}
				continue;
			}
			if(value==COLONV)
			{
				if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
				{
					fprintf(dest, "%s", ProcessName);
					fprintf(dest, "113\tspecial_character:\t:\n");
					semerror(ERRUNEXEOF, NULL);
					break;
				}
				value=getvalue(bufor, dest);
				if(value==ASSERTV || value==ATTRIBV)
				{
					ReadToSemiC();
					if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
					{
						fprintf(dest, "%s", ProcessName);
						fprintf(dest, "113\tspecial_character:\t:\n");
						semerror(ERRUNEXEOF, NULL);
						break;
					}
					value=getvalue(bufor, dest);
				}
				if(value==BLOCKV)//blok z nazwa (nazwa bloku musi wystapic)
				{
					char blname[NAMESIZE];
					int ModifiedValue;
					getname(ProcessName, blname);
					ModifiedValue=AddToBlockList(blname);
					if(ModifiedValue)
						NameValue=ModifiedValue;
					if(ileproc==0)
						ileproc=1;
					level++;
					sprintf(dpath, "%s\\%s_%d", destpath, funcname, level);
					strcpy(SemGlobal->actroot, funcname);
					if(BlockParse(dpath, level, blname)==-1)
						return -1;
					if(*blname!='\\')
					{
						fprintf(dest,
							"%d\tkeyword:\tblock\n%d\tidentifier:\t%s\n",
							BLOCKV, NameValue, blname);
					}
					else
					{
						fprintf(dest,
							"%d\tkeyword:\tblock\n%d\tidentifier:\t%s\n",
							BLOCKV, SemGlobal->LastLexemId+200, SemGlobal->LastFoundName);
						WriteLastLexemId(SemGlobal->LastFoundName);

					}
					level--;
					continue;				   
				}
				if(value==FORV || value==IFV)//generate z etykieta
				{
					if(ileproc==0)
						ileproc=1;
					level++;
					sprintf(dpath, "%s\\%s_%d", destpath, funcname, level);
					//sprintf(dpath, "%s\\%s_%d\\%s_%d", destpath, entityname, level-1, archname, level);
					strcpy(SemGlobal->actroot, funcname);
					if(NoNamedGenerateParse(dpath, level, bufor)==-1)
						return -1;
					fprintf(dest,
						"%d\tkeyword:\tgenerate\n%d\tidentifier:\t%s\n",
						GENERATEV, SemGlobal->LastLexemId+200, SemGlobal->LastFoundName);
					WriteLastLexemId(SemGlobal->LastFoundName);
					level--;
					continue;
				}
				if(value==PROCESSV)//proces z nazwa
				{
					char prname[NAMESIZE];
					int ModifiedValue;
					getname(ProcessName, prname);
					ModifiedValue=AddToProcessList(prname);
					if(ModifiedValue)
						NameValue=ModifiedValue;
					if(ileproc==0)
						ileproc=1;
					level++;
					sprintf(dpath, "%s\\%s_%d", destpath, funcname, level);
					strcpy(SemGlobal->actroot, funcname);
					if(ProcessParse(dpath, level, prname)==-1)
						return -1;
					if(*prname!='\\')
						fprintf(dest,
							"%d\tkeyword:\tprocess\n%d\tidentifier:\t%s\n",
							PROCESSV, NameValue, prname);
					else
					{
						fprintf(dest,
							"%d\tkeyword:\tprocess\n%d\tidentifier:\t%s\n",
							PROCESSV, SemGlobal->LastLexemId+200, SemGlobal->LastFoundName);
						WriteLastLexemId(SemGlobal->LastFoundName);
					}
					level--;
					continue;				   
				}
				else//prawdopodobnie nazwa componentu lub zwykla etykieta
				{
					fprintf(dest, "%s", ProcessName);
					fprintf(dest, "113\tspecial_character:\t:\n");
					value=getvalue(bufor, dest);
					if(value>200 || value==COMPONV)
					{
						if(value==COMPONV)
						{
							if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
							{
								fprintf(dest, "%d", value);
								fprintf(dest, "\tidentifier:\t:%s\n", componname);
								semerror(ERRUNEXEOF, NULL);
								break;
							}
							value=getvalue(bufor, NULL);
							fprintf(dest, "17\tkeyword:\tcomponent\n");
						}
						getname(bufor, componname);
						if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
						{
							fprintf(dest, "%d", value);
							fprintf(dest, "\tidentifier:\t:%s\n", componname);
							semerror(ERRUNEXEOF, NULL);
							break;
						}
						fprintf(dest, "%d", value);
						fprintf(dest, "\tidentifier:\t%s\n", componname);
						value=getvalue(bufor, dest);
						if(value!=PORTV)
						{
							fprintf(dest, "%s", bufor);
							continue;
						}
						else
							fprintf(dest, "%s", bufor);
						if(!fgets(bufor, BUFSIZE, SemGlobal->srcf))
						{
							semerror(ERRUNEXEOF, NULL);
							break;
						}
						value=getvalue(bufor, dest);
						if(value!=MAPV)
						{							
							sprintf(errmesg, ": %s", bufor);
							semerror(ERRUNSCONSTRUCT, errmesg);
						}
						else
							fprintf(dest, "%s", bufor);
						CopyToSemicolon(dest);
						NewComponentName(componname);
						continue;
					}
				}
			}
			else
			{
				AliasCheckAndReplace(ProcessName);
				fprintf(dest, "%s", ProcessName);
				fprintf(dest, "%s", bufor);
				continue;
			}
		}
		if(value==PROCESSV)//process bez nazwy
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d", destpath, funcname, level);
			strcpy(SemGlobal->actroot, funcname);
			if(NoNamedProcessParse(dpath, level)==-1)
				return -1;
			fprintf(dest,
				"%d\tkeyword:\tprocess\n%d\tidentifier:\t%s\n",
				PROCESSV, SemGlobal->LastLexemId+200, SemGlobal->LastFoundName);
			WriteLastLexemId(SemGlobal->LastFoundName);
			level--;
			continue;
		}
		if(value==FORV || value==IFV)//generate bez nazwy
		{
			if(ileproc==0)
				ileproc=1;
			level++;
			sprintf(dpath, "%s\\%s_%d", destpath, funcname, level);
			strcpy(SemGlobal->actroot, funcname);
			if(NoNamedGenerateParse(dpath, level, bufor)==-1)
				return -1;
			fprintf(dest,
				"%d\tkeyword:\tgenerate\n%d\tidentifier:\t%s\n",
				GENERATEV, SemGlobal->LastLexemId+200, SemGlobal->LastFoundName);
			WriteLastLexemId(SemGlobal->LastFoundName);
			level--;
			continue;
		}
		if(value==ASSERTV || value==ATTRIBV)
		{
			ReadToSemiC();
			continue;
		}
		if(value==ENDV)
		{
			if(!ileicf)
				ilebegin--;
			else
			{
				fprintf(dest, "%s", bufor);
				fgets(bufor, BUFSIZE, SemGlobal->srcf);
				value=getvalue(bufor, dest);
				if(value==IFV || value==CASEV || value==FORV || value==LOOPV || value==SEMICV || value==GENERATEV)
					ileicf--;
				fprintf(dest, "%s", bufor);
				continue;
			}
		}
		if(ilebegin==0)
			break;
		if(value==BEGINV)
			ilebegin++;
		if(value==IFV || value==CASEV || value==FORV || value==LOOPV)
			ileicf++;
		fprintf(dest, "%s", bufor);
	}
	fprintf(dest, "%s", bufor);
	while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
	{
		value=getvalue(bufor, dest);
		if(value==SEMICV)
		{
			fprintf(dest, "%s", bufor);
			break;
		}
		fprintf(dest, "%s", bufor);
	}
	fclose(dest);
  if(hasTypes)
    enumsdef2enumsid(defpath, idpath);

	strcpy(SemGlobal->LastFoundName, prname);
	SemGlobal->LastFoundValue=prnumber;
	//SemGlobal->LastFunctionNo=prnumber;
	SemGlobal->varlevel--;
	return 0;
}//BlockParse



int Zasiegi(char *srcpath)
{
	char bufor[BUFSIZE];
	unsigned value;
	//odczyt do slowa kluczowego architechture, entity itd.
	if((SemGlobal->srcf=fopen(srcpath, "rt")))
	{
		while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
		{
			value=getvalue(bufor, NULL);
			if(value==ENTITYV)
			{
				strcpy(SemGlobal->actroot, "toplevel");
				if(entityparse("toplevel", 1)==-1)
					return -1;
			}
			if(value==ARCHV)
			{
				strcpy(SemGlobal->actroot, "toplevel");
				if(archparse("toplevel", 1)==-1)
					return -1;
			}
			if(value==CONFV)
			{
				if(configparse()==-1)
					return -1;
			}
			if(value==PACKV)
			{
				if(packparse("library", 1)==-1)
					return -1;
			}
			if(value==USEV)
			{
				if(UseParse()==-1)
					return -1;
			}
			if(value==ASSERTV || value==ATTRIBV)
			{
				ReadToSemiC();
				continue;
			}
		}
		fclose(SemGlobal->srcf);
	}
	else
		return -1;
	return 0;
}//Zasiegi
int UsePreParseInPackage(void)
{
	return 0;
}//UsePreParseInPackage


int AddToPackages(char *destpath, int level)
{
	char packname[NAMESIZE], bufor[BUFSIZE];
	int value, LastWasFor=0;
	unsigned ilebegin=0, ileicf=0;//wspolny licznik dla if, case i for
	//na poczatku pakietu znajduje sie jego nazwa, ktora bedzie nazwa pliku
	fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt nazwy pakietu lub slowa body
	value=getvalue(bufor, NULL);
	if(value==BODYV)
	{
		//!!!Dodatkowy mechanizm sprawdzania czy jest plik z wartosciami
		//semantycznymi dla pakietu
		fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt nazwy pakietu
		getname(bufor, packname);
		fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt is
		while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
		{
			value=getvalue(bufor, NULL);
			if(value==USEV)
			{
				if(UsePreParseInPackage()==-1)
					return -1;
			}
			if(value==ENDV)
			{
				LastWasFor=0;
				if(!ileicf)
					ilebegin--;
				else
				{
					fgets(bufor, BUFSIZE, SemGlobal->srcf);
					value=getvalue(bufor, NULL);
					if(value==IFV || value==CASEV || value==FORV || value==SEMICV || value==LOOPV || value==GENERATEV)
						ileicf--;
					continue;
				}
			}
			if(ilebegin==0)
				break;
			if(value==BEGINV)
			{
				LastWasFor=0;
				ilebegin++;
			}
			if(value==IFV || value==CASEV || value==FORV || value==LOOPV)
			{
				if(LastWasFor && value==LOOPV)
					ileicf--;
				LastWasFor=0;
				ileicf++;
			}
			if(value==FORV)
				LastWasFor=1;
		}
	}
	else
	{
		FILE *packsf;
#define WORKPACKAGES "packages.pck"
		getname(bufor, packname);
		packsf=fopen(WORKPACKAGES, "at");
		if(packsf==NULL)
			return -1;
		fprintf(packsf, "work.%s.all\tlibrary\\%s\\%s\n", packname, packname, packname);
		fclose(packsf);
		strcpy(SemGlobal->rootsname, packname);
		fgets(bufor, BUFSIZE, SemGlobal->srcf);//odczyt is
		while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
		{
			value=getvalue(bufor, NULL);
			//!!!!!!//pomijanie tylko czesci deklaracyjnej procedury lub funkcji
			if(value==USEV)
				if(UsePreParseInPackage()==-1)
					return -1;
				if(value==ENDV)
					break;
		}
	}
	return 0;
}//AddToPackages


int FindPackage(char *srcpath)
{
	char bufor[BUFSIZE];
	unsigned value;
	if((SemGlobal->srcf=fopen(srcpath, "rt")))
	{
		while(fgets(bufor, BUFSIZE, SemGlobal->srcf))
		{
			value=getvalue(bufor, NULL);
			if(value==PACKV)
			{
				if(AddToPackages("library", 1)==-1)
					return -1;
				SemGlobal->PackFound=1;
			}
		}
		fclose(SemGlobal->srcf);
	}
	else
		return -1;
	return 0;
}//FindPackages



int LexFiles(void)
{
	char bufor[BUFSIZE];
	int len, i;
	for(i=0 ; i<SemGlobal->AllFilesCnt ; i++)
	{
		strcpy(bufor, SemGlobal->AllFiles[i].path);
		len=strlen(bufor)-4;
		strncpy(SemGlobal->actroot, bufor, len);
		SemGlobal->actroot[len]=0;
		strncpy(SemGlobal->actfile, bufor, len);
		SemGlobal->actfile[len]=0;
		strcat(SemGlobal->actfile, ".vhd");
		if(Zasiegi(bufor)==-1)
			return -1;
	}
	return 0;
}//lexfiles


int LookForPackages(void)
{
	FILE *designf;
	char bufor[BUFSIZE];
	int len, i;
	chdir("work");
	designf=fopen("design.dgn", "rt");
	while(fgets(bufor, BUFSIZE, designf))
	{
		len=strlen(bufor)-5;
		strncpy(SemGlobal->actroot, bufor, len);
		strncpy(SemGlobal->actfile, bufor, len+4);
		SemGlobal->actfile[len+4]=0;
		SemGlobal->actroot[len]=0;
		sprintf(bufor+len, ".lex");
		SemGlobal->PackFound=0;
		if(FindPackage(bufor)==-1)
		{
			fclose(designf);
			return -1;
		}
		if(SemGlobal->PackFound)
		{
			if(SemGlobal->PackFilesCnt==0)
			{
				SemGlobal->PackFilesCnt=1;
				SemGlobal->PackFiles=(ALLFILES*)malloc(sizeof(ALLFILES));
			}
			else
			{
				SemGlobal->PackFilesCnt++;
				SemGlobal->PackFiles=(ALLFILES*)realloc((ALLFILES*)SemGlobal->PackFiles, SemGlobal->PackFilesCnt*sizeof(ALLFILES));
			}
			SemGlobal->PackFiles[SemGlobal->PackFilesCnt-1].path=(char*)malloc((strlen(bufor)+1)*sizeof(char));
			strcpy(SemGlobal->PackFiles[SemGlobal->PackFilesCnt-1].path, bufor);
		}
		else
		{
			if(SemGlobal->NoPackFilesCnt==0)
			{
				SemGlobal->NoPackFilesCnt=1;
				SemGlobal->NoPackFiles=(ALLFILES*)malloc(sizeof(ALLFILES));
			}
			else
			{
				SemGlobal->NoPackFilesCnt++;
				SemGlobal->NoPackFiles=(ALLFILES*)realloc((ALLFILES*)SemGlobal->NoPackFiles, SemGlobal->NoPackFilesCnt*sizeof(ALLFILES));
			}
			SemGlobal->NoPackFiles[SemGlobal->NoPackFilesCnt-1].path=(char*)malloc((strlen(bufor)+1)*sizeof(char));
			strcpy(SemGlobal->NoPackFiles[SemGlobal->NoPackFilesCnt-1].path, bufor);
		}
	}
	fclose(designf);
	SemGlobal->AllFilesCnt=SemGlobal->NoPackFilesCnt+SemGlobal->PackFilesCnt;
	SemGlobal->AllFiles=(ALLFILES*)malloc(SemGlobal->AllFilesCnt*sizeof(ALLFILES));
	for(i=0 ; i<SemGlobal->PackFilesCnt ; i++)
	{
		SemGlobal->AllFiles[i].path=(char*)malloc((strlen(SemGlobal->PackFiles[i].path)+1)*sizeof(char));
		strcpy(SemGlobal->AllFiles[i].path, SemGlobal->PackFiles[i].path);
		free(SemGlobal->PackFiles[i].path);
	}
	free(SemGlobal->PackFiles);
	for(i=0 ; i<SemGlobal->NoPackFilesCnt ; i++)
	{
		SemGlobal->AllFiles[i+SemGlobal->PackFilesCnt].path=(char*)malloc((strlen(SemGlobal->NoPackFiles[i].path)+1)*sizeof(char));
		strcpy(SemGlobal->AllFiles[i+SemGlobal->PackFilesCnt].path, SemGlobal->NoPackFiles[i].path);
		free(SemGlobal->NoPackFiles[i].path);
	}
	free(SemGlobal->NoPackFiles);
	return 0;
}//LookForPackages

int RecordBitCnt(int msb, int lsb)
{
	int rsb, ValidBits;
	char tokstr[BUFSIZE];
	if(msb>lsb)
	{
		rsb=lsb;
		lsb=msb;
		msb=rsb;
	}
	ValidBits=0;
	for(rsb=msb ; rsb<=lsb ; rsb++)
	{
		int r1, r2, r3, r4, r5, r6;
		strcpy(SemGlobal->tmpactpath, SemGlobal->actpath);
		if(FindLexemInformation(rsb, tokstr))
		{
			sscanf(tokstr, "%d\t%d\t%d\t%d\t%d\t%d\t", &r1, &r2, &r3, &r4, &r5, &r6);
			ValidBits+=r6;
		}
		if(rsb==lsb)
			break;
	}
	return ValidBits;
}


void WritePortInfo(FILE *topfile, char *sporttype, int msb, int lsb, char *portname, int TypeOf, int ValidBits)
{
	char typestr[BUFSIZE], tokstr[BUFSIZE];
	switch(TypeOf)
	{
	case 202:
	case 204:
	case 209:
	case 210:
	case 215:
	case 216:
	case 217:
	case 218:
	case 219:
		fprintf(topfile, "\n%s %s", sporttype, portname);
		break;
	case 207:
	case 208:
	case 201:
		if(ValidBits==1)
		{
			fprintf(topfile, "\n%s %s", sporttype, portname);
		}	
		else
		{
			if(ValidBits==32)
				fprintf(topfile, "\n%s [31:0] %s", sporttype, portname);
			else
				fprintf(topfile, "\n%s [%d:%d] %s", sporttype, HowManyBitsInt(msb, lsb)-1, 0, portname);
		}
		break;
	case 205:
		fprintf(topfile, "\n%s [0:7] %s", sporttype, portname);
		break;
	case 206:
		fprintf(topfile, "\n%s [%d:%d][0:7] %s", sporttype, msb, lsb, portname);
		break;
	case 203:
	case 211:
	case 212:
	case 213:
	case 214:
		fprintf(topfile, "\n%s [%d:%d] %s", sporttype, msb, lsb, portname);
		break;
	default:
		strcpy(SemGlobal->tmpactpath, SemGlobal->actpath);
		if(FindTypeString(TypeOf, tokstr, typestr))
		{
			if(msb>300 && lsb>300 && TypeOf!=203 && TypeOf!=211 && TypeOf!=212 && TypeOf!=213 && TypeOf!=214)
			{
				ValidBits=RecordBitCnt(msb, lsb);
				fprintf(topfile, "\n%s [%d:%d] %s", sporttype, ValidBits-1, 0, portname);
			}
			else
			{
				if(ValidBits+1!=lsb-msb && ValidBits+1!=msb-lsb)
				{
					if(msb==0 || lsb==0)
						fprintf(topfile, "\n%s [0:%d] %s", sporttype, ValidBits-1, portname);
					else
						fprintf(topfile, "\n%s [%d:%d] %s", sporttype, msb, lsb, portname);
				}
				else
					fprintf(topfile, "\n%s [0:%d] %s", sporttype, ValidBits-1, portname);
			}
		}
		else
		{
			if(msb>300 && lsb>300 && TypeOf!=203 && TypeOf!=211 && TypeOf!=212)
				fprintf(topfile, "\n%s [%d:%d] %s", sporttype, 0, ValidBits-1, portname);
			else
				fprintf(topfile, "\n%s [%d:%d] %s", sporttype, msb, lsb, portname);
		}
		break;
	}
}//WritePortInfo


void WritePortExtendedInfo(FILE *topfile, int msb, int lsb, int TypeOf, int ValidBits)
{
	char typestr[BUFSIZE], tokstr[BUFSIZE];
	char tostr[8];
	fprintf(topfile, " :");
	if(msb<lsb)
		strcpy(tostr, "to");
	else
		strcpy(tostr, "downto");
	switch(TypeOf)
	{
	case 201:
		if(ValidBits==32)
			fprintf(topfile, " integer");
		else
			fprintf(topfile, " %s %d %s %d", "integer range", msb, tostr, lsb);
		break;
	case 202:
		fprintf(topfile, "%s", " bit");
		break;
	case 203:
		fprintf(topfile, " %s (%d %s %d)", "bit_vector", msb, tostr, lsb);
		break;
	case 204:
		fprintf(topfile, "%s", " boolean");
		break;
	case 205:
		fprintf(topfile, "%s (0 to 7)", " character");
		break;
	case 206:
		fprintf(topfile, " %s (%d %s %d)"/*(0 to 7)*/, "string", msb, tostr, lsb);
		break;
	case 207:
		if(ValidBits==31)
			fprintf(topfile, " natural");
		else
			fprintf(topfile, " %s %d %s %d", "natural range", msb, tostr, lsb);
		break;
	case 208:
		if(ValidBits==31)
			fprintf(topfile, " positive");
		else
			fprintf(topfile, " %s %d %s %d", "positive range", msb, tostr, lsb);
		break;
	case 209:
		fprintf(topfile, "%s", " std_ulogic");
		break;
	case 210:
		fprintf(topfile, "%s", " std_logic");
		break;
	case 211:
		fprintf(topfile, " %s (%d %s %d)", "std_ulogic_vector", msb, tostr, lsb, "std_ulogic_vector");
		break;
	case 212:
		fprintf(topfile, " %s (%d %s %d)", "std_logic_vector", msb, tostr, lsb);
		break;
	case 213:
		fprintf(topfile, " %s (%d %s %d)", "signed", msb, tostr, lsb);
		break;
	case 214:
		fprintf(topfile, " %s (%d %s %d)", "unsigned", msb, tostr, lsb);
		break;
	case 219:
		fprintf(topfile, "%s", " X01");
		break;
	case 218:
		fprintf(topfile, "%s", " X01Z");
		break;
	case 216:
		fprintf(topfile, "%s", " UX01");
		break;
	case 217:
		fprintf(topfile, "%s", " UX01Z");
		break;
	default:
		strcpy(SemGlobal->tmpactpath, SemGlobal->actpath);
		if(FindTypeString(TypeOf, tokstr, typestr))
		{
			if(msb>300 && lsb>300)
			{
				ValidBits=RecordBitCnt(msb, lsb);
				fprintf(topfile, " %s (%d %s %d)", SemGlobal->LastTypeName, ValidBits-1, "downto", 0);
			}
			else
			{
				if(ValidBits+1!=lsb-msb && ValidBits+1!=msb-lsb)
				{
					if(msb==0 || lsb==0)
						fprintf(topfile, " %s (0 %s %d)", SemGlobal->LastTypeName, "to", ValidBits-1);
					else
						fprintf(topfile, " %s (%d %s %d)", SemGlobal->LastTypeName, msb, tostr, lsb);
				}
				else
					fprintf(topfile, " %s (%d %s %d)", SemGlobal->LastTypeName, 0, tostr, ValidBits-1);
			}
		}
		else
		{
			if(msb>300 && lsb>300)
				fprintf(topfile, " %s (%d %s %d)", "USERTYPE", 0, tostr, ValidBits-1);
			else
				fprintf(topfile, " %s (%d %s %d)", "USERTYPE", msb, tostr, lsb);
		}
		break;
	}
}//WritePortExtendedInfo



void WriteEntityPortDescription(FILE *topfile, int i)
{
	FILE *entityfile;
	char entitypath[256], bufor[BUFSIZE], portname[256], sporttype[10];
	int oldporttype, oldmsb, oldlsb, oldTypeOf=0, Same=0, NotFirst=0, TypeOf=0;
	int lexval, porttype, msb, lsb, ValidBits=0, res4, IsSemValues, j, What=0;
	if(!SemGlobal->entities[i].isentity)//Brak jednostki o nazwie podanej w architekturze
		return;
	sprintf(entitypath, "toplevel\\%s.val", SemGlobal->entities[i].name);
	entityfile=fopen(entitypath, "rt");
	IsSemValues=0;
	fprintf(topfile, "entity %s", SemGlobal->entities[i].name);
	while(fgets(bufor, BUFSIZE, entityfile))
	{
		if(IsSemValues)
		{
			sscanf(bufor, "%d\t%s\t%d\t%d\t%d\t%d\t%d\t%d\t%d", &lexval, portname, &TypeOf, &What, &porttype, &ValidBits, &msb, &lsb, &res4);
			//dla wszystkich portow danej jednostki
			if(What==GENERICLV)
				continue;
			for(j=0 ; j<SemGlobal->entities[i].portcnt ; j++)
			{
				if(SemGlobal->entities[i].portids[j]==lexval)
				{
					if(SemGlobal->entities[i].isbuffers[j])
						porttype=4;
					switch(porttype)
					{
					case 1:strcpy(sporttype, "input");break;
					case 2:strcpy(sporttype, "output");break;
					case 3:strcpy(sporttype, "inout");break;
					case 4:strcpy(sporttype, "inout");break;
					default:strcpy(sporttype, "linkage");break;
					}
					if(NotFirst)
					{
						Same=0;
						if(oldporttype==porttype && oldmsb==msb && oldlsb==lsb && TypeOf==oldTypeOf)
							Same=1;
					}
					if(Same)
					{
						fprintf(topfile, ", %s", portname);
					}
					else
					{
						WritePortInfo(topfile, sporttype, msb, lsb, portname, TypeOf, ValidBits);
						oldporttype=porttype;
						oldmsb=msb;
						oldlsb=lsb;
						oldTypeOf=TypeOf;
					}
					NotFirst=1;
					break;
				}
			}
			continue;
		}
		if(bufor[0]=='#' && bufor[2]=='5')
		{
			IsSemValues=1;
		}
	}
	fprintf(topfile, "\n");
	fclose(entityfile);
}//WriteEntityPortDescription



void WriteEntityPortExtendedDescription(FILE *topfile, int i)
{
	FILE *entityfile;
	char entitypath[256], bufor[BUFSIZE], portname[256], sporttype[10];
	int oldporttype, oldmsb, oldlsb, Same=0, NotFirst=0, TypeOf=0, oldTypeOf;
	int lexval, porttype, msb, lsb, What=0, res4, IsSemValues, j;
	int ValidBits=0, oldValidBits=0;
	if(!SemGlobal->entities[i].isentity)//Brak jednostki o nazwie podanej w architekturze
		return;
	sprintf(entitypath, "toplevel\\%s.val", SemGlobal->entities[i].name);
	entityfile=fopen(entitypath, "rt");
	IsSemValues=0;
	fprintf(topfile, "entity %s", SemGlobal->entities[i].name);
	fprintf(topfile, " (%s)", SemGlobal->entities[i].archname);
	while(fgets(bufor, BUFSIZE, entityfile))
	{
		if(IsSemValues)
		{
			sscanf(bufor, "%d\t%s\t%d\t%d\t%d\t%d\t%d\t%d\t%d", &lexval, portname, &TypeOf, &What, &porttype, &ValidBits, &msb, &lsb, &res4);

			if(What==GENERICLV)
				continue;
			//dla wszystkich portow danej jednostki
			for(j=0 ; j<SemGlobal->entities[i].portcnt ; j++)
			{
				if(SemGlobal->entities[i].portids[j]==lexval)
				{
					if(SemGlobal->entities[i].isbuffers[j])
						porttype=4;
					switch(porttype)
					{
					case 1:strcpy(sporttype, "input");break;
					case 2:strcpy(sporttype, "output");break;
					case 3:strcpy(sporttype, "inout");break;
					case 4:strcpy(sporttype, "inout");break;
					default:strcpy(sporttype, "linkage");break;
					}
					if(NotFirst)
					{
						Same=0;
						if(oldporttype==porttype && oldmsb==msb && oldlsb==lsb && oldTypeOf==TypeOf && ValidBits==oldValidBits)
							Same=1;
					}
					if(Same)
					{
						fprintf(topfile, ", %s", portname);
					}
					else
					{
//						if(!NotFirst)
//							fprintf(topfile, "%s", portname);
						if(NotFirst)
							WritePortExtendedInfo(topfile, oldmsb, oldlsb, oldTypeOf, oldValidBits);
						oldporttype=porttype;
						oldmsb=msb;
						oldlsb=lsb;
						oldTypeOf=TypeOf;
						oldValidBits=ValidBits;
						fprintf(topfile, "\n%s:\t", sporttype);
						fprintf(topfile, "%s", portname);

					}
					NotFirst=1;
					break;
				}
			}
			continue;
		}
		if(bufor[0]=='#' && bufor[2]=='5')
		{
			IsSemValues=1;
		}
	}
	if(NotFirst)
		WritePortExtendedInfo(topfile, oldmsb, oldlsb, oldTypeOf, oldValidBits);
	fprintf(topfile, "\n");
	fclose(entityfile);
}//WriteEntityPortExtendedDescription


int IsInToplevelEntity(char *path, char *topentname)
{
	char tmppath[PATHSIZE];
	int i,j;
	for(i=9, j=0 ; path[i]!='\\' && path[i]!=0 ; i++, j++)
		tmppath[j]=path[i];
	tmppath[j]=0;
	if(path[i]!=0)//nie byla to sama jednostka
	{
		if(j>0)
			while(tmppath[j--]!='_' && j>0);
		tmppath[j+1]=0;
	}
	for(i=0 ; i<SemGlobal->ileencji ; i++)
	{
		if(SemGlobal->entities[i].isTopLevel && !stricmp(SemGlobal->entities[i].name, tmppath))
		{
			strcpy(topentname, tmppath);
			return 1;
		}
	}
	return 0;
}//IsInToplevelEntity


int IsInEntity(char *path, char *topentname)
{
	char tmppath[PATHSIZE];
	int i,j;
	for(i=9, j=0 ; path[i]!='\\' && path[i]!=0 ; i++, j++)
		tmppath[j]=path[i];
	tmppath[j]=0;
	if(path[i]!=0)//nie byla to sama jednostka
	{
		if(j>0)
			while(tmppath[j--]!='_' && j>0);
		tmppath[j+1]=0;
	}
	for(i=0 ; i<SemGlobal->ileencji ; i++)
	{
		if(SemGlobal->entities[i].isTopLevel)
		{
			strcpy(topentname, SemGlobal->entities[i].name);
			return 1;
		}
	}
	return 0;
}//IsInEntity

int IsEntityHasResolvFile(char *topentname)
{
	int i;
	for(i=0 ; i<SemGlobal->ileencji ; i++)
	{
		if(!stricmp(SemGlobal->entities[i].name, topentname))
		{
			if(SemGlobal->entities[i].HasResolvFile)
				break;
			else
			{
				SemGlobal->entities[i].HasResolvFile=1;
				return 0;
			}
		}
	}
	return 1;
}

int WriteResolvedVariables(void)
{
	int IsFirst=0, nr, type, IsSemValues=0;
	int lexval, res1, TypeOf, porttype, res3, msb, lsb, res4, i;
	char bufor[BUFSIZE], name[NAMESIZE], path[PATHSIZE], portname[NAMESIZE], topentname[NAMESIZE];
	FILE *idxffile, *resfile, *valfile;
	idxffile=fopen("index.idx", "rt");
	if(!idxffile)
		return -1;
	while(fgets(bufor, BUFSIZE, idxffile))
	{
		sscanf(bufor, "%d %d %s\t%s", &nr, &type, name, path);
		//if(IsInToplevelEntity(path, topentname))
		if(IsInEntity(path, topentname))
		{
			strcat(path, ".val");
			valfile=fopen(path, "rt");
			if(!valfile)
				return -1;
			if(IsEntityHasResolvFile(topentname))
			{
				strcat(topentname, ".res");
				resfile=fopen(topentname, "at");
			}
			else
			{
				strcat(topentname, ".res");
				resfile=fopen(topentname, "wt");
			}
			while(fgets(bufor, BUFSIZE, valfile))
			{
				if(bufor[0]=='#')
				{
					if(bufor[2]=='5')
						IsSemValues=1;
					else
						IsSemValues=0;
					continue;
				}
				if(IsSemValues)
				{
					sscanf(bufor, "%d\t%s\t%d\t%d\t%d\t%d\t%d\t%d\t%d", &lexval, portname, &res1, &TypeOf, &porttype, &res3, &msb, &lsb, &res4);
					if(TypeOf==0)
						continue;
					if(res1==210)
						fprintf(resfile, "%s THREE_STATE 0;\n", portname);
					if(res1==212)
					{
						if(res3==32)
							fprintf(resfile, "%s THREE_STATE %d;\n", portname, res3);
						else
						{
							if(msb>lsb)
								fprintf(resfile, "%s THREE_STATE %d;\n", portname, 1+msb-lsb);
							else
								fprintf(resfile, "%s THREE_STATE %d;\n", portname, 1+lsb-msb);
						}
					}
					/* Oddzielnie w celu umozliwienia pozniejszych zmian */
					if(res1==213)
						fprintf(resfile, "%s THREE_STATE %d;\n", portname, res3);
					if(res1==214)
						fprintf(resfile, "%s THREE_STATE %d;\n", portname, res3);
					/* To samo dla typow UX01, UX01Z, X01Z, X01 */
					if(res1>=216 && res1<=219)
						fprintf(resfile, "%s THREE_STATE 0;\n", portname);
					for(i=0 ; i<SemGlobal->ResolvedCnt ; i++)
					{
						if(res1==SemGlobal->ListOfResolved[i].TypeNo)
						{
							switch(SemGlobal->ListOfResolved[i].ResMethod)
							{
							case WIREDOR:fprintf(resfile, "%s OR 0;\n", portname);break;
							case WIREDAND:fprintf(resfile, "%s AND 0;\n", portname);break;
							case THREESTATE:
							default:	fprintf(resfile, "%s THREE_STATE 0;\n", portname);
							}
						}
					}
				}
			}
			fclose(resfile);
			fclose(valfile);
		}
		IsFirst=0;
	}
	return 0;
}

int EntityAdj(void)
{
	int i, j, k;
	FILE *adjfile, *topfile;
	if(SemGlobal->ileencji==0)
	{
		semwarn("There is no entity defined");
		return 0;
	}
	if((adjfile=fopen("entity.adj", "wt"))==NULL)
		return -1;
	for( i=0 ; i<SemGlobal->configcnt ; i++)//Dla wszystkich konfiguracji
	{
		for( j=0 ; j<SemGlobal->archcnt ; j++)//Wszystkich architektur
		{
			if(!stricmp(SemGlobal->archs[j].entityname, SemGlobal->configs[i].entityname))//szukamy odpowiedniej architektury
			{
				for( k=0 ; k<SemGlobal->ileencji ;k++)//i dla wszystkich jednostek
				{
					if(!stricmp(SemGlobal->archs[j].entityname, SemGlobal->entities[k].name)
						&&
						!stricmp(SemGlobal->archs[j].archname, SemGlobal->configs[i].archname)
					  )//wyszukujemy ta, dla ktorej byla konfiguracja
					{
						fprintf(adjfile, "%d:%s %d:%s\n", 
							SemGlobal->entities[k].entityno,
							SemGlobal->entities[k].name,
							SemGlobal->archs[j].archno,
							SemGlobal->archs[j].archname);
						SemGlobal->entities[k].isconfig=1;
					}
				}
			}
		}
	}
	for( i=0 ; i<SemGlobal->ileencji ; i++)
	{
		if(!SemGlobal->entities[i].isconfig && SemGlobal->entities[i].isentity && SemGlobal->entities[i].isarch)
		{
			for( j=0 ; j<SemGlobal->archcnt ; j++)
			{
				if(!stricmp(SemGlobal->archs[j].entityname, SemGlobal->entities[i].name)
					&&
					!stricmp(SemGlobal->archs[j].archname, SemGlobal->entities[i].archname)
					)
				{
					
					fprintf(adjfile, "%d:%s %d:%s\n", 
						SemGlobal->entities[i].entityno,
						SemGlobal->entities[i].name,
						SemGlobal->archs[j].archno,
						SemGlobal->archs[j].archname);
				}
			}
		}
	}
	fclose(adjfile);
	//petla eliminujaca jednostki nie bedace typu TopLevel
	for(i=0 ; i<SemGlobal->componcnt ; i++)
	{
		for(j=0 ; j<SemGlobal->ileencji ; j++)
		{
			if(!stricmp(SemGlobal->entities[j].name, SemGlobal->components[i].name))
			{
				SemGlobal->entities[j].isTopLevel=0;
				break;
			}
		}
	}
	if((topfile=fopen("topent.idx", "wt"))==NULL)
		return -1;
	for(i=0 ; i<SemGlobal->ileencji ; i++)
	{
		if(SemGlobal->entities[i].isTopLevel)
			WriteEntityPortDescription(topfile, i);
	}
	fclose(topfile);
	if((topfile=fopen("topentex.idx", "wt"))==NULL)
		return -1;
	for(i=0 ; i<SemGlobal->ileencji ; i++)
	{
		if(SemGlobal->entities[i].isTopLevel)
			WriteEntityPortExtendedDescription(topfile, i);
	}
	fclose(topfile);
	return 0;
}//EntityAdj


int UseFinal(void)
{
	int i, found;
	char bufor[BUFSIZE], *token;
	FILE *dest, *packfile=NULL, *syspackfile=NULL;
	for(i=0 ; i<SemGlobal->nfoundcnt ; i++)
	{
		if( ((packfile=fopen("packages.pck", "rt"))==NULL) && ((syspackfile=fopen("..\\syspack.pck", "rt"))==NULL))
			return -1;
		if((dest=fopen(SemGlobal->usesnfound[i].path, "at"))==NULL)
			return -1;
		//!!!!!!!!!!!!!!!!!!!!!!!!!!
		//Obsluzyc przypadek gdy w jednym pliku maja byc zapisane 2 use-clause
		//albo nie bedzie zadnego
		fprintf(dest, "#7 Dodatkowe sciezki do pakietow z work\n");
		found=0;
		if(packfile!=NULL)
		{
			while(fgets(bufor, BUFSIZE, packfile))
			{
				token=strtok(bufor, " \t");
				if(!stricmp(token, SemGlobal->usesnfound[i].useclause))
				{
					token=strtok(NULL, "\n");
					fprintf(dest, "%s\n", token);
					found=1;
					break;
				}
			}
			fclose(packfile);
		}
		if(syspackfile!=NULL)
		{
			if(!found)
			{
				while(fgets(bufor, BUFSIZE, syspackfile))
				{
					token=strtok(bufor, " \t");
					if(!stricmp(token, SemGlobal->usesnfound[i].useclause))
					{
						token=strtok(NULL, "\n");
						fprintf(dest, "%s\n", token);
						found=1;
						break;
					}
				}
			}
			fclose(syspackfile);
		}
		if(!found)
		{
			char errmesg[BUFSIZE];
			SemGlobal->lineno=SemGlobal->usesnfound[i].lineno;
			strcpy(SemGlobal->actfile, SemGlobal->usesnfound[i].filename);
			sprintf(errmesg, ": %s", SemGlobal->usesnfound[i].useclause);
			semerror(ERRNFPACKAGE, errmesg);
		}
		fclose(dest);
	}
	if(SemGlobal->nfoundcnt)
		free(SemGlobal->usesnfound);
	return 0;
}//UseFinal

int FindLastLexemId(void)
{
	FILE *varfile;
	int i;
	char bufor[BUFSIZE];
	if((varfile=fopen("alu.var", "rt"))==NULL)
		return -1;
	while(fgets(bufor, BUFSIZE, varfile));
	sscanf(bufor, "%d", &i);
	SemGlobal->LastLexemId=i+1;
	return 0;
}//FindLastLexemId

void InitializeSemGlobal(void)
{
	SemGlobal->lineno=1;
	SemGlobal->errcnt=0;
	SemGlobal->configcnt=0;
	SemGlobal->configs=NULL;
	SemGlobal->archcnt=0;
	SemGlobal->archs=NULL;
	SemGlobal->ileencji=0;
	SemGlobal->entities=NULL;
	SemGlobal->componcnt=0;
	SemGlobal->components=NULL;
	SemGlobal->actcompo[0]=0;
	SemGlobal->pathcnt=0;
	SemGlobal->usepaths=NULL;
	SemGlobal->nfoundcnt=0;
	SemGlobal->indexf=fopen("work\\index.idx", "wt");
	SemGlobal->indexno=0;
	SemGlobal->varlevel=0;
	SemGlobal->isIEEE1164avail=0;
	SemGlobal->hasTypes=0;
	SemGlobal->hasNonComputables=0;
	SemGlobal->NoPackFilesCnt=0;
	SemGlobal->NoPackFiles=NULL;
	SemGlobal->PackFilesCnt=0;
	SemGlobal->PackFiles=NULL;
	SemGlobal->AllFiles=NULL;
	SemGlobal->LastFunctionNo=0;
	SemGlobal->LastLexemId=0;
	SemGlobal->protoscnt=0;
	SemGlobal->fprotos=NULL;
	SemGlobal->defaultmode=3;/* default mode INOUT */
	SemGlobal->ListForAlias=NULL;
	SemGlobal->FullListForAlias=NULL;
	SemGlobal->ListOfPrcs=NULL;
	SemGlobal->ListOfBlocks=NULL;
	SemGlobal->ListOfGenerics=NULL;
	SemGlobal->ListOfYY=NULL;
	SemGlobal->ProcessCnt=0;
	SemGlobal->BlockCnt=0;
	SemGlobal->isInPackage=0;
	SemGlobal->isInValFile=0;
	SemGlobal->IsSTLACompiled=0;
	SemGlobal->IsParentNonComputable=0;
	SemGlobal->IsResolvedFunction=0;
	SemGlobal->AliasCnt=0;
	SemGlobal->NCofNC=0;
	SemGlobal->LastResolvedType=0;
	SemGlobal->ResolvedCnt=0;
	SemGlobal->ListOfResolved=NULL;
	SemGlobal->isComponentPort=0;
	SemGlobal->TypeCollected=0;
	SemGlobal->GuardCnt=0;
	SemGlobal->LastFullType[0]=0;
	SemGlobal->actEntityContext[0]=0;
	SemGlobal->actProcessContext[0]=0;
	SemGlobal->ModifiersCnt.real=0;
	SemGlobal->ModifiersCnt.last=0;
	SemGlobal->NamesCnt=0;
	SemGlobal->isArchitectureParsed=0;
	SemGlobal->ChangedComcnt=0;
	SemGlobal->errfile=fopen("work\\errors.log", "at");
}

int main(void)
{
	int i;
	struct PBGLOBAL Global;
	SemGlobal=&Global;
	InitializeSemGlobal();
	if(LookForPackages()==-1)
	{
		fclose(SemGlobal->indexf);
		return -1;
	}
	if(FindLastLexemId()==-1)
		return -1;
	if(LexFiles()==-1)
	{
		fclose(SemGlobal->indexf);
		semerror(ERRFILEOPEN, NULL);
		return -1;
	}
	fclose(SemGlobal->indexf);
	if(EntityAdj()==-1)
		return -1;
	if(UseFinal()==-1)
		return -1;
	if(WriteResolvedVariables()==-1)
		return -1;

/*Zwalnianie pamieci*/
	if(SemGlobal->ileencji)
	{
		for(i=0 ; i<SemGlobal->ileencji ; i++)
		{
			if(SemGlobal->entities[i].portcnt)
			{
				;
//				free(SemGlobal->entities[i].portids);
//				free(SemGlobal->entities[i].isbuffers);
			}
		}
		free(SemGlobal->entities);
	}
	if(SemGlobal->ProcessCnt)
	{
		for(i=0 ; i<SemGlobal->ProcessCnt ; i++)
			free(SemGlobal->ListOfPrcs[i].name);
		free(SemGlobal->ListOfPrcs);
	}
	if(SemGlobal->BlockCnt)
	{
		for(i=0 ; i<SemGlobal->BlockCnt ; i++)
			free(SemGlobal->ListOfBlocks[i].name);
		free(SemGlobal->ListOfBlocks);
	}
	if(SemGlobal->pathcnt)
		free(SemGlobal->usepaths);
	if(SemGlobal->protoscnt)
	{
		for(i=0 ; i<SemGlobal->protoscnt ; i++)
		{
			if(SemGlobal->fprotos[i].parcnt)
				free(SemGlobal->fprotos[i].params);
		}
		free(SemGlobal->fprotos);
	}
	if(SemGlobal->AllFilesCnt)
	{
//		for(i=0 ; i<SemGlobal->AllFilesCnt ; i++)
//			free(SemGlobal->AllFiles[i].path);
		free(SemGlobal->AllFiles);
	}
	if(SemGlobal->archs!=NULL)
		free(SemGlobal->archs);
	fclose(SemGlobal->errfile);
	//FreeAllGlobalMem();
	if(!SemGlobal->errcnt)
		printf("Semantic analysis successful\n");
	if(SemGlobal->ileencji==0)
		exit(254);//Prawidlowe wykonanie analizatora, brak toplevel entities
	exit(SemGlobal->errcnt);
	return 0;
}//main
